#ifndef RAWNANOSLICE_H
#define RAWNANOSLICE_H

#include "DAQDataFormats/RawNanoSliceHeader.h"

#define RAWNANOSLICE_SIZE 3

namespace daqdataformats {

  //! Data structure for holding all the nanoslice data of interest
  typedef struct {
    NanoSliceHeader Header;
    uint32_t    TimeStamp;
    uint32_t    PulseHeight;
    uint32_t    Reserved;
  } NanoSlice;

    namespace rawnano {

        static const uint32_t cafe_word = 0xCAFE;

        enum NanoSliceMASKS{
            MASK_TimeStamp    = 0xFFFFFFFF,
            MASK_PulseHeight  = 0x0FFF0000,
            MASK_Reserved     = 0x0000FFFF
        };

        enum NanoSliceSHIFTS{
            SHIFT_TimeStamp   = 0,
            SHIFT_PulseHeight = 16,
            SHIFT_Reserved    = 0
        };

    };// end of namespace rawnano

  //! Class to hold the data from the FEBs in correct formats
  /*!
    The RawNanoSlice class is designed to mimic the format of the actual data
    block that the NOvA Front End6
    Boards (FEBS) produce when they detect a signal
    above the zero suppression threshold.
    
    The class supports general initialization with or with out data blocks as
    well as copyin style initialization for use with STL containers.
    
    Pack and unpack methods for the data rely on two types of structured data
    blocks that are provided in association with the class.
    
    The NanoSliceBlock structure breaks the raw data block out into its related
    high and low half words to give the headers, time stamps and pulseheight information
    
    The NanoSliceHeader structure further breaks out the header blocks into the
    individual data information parts.
  */

  class RawNanoSlice: public RawDAQData {
    
  public:
      
    // Constructors
    RawNanoSlice ();                                //! The default constructor
    RawNanoSlice (NanoSlice nanoslice_parameters);  //! Constructor with Data initialization
    RawNanoSlice (const RawNanoSlice& copy_in);           //! Deep copy constructor
    RawNanoSlice (void* nanoslice_buffer);          //! Shallow copy constructor


    /// Destructor
    ~RawNanoSlice();
    /// Initialization of the Data members
    void init();
    void initwithData(NanoSlice nanoslice_parameters);

    /// Deep Copy
    void  copy(const RawNanoSlice& copy_in);

    /// Print the nanoslice structure including Data and _Buffer
    void print() const;

    // Getters
    RawNanoSliceHeader* getHeader();              //! Get the NanoSlice header
    uint32_t            getPulseHeight() const;   //! Get method for pulseheight information
    uint32_t            getReserved() const;      //! Get the reserved word
    uint32_t            getTimeStamp() const;     //! Get method for time stamp information

    // Setters
    bool setTimeStamp  (uint32_t);  //! Set method for timestamp information
    bool setPulseHeight(uint32_t);  //! Set method for pulse height information
    bool setReserved   (uint32_t);  //! Set method for Reserved word
    bool setParameters (NanoSlice); //! Set method for all parameters of the NanoSlice

    /// Check if the reserved word is 0xcafe
    bool checkReservedForCafe();

    bool setBufferSource (void* pointer); //!< Set The Buffer pointer to a given location. Don't use it if you are not sure
    bool resetBufferSource();             //!< Reset the class to use its internal buffer
    bool clear();                         //!< Clear the buffer
    bool reset();                         //!< Reset the class data
    bool isInternalBuffer() const;              //!< Is the buffer internal?

    /// General sizeof method which returns the size of a DAQ formated data block in words (not bytes)
    uint32_t sizeofdata() const;

  private:
    /// The Data block
    RawNanoSliceHeader  _Header;       //! Header

    /// The actual internal buffer (not a pointer)
    uint32_t  _InternalBuffer[RAWNANOSLICE_SIZE];


  };// end of RawNanoSlice
  
  
} // END NAMESPCE daqdataformats
#endif /* RAWNANOSLICE_H */
