#ifndef RAWTIMINGMARKER_H
#define RAWTIMINGMARKER_H

#include "DAQDataFormats/RawDAQData.h"

#define RAWTIMINGMARKER_SIZE 2

namespace daqdataformats {

  //! Data structure to hold the parts of the timing word
  struct TimingMarker {
    uint32_t MarkerLowWord;
    uint32_t MarkerHighWord;
  };
     namespace timmarker {
        enum TimingMarkerMASKS{
            MASK_Lo = 0xFFFFFFFF,
            MASK_Hi = 0x00FFFFFF
        };

        enum TimingMarkerSHIFTS{
            SHIFT_Lo = 0,
            SHIFT_Hi = 0
        };

        enum TimingMarkerWord{
            WORD_Lo = 0,
            WORD_Hi = 1
        };


    };// end of namespace timmarker


  //! Class to hold the timing markers used in the construction of Microslices in the DCMs
  /*!
    The RawTimingMarker class impliments the data formats used to create proper timing marker
    which is used in the construction of a data microslice in the Data concentrator modules 
    data collection process.  These markers are designed to separate blocks of data coming 
    out of the front end boards at roughly 5us intervals.  All the data between time markers
    is assume to have occured after the first marker and before the second marker.

    Each marker is 56 bits (0-55) long divided into a high and low word.

    This class provides a set of generator methods which will produce a valid time marker
    based off of the current system clock of the host system on which it is running.  This marker
    is generated using the CPU specific real time clock system and may not be available 
    on all platforms.
   */
  class RawTimingMarker: public RawDAQData {
    
  public:
    
    // Constructors
    RawTimingMarker();		               //! The default constructor
    RawTimingMarker(const RawTimingMarker& copy_in); //! Deep copy constructor
    RawTimingMarker(void* timingmarker_buffer);//! Shallow copy constructor

    /// Destructor
    ~RawTimingMarker();

    /// Initialize the Timing Marker
    void init();
    /// Deep Copy
    void  copy(const RawTimingMarker& copy_in);
    
    uint32_t  getLowWord() const;  //! Get method for lower half of timing word
    uint32_t  getHighWord() const; //! Get method for upper half of timing word
    uint64_t  getTime() const;     //! Get the time-stamp as a single 64-bit value

    bool setLowWord (uint32_t); //! Set method for lower half of timing word
    bool setHighWord(uint32_t); //! Set method for upper half of timing word
    bool setTime    (uint64_t); //! Set the time-stamp as a single 64-bit value

    bool setBufferSource (void* pointer); //!< Set The Buffer pointer to a given location. Don't use it if you are not sure
    bool resetBufferSource();             //!< Reset the class to use its internal buffer
    bool clear();                         //!< Clear the buffer
    bool reset();                         //!< Reset the class data
    bool isInternalBuffer() const;              //!< Is the buffer internal?


    /// Print the Timing marker
    void print() const;

    /// Method to generate a valid timing marker from the system clocks
    void generate();

    
    /// Sizeof method that returns the size of the packed data members in words
    uint32_t sizeofdata() const;   // Returns the size of the data block in words

  private:

      /// The actual internal buffer (not a pointer)
      uint32_t  _InternalBuffer[RAWTIMINGMARKER_SIZE];
  };
  
  
} // END NAMESPCE daqdataformats
#endif /* RAWTIMINGMARKER_H */
