#ifndef DAQRAWTRIGGERHEADER_H
#define DAQRAWTRIGGERHEADER_H

#include <unistd.h>

#include "DAQDataFormats/RawDAQData.h"
#include "DAQDataFormats/TriggerDefines.h"

#define COPY_COPIES_INTERNAL	/* Define the behavior of copy constructors */

namespace daqdataformats {

  const static uint32_t TriggerHeaderDelimiter = 0xAAAA;

  enum TriggerHeaderMASKS{
    MASK_DELIMITER          = 0xFFFF0000,
    MASK_VERSION            = 0x00000FFF,
    MASK_HWFLAG             = 0x80000000,
    MASK_SOURCE             = 0x78000000,
    MASK_SOURCE_SUBID       = 0x00FF0000,
    MASK_SOURCE_ID          = 0x0000FFFF,
    MASK_TRIG_LOW           = 0xFFFFFFFF,
    MASK_TRIG_HIGH          = 0x000000FF,
    MASK_MASTERTRIGGER_LOW  = 0xFFFFFFFF,
    MASK_MASTERTRIGGER_HIGH = 0x000000FF,

    SHIFT_DELIMITER          = 16,
    SHIFT_VERSION            = 0,
    SHIFT_HWFLAG             = 31,
    SHIFT_SOURCE             = 27,
    SHIFT_SOURCE_SUBID       = 16,
    SHIFT_SOURCE_ID          = 0,
    SHIFT_TRIG_LOW           = 0,
    SHIFT_TRIG_HIGH          = 0,
    SHIFT_MASTERTRIGGER_LOW  = 0,
    SHIFT_MASTERTRIGGER_HIGH = 0,

    WORD_DELIMITER          = 0,
    WORD_VERSION            = 0,
    WORD_HWFLAG             = 1,
    WORD_SOURCE             = 1,
    WORD_SOURCE_SUBID       = 1,
    WORD_SOURCE_ID          = 1,
    WORD_TRIG_LOW           = 2,
    WORD_TRIG_HIGH          = 3,
    WORD_MASTERTRIGGER_LOW  = 4,
    WORD_MASTERTRIGGER_HIGH = 5
  };

  //! Class to hold the data from the global trigger header in the correct format
  /*!
    The RawTriggerHeader contains the basic information required to
    reconstruct where and how the trigger was generated.

    The class supports general initialization with an internal or external
    data buffer as well as copyin style initialization for use with
    STL containers.
  */
  class RawTriggerHeader: public RawDAQData {
  
  public:
      
    RawTriggerHeader();		                            //!< Constructor
    RawTriggerHeader(void* input_buffer, uint32_t size=0);  //!< Constructor with data intialization
    RawTriggerHeader(const RawTriggerHeader& copy_in);	    //!< Copyin Constructor
    ~RawTriggerHeader();                                    //!< Destructor
    
    bool copy(void* input_buffer, uint32_t size=0);       //!< Copy method from an external data block (non-deep)
    bool copy(const RawTriggerHeader& copy_in);	          //!< Copy method from RawTriggerHeader format (deep copy)

    uint32_t   getDelimiter() const;	          //!< Get method for the block delimiter (normally 0xAAAA)
    uint32_t   getHWFlag() const;                         //!< Get method for the hardware generation flag
    uint32_t   getVersion() const;                        //!< Get method for the verion of the trigger header block
    uint32_t   getSource() const;                         //!< Get method for the trigger source
    uint32_t   getSourceID() const;                       //!< Get meethod for the trigger source ID
    uint32_t   getSourceSubID() const;                    //!< Get method for the trigger sub source ID
    uint32_t   getTriggerNumberLow() const;               //!< Get method for the low bits (0-31) of the trig number
    uint32_t   getTriggerNumberHigh() const;              //!< Get method for the high bits (32+) of the trig number
    uint64_t   getTriggerNumber() const;                  //!< Get method for trigger number
    uint32_t   getMasterTriggerNumberLow() const;         //!< Get method for the low bits (0-31) of the master trig number
    uint32_t   getMasterTriggerNumberHigh() const;        //!< Get method for the high bits (32+) of the master trig number
    uint64_t   getMasterTriggerNumber() const;            //!< Get method for master trigger number

    bool setDelimiter(uint32_t delim);	                  //!< Set method for the block delimiter (normally 0xAAAA)
    bool setHWFlag(uint32_t flag);		          //!< Set method for the hardware generation flag
    bool setVersion(uint32_t version);                    //!< Set method for pulseheight information
    bool setSource(trigSource source);	                  //!< Set method for pulseheight information
    bool setSourceID(trigSourceID source);                //!< Set method for pulseheight information
    bool setSourceSubID(trigSourceSubID source);          //!< Set method for pulseheight information
    bool setTriggerNumberLow(uint32_t trig_low);          //!< Set method for the low bits (0-31) of the trig number
    bool setTriggerNumberHigh(uint32_t trig_hi);          //!< Set method for the high bits (32+) of the trig number
    bool setTriggerNumber(uint64_t trig_num);             //!< Set method for pulseheight information
    bool setMasterTriggerNumberLow(uint32_t trig_low);          //!< Set method for the low bits (0-31) of the trig number
    bool setMasterTriggerNumberHigh(uint32_t trig_hi);          //!< Set method for the high bits (32+) of the trig number
    bool setMasterTriggerNumber(uint64_t trig_num);             //!< Set method for pulseheight information

    bool reset();                                         //!< Reset the class data
    bool clear();                                         //!< Clear the buffer
    bool resetBufferSource();	                          //!< Reset the class to use its internal buffer
    bool setBufferSource(void* input_buffer);             //!< Set the buffer to use for storage (can be external)

    bool isInternalBuffer() const;			  //!< Inquiry method for internal buffering type
        
    //! General write method that writes the raw Trigger header formated data block to the target memory
    void* writeData(void* buffer);

    //! General read method that read the raw Trigger header formated data block from the target memory
    void* readData(void* buffer);


    //! Print Trigger Header
    void print() const;
    
    //! General sizeof method which returns the size of the Trigger header formated data block in words (not bytes)
    uint32_t sizeofdata() const;

    //! Assignment Operator
    bool operator=(const RawTriggerHeader& copy_in);
    
  private:

    void init(void* input_buffer=0, uint32_t size=0); //!< Class initialization.  This method will zero associated buffers.
    
    //! The trigger header is a fixed size
    const static uint32_t HEADER_WORDS = 6;

    //! Maximum size of the current buffer (used for external buffers)
    uint32_t _max_buffer_size;

    //! Raw block structure containing the packed information
    uint32_t _header[HEADER_WORDS];

  };
  
  
} // END NAMESPCE daqdataformats
#endif /* DAQRAWTRIGGERHEADER_H */
