#include "DAQDataFormats/RawDataBlock.h"
#include "DAQDataFormats/RawMilliSliceHeader.h"
#include "DAQDataFormats/DataFormatException.h"
#include <sys/time.h>		// gettimeofday
#include <boost/lexical_cast.hpp>
#include <cstdio>


#ifndef RAWBUFF32
// Define a macros to make accessing the raw buffer easier
#define RAWBUFF32 ((uint32_t*)this->_Buffer)
#define RAWBUFF16 ((uint16_t*)this->_Buffer)
#endif

using namespace daqdataformats;

//------------------------------------------------------------------------------
// Default constructor
RawDataBlock::RawDataBlock(uint32_t size, uint32_t tlvl) : _tlvl(tlvl)
{
    tprintf( 24,"RawDataBlock(uint32_t size) construct (this=%p)\n",this );

    if (size < RAWDATABLOCKHEADER_SIZE) {
            std::string msg("An external input buffer was given to the ");
            msg.append("RawDataBlock constructor that was too small (size=");
            msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
            msg.append(boost::lexical_cast<std::string>(RAWDATABLOCKHEADER_SIZE) + ").");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
        }

    // Set the maximum size
    _max_buffer_size = size;

    _InternalBuffer.clear();
    _InternalBuffer.resize(_max_buffer_size, 0);

    // Set pointers to point to the internal buffer
    this->resetBufferSource();

    this->init();
}

//------------------------------------------------------------------------------
RawDataBlock::RawDataBlock(const RawDataBlock& copy_in){
    this->copy(copy_in);
}

//------------------------------------------------------------------------------
RawDataBlock::RawDataBlock(void* datablock_buffer, uint32_t size){
    if (size < RAWDATABLOCKHEADER_SIZE) {
            std::string msg("An external input buffer was given to the ");
            msg.append("RawDataBlock constructor that was too small (size=");
            msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
            msg.append(boost::lexical_cast<std::string>(RAWDATABLOCKHEADER_SIZE) + ").");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
        }

    // First, setup the internal buffer
        // Set the maximum size
        _max_buffer_size = size;

        _InternalBuffer.clear();
        _InternalBuffer.resize(_max_buffer_size, 0);

        // Set pointers to point to the internal buffer
        this->resetBufferSource();


        this->init();

    // Now, setup the pointer to the external buffer
    this->setBufferSource(datablock_buffer);
}

//------------------------------------------------------------------------------
// Destructor
RawDataBlock::~RawDataBlock() { tprintf(24,"~RawDataBlock (this=%p)\n",this);
}

//------------------------------------------------------------------------------
// General initialization
void RawDataBlock::init(){

    _Header.init();
    // Initialize _FloatingMicroBlockNumber
    _FloatingMicroBlockNumber = -1;

    _fd_read_bytes = 0;
}

//------------------------------------------------------------------------------
void RawDataBlock::copy(const RawDataBlock& copy_in){


    _max_buffer_size = copy_in._max_buffer_size;
    _InternalBuffer.reserve((long unsigned int)_max_buffer_size);

      // Copy the data from one buffer to the other
    for(uint32_t i = 0; i < _max_buffer_size; ++i)
        _InternalBuffer[i] = copy_in._InternalBuffer[i];

    // Setup the main buffer pointer
    // If the copy_in's buffer is external, point to that external memory
    if(copy_in.isInternalBuffer()) this->resetBufferSource();
    else                           this->setBufferSource(copy_in._Buffer);

    // Setup the FloatingMicroBlock
    _FloatingMicroBlockNumber = copy_in._FloatingMicroBlockNumber;
    if(_FloatingMicroBlockNumber > -1)this->setFloatingMicroBlock(copy_in._FloatingMicroBlockNumber);


}

//------------------------------------------------------------------------------
// Add a MicroBlock into the DataBlock
void RawDataBlock::addMicroBlock(void* microblock_buffer, bool check_mc_flag){

   // Check if the size of the MicroBlock would exceed the FreeMemory.
   // Check it only if the DataBlock buffer is Internal.
   if(this->isInternalBuffer()){
                uint32_t* pp = ((uint32_t*)microblock_buffer) + RAWMICROBLOCKHEADER_SIZE;
                RawMicroSliceHeader header_check(pp);
                uint32_t slicesize = header_check.getByteCount()/4 + header_check.sizeofdata();
                if (this->getFreeMemory() < slicesize) {
                    std::string msg("Cannot add a MicroBlock. ");
                    msg.append("DataBlock buffer is full (total DataBlock buffer size=");
                    msg.append(boost::lexical_cast<std::string>(_max_buffer_size) + ", current data size=");
                    msg.append(boost::lexical_cast<std::string>(this->sizeofdata()) + ", MicroBlock size = ");
                    msg.append(boost::lexical_cast<std::string>(slicesize) + ")");
                    GENERATE_DATAFORMAT_EXCEPTION(msg);
                }
    }

    // Set the pointer of the _FloatingMicroBlock into the end of the currect Buffer
    _FloatingMicroBlock.setBufferSource(&RAWBUFF32[this->sizeofdata()]);
    // Read the MicroBlock into the _Buffer
    _FloatingMicroBlock.readData(microblock_buffer);

    // Setup the DataBlock header by adding the size and the number of MicroBlocks in the DataBlock
    _Header.addDataSize(_FloatingMicroBlock.sizeofdata());
    _Header.advanceNumMicroBlocks();

    // Check if the MicroSlice in the MicroBlock is MC generated
    // If it is, set the appropriate DataBlock MC flag
    if(!check_mc_flag)return;
    this->setMonteCarloFlag(_FloatingMicroBlock.getMicroSlice()->getMonteCarloFlag());
}

//------------------------------------------------------------------------------
// Add a MicroBlock of a given size into the DataBlock
void RawDataBlock::addMicroBlock(void* microblock_buffer, uint32_t microblock_size, bool check_mc_flag){

    // Check if the size of the MicroBlock would exceed the FreeMemory
    if (this->isInternalBuffer() && this->getFreeMemory() < microblock_size) {
            std::string msg("Cannot add a MicroBlock. ");
            msg.append("DataBlock buffer is full (total DataBlock buffer size=");
            msg.append(boost::lexical_cast<std::string>(_max_buffer_size) + ", current data size=");
            msg.append(boost::lexical_cast<std::string>(this->sizeofdata()) + ", MicroBlock size = ");
            msg.append(boost::lexical_cast<std::string>(microblock_size) + ")");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
    }

    //Read the MicroBlock to the _Buffer starting from the previous position
    this->readData(microblock_buffer, microblock_size , this->sizeofdata());

    // Change the DataBlock header appropriately
    _Header.addDataSize(microblock_size);
    _Header.advanceNumMicroBlocks();

    // Check if the MicroSlice in the MicroBlock is MC generated
    // If it is, set the appropriate DataBlock MC flag
    if(!check_mc_flag)return;
    RawMicroBlock microblock(microblock_buffer);
    this->setMonteCarloFlag(microblock.getMicroSlice()->getMonteCarloFlag());
}
//------------------------------------------------------------------------------
/// Add the MicroBlock to _Buffer by passing the references of
/// the MilliSlice and the MicroSlice. MilliSlice is the one that contains the MicroSlice.
void RawDataBlock::addMicroBlock(void* millislice_buffer, void* microslice_buffer, bool check_mc_flag){

   // Check if the size of the MicroBlock would exceed the FreeMemory.
   // Check it only if the DataBlock buffer is Internal.
   if(this->isInternalBuffer()){
                RawMicroSliceHeader header_check(microslice_buffer);
                uint32_t slicesize = header_check.getByteCount()/4 + header_check.sizeofdata();
                if (this->getFreeMemory() < slicesize) {
                    std::string msg("Cannot add a MicroBlock. ");
                    msg.append("DataBlock buffer is full (total DataBlock buffer size=");
                    msg.append(boost::lexical_cast<std::string>(_max_buffer_size) + ", current data size=");
                    msg.append(boost::lexical_cast<std::string>(this->sizeofdata()) + ", MicroBlock size = ");
                    msg.append(boost::lexical_cast<std::string>(slicesize) + ")");
                    GENERATE_DATAFORMAT_EXCEPTION(msg);
                }
    }

   // Set the pointer of the _FloatingMicroBlock into the end of the currect Buffer
    _FloatingMicroBlock.setBufferSource(&RAWBUFF32[this->sizeofdata()]);
    // Read the MicroBlockHeader and the MicroSlice
    _FloatingMicroBlock.setMicroBlockHeader(millislice_buffer);
    _FloatingMicroBlock.setMicroSlice(microslice_buffer);

    // Setup the DataBlock header by adding the size and the number of MicroBlocks in the DataBlock
    _Header.addDataSize(_FloatingMicroBlock.sizeofdata());
    _Header.advanceNumMicroBlocks();

    // Check if the MicroSlice in the MicroBlock is MC generated
    // If it is, set the appropriate DataBlock MC flag
    if(!check_mc_flag)return;
    this->setMonteCarloFlag(_FloatingMicroBlock.getMicroSlice()->getMonteCarloFlag());
}

//------------------------------------------------------------------------------
/// Add the MicroBlock to _Buffer by passing the references of
/// the MilliSlice and the MicroSlice of a known size.
/// MilliSlice is the one that contains the MicroSlice.
void RawDataBlock::addMicroBlock(void* millislice_buffer, void* microslice_buffer, uint32_t microslice_size, bool check_mc_flag){

    // Calculate the size of the MicroBlock that we are adding
    uint32_t microblock_size = RAWMICROBLOCKHEADER_SIZE + microslice_size;

    // Check if the size of the MicroBlock would exceed the FreeMemory
    if (this->isInternalBuffer() && this->getFreeMemory() < microblock_size) {
            std::string msg("Cannot add a MicroBlock. ");
            msg.append("DataBlock buffer is full (total DataBlock buffer size=");
            msg.append(boost::lexical_cast<std::string>(_max_buffer_size) + ", current data size=");
            msg.append(boost::lexical_cast<std::string>(this->sizeofdata()) + ", MicroBlock size = ");
            msg.append(boost::lexical_cast<std::string>(microblock_size) + ")");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
    }

    //Read the MicroBlockHeader
    uint32_t* pp = ((uint32_t*)millislice_buffer) + rawmillihdr::WORD_DCMChan;
    this->readData(pp               , RAWMICROBLOCKHEADER_SIZE , this->sizeofdata());

    //Read the MicroSlice
    this->readData(microslice_buffer,  microslice_size         , this->sizeofdata() + RAWMICROBLOCKHEADER_SIZE);

    // Change the DataBlock header appropriately
    _Header.addDataSize(microblock_size);
    _Header.advanceNumMicroBlocks();

    // Check if the MicroSlice in the MicroBlock is MC generated
    // If it is, set the appropriate DataBlock MC flag
    if(!check_mc_flag)return;
    RawMicroSliceHeader microslice_header(microslice_buffer);
    this->setMonteCarloFlag(microslice_header.getMonteCarloFlag());
}

// Closing the DataBlock by adding the CRC and increasing the DataBlock size by one (CRC).
bool RawDataBlock::close(bool produce_crc){

    // Add one word to the Size of the DataBlock for the CRC
    _Header.addDataSize(1);

    // Add CRC at the very end of _Buffer
    if(produce_crc){
        _Header.setCRCCalculationUsed(true);
        this->addCheckSum();
    }
    return true;
}


//------------------------------------------------------------------------------
//Getters
      RawDataBlockHeader* RawDataBlock::getHeader()              { return &_Header; }
      RawMicroBlock*      RawDataBlock::getFloatingMicroBlock()  { return &_FloatingMicroBlock; }
      uint32_t            RawDataBlock::getVersion()                 const { return _Header.getVersion();}
      uint32_t            RawDataBlock::getMarker()                  const { return _Header.getMarker();}
      uint32_t            RawDataBlock::getTrigNumLo()               const { return _Header.getTrigNumLo();}
      uint32_t            RawDataBlock::getTrigNumHi()               const { return _Header.getTrigNumHi();}
      uint64_t            RawDataBlock::getTrigNum()                 const { return _Header.getTrigNum();}
      uint16_t            RawDataBlock::getNumMicroBlocks()          const { return _Header.getNumMicroBlocks();}
      uint16_t            RawDataBlock::getBuffId()                  const { return _Header.getBuffId();}
      uint32_t            RawDataBlock::getDataSize()                const { return _Header.getDataSize();}
      bool                RawDataBlock::getMonteCarloFlag()          const { return _Header.getMonteCarloFlag();}
      bool                RawDataBlock::isCRCCalculationUsed()       const { return _Header.isCRCCalculationUsed();}
      bool                RawDataBlock::isConnectionInitialization() const { return _Header.isConnectionInitialization();}
      bool                RawDataBlock::isCloseConnection()          const { return _Header.isCloseConnection();}
      bool                RawDataBlock::isMissingData()              const { return _Header.isMissingData();}


      uint32_t            RawDataBlock::getFreeMemory() const {
            if(this->sizeofdata() > _max_buffer_size)return 0;
            return _max_buffer_size - this->sizeofdata();
      }

      //------------------------------------------------------------------------------
//Setters
      bool RawDataBlock::setVersion                  (uint32_t v)   { return _Header.setVersion(v);}
      bool RawDataBlock::setTrigNumLo                (uint32_t tr)  { return _Header.setTrigNumLo(tr);}
      bool RawDataBlock::setTrigNumHi                (uint32_t tr)  { return _Header.setTrigNumHi(tr);}
      bool RawDataBlock::setTrigNum                  (uint64_t tr)  { return _Header.setTrigNum(tr);}
      bool RawDataBlock::setBuffId                   (uint16_t b)   { return _Header.setBuffId(b);}
      bool RawDataBlock::setMonteCarloFlag           (bool ft)      { return _Header.setMonteCarloFlag(ft);}
      bool RawDataBlock::setConnectionInitialization (bool ft)      { return _Header.setConnectionInitialization(ft);}
      bool RawDataBlock::setCloseConnection          (bool ft)      { return _Header.setCloseConnection(ft);}
      bool RawDataBlock::setMissingData              (bool ft)      { return _Header.setMissingData(ft);}

      bool RawDataBlock::setBufferSource(void* pointer){
        // Set the Buffer and its header pointers
        _Buffer  = pointer;
        _Header.setBufferSource(pointer);

        // reset the Floating MicroBlock number
        _FloatingMicroBlockNumber = -1;
        return true;
      }
    bool RawDataBlock::resetBufferSource(){
        this->setBufferSource(&_InternalBuffer[0]);
        return true;
    }

// clears all data in the buffer
    bool RawDataBlock::clear(){
        uint32_t max_data = this->sizeofdata();

        // If it exceeds, then it means that External buffer was used.
        if(this->isInternalBuffer() && max_data > _max_buffer_size)max_data = _max_buffer_size;

        // Loop over the buffer and zero it
        for(uint32_t i = 0; i < max_data; ++i)
            RAWBUFF32[i]=0;

        _fd_read_bytes = 0;

        return true;
    }
    bool RawDataBlock::reset(){
        this->resetBufferSource();
        this->clear();
        return true;
    }
    bool RawDataBlock::isInternalBuffer() const {

        if((this->_Buffer) == (void*)(&_InternalBuffer[0]))
            return true;
        return false;
    }


      
    bool RawDataBlock::setFloatingMicroBlock (uint32_t imicro){
        if(imicro == 0){
            _FloatingMicroBlockNumber = imicro;
            return _FloatingMicroBlock.setBufferSource(&RAWBUFF32[_Header.sizeofdata()]);
            }

        if((int)imicro == _FloatingMicroBlockNumber)return true;

        if (imicro >= this->getNumMicroBlocks()) return false;
        

        if((int)imicro > _FloatingMicroBlockNumber && _FloatingMicroBlockNumber > -1){
            uint32_t* pointer = (uint32_t*)_FloatingMicroBlock.getBuffer();
            for(; _FloatingMicroBlockNumber < (int)imicro; _FloatingMicroBlockNumber++){
                 pointer += _FloatingMicroBlock.sizeofdata();
                _FloatingMicroBlock.setBufferSource(pointer);
            }
            return true;
        }

        // Firrst point to the first MicroBlock
        uint32_t* pointer = (uint32_t*)&RAWBUFF32[_Header.sizeofdata()];
        _FloatingMicroBlock.setBufferSource(pointer);

        for(_FloatingMicroBlockNumber = 0; _FloatingMicroBlockNumber < (int)imicro; _FloatingMicroBlockNumber++){
            pointer += _FloatingMicroBlock.sizeofdata();
            _FloatingMicroBlock.setBufferSource(pointer);
            }
        return true;
    }
//------------------------------------------------------------------------------
// Read an entire DataBlock from datablock_buffer into _Buffer
void* RawDataBlock::readData(void* datablock_buffer){
    //Read the header and unpack it
    datablock_buffer = _Header.readData(datablock_buffer);

    // read the _Buffer from scratch
    return this->readData(datablock_buffer, this->sizeofdata() - _Header.sizeofdata(), _Header.sizeofdata());
}

//------------------------------------------------------------------------------
void* RawDataBlock::readData(void* void_buffer, uint32_t size){

  uint32_t* buffer = (uint32_t*)void_buffer;

  if(buffer == 0){return 0;}

  for(uint32_t i = 0; i < size; ++i )
    RAWBUFF32[i] = buffer[i];

  // Return the address of the buffer advanced by 1 word;
  return (void*)(buffer + size);
}

// Method to read data in to a file descriptor
// The assumption is that the read will not block because a poll or select
// (in the case that the fd is a socket) has determined that data is ready.
// (Of course, the fd is always "ready" if the fd is associate with a file
// on disk).  But no assumption is made that a 2nd read will not block, so
// only one read (within this readData routine) is every done. The caller
// is responsible for doing the necessary number of readData calls to get a
// "complete" object. 
readDataStatus RawDataBlock::readData(int fd)
{
    if (fd >= 0)
    {
        if (isInternalBuffer())
        {   ssize_t ss;
	    size_t bytes_to_read;
            if (_fd_read_bytes < RAWDATABLOCKHEADER_SIZE*sizeof(uint32_t))
            {   size_t min_words = RAWDATABLOCKHEADER_SIZE;
                size_t min_bytes =min_words*sizeof (uint32_t);
                if (_InternalBuffer.capacity() < min_words)
                {
		    _InternalBuffer.resize(min_words);
                    this->resetBufferSource();
                }

		bytes_to_read = min_bytes - _fd_read_bytes;
            }
            else
            {   uint32_t ss=sizeofdata();
		// check before using size
		if (   (getMarker() != datablock::DataBlock_Marker)
		    || (ss           > 0x400000/*arbitrary val*/) )
		{   return rds_error;
		}
                if (_InternalBuffer.capacity() < (size_t)ss)
                {   _InternalBuffer.resize( ss );
                    this->resetBufferSource();
                }
                bytes_to_read = (size_t)ss*sizeof(uint32_t) - _fd_read_bytes;
            }

	    ss = read(fd, (char*)_Buffer+_fd_read_bytes, bytes_to_read );
	    if (ss==-1)
	    {   perror("read");
		return rds_error;
	    }
	    else if (ss==0)
	    {   tprintf( 5, "connection closed\n" );
		return rds_error;
	    }

	    _fd_read_bytes += ss;
	    if (ss != (int)bytes_to_read)
	    {   return rds_incomplete;
	    }

	    // at this point, the header should have been read in and the
	    // size from within can be examined.
	    // check for reasonableness
	    if (   (getMarker() != datablock::DataBlock_Marker)
		|| (sizeofdata() > 0x400000/*arbitrary val*/) )
	    {   return rds_error;
	    }

	    if (((size_t)this->sizeofdata()*sizeof (uint32_t)) == _fd_read_bytes)
		return rds_complete;
	    return rds_incomplete;
        }
        else
        {
	    printf("external buffer\n");
            // assume there is enough space???
	    return RawDAQData::readData(fd);
        }
    }
    return rds_error;
}

//------------------------------------------------------------------------------
void* RawDataBlock::readData(void* buffer_void, uint32_t size, uint32_t _Buffer_start_position){

  uint32_t* buffer = (uint32_t*)buffer_void;

  if(buffer == 0){return 0;}

  for(uint32_t i = 0; i < size; ++i )
    RAWBUFF32[i+_Buffer_start_position] = buffer[i];

  // Return the address of the buffer advanced by 1 word;
  return (void*)(buffer+size);
}


//------------------------------------------------------------------------------
// Print the DataBlock structure
void RawDataBlock::print() {


    printf("\n---RawDataBlock----------------------------------------------------size=%u\n",this->sizeofdata());

    _Header.print();

    /// Print all MicroBlocks
    for(uint32_t imicroblock = 0; imicroblock < this->getNumMicroBlocks(); ++imicroblock){
        this->setFloatingMicroBlock(imicroblock);
        this->getFloatingMicroBlock()->print();
    }



    /*
    std::cout<<"\nRAW:";
    for(uint32_t i = 0; i < this->sizeofdata(); ++i){
        printf(" ");
        this->printWord(i,false);
    }

    std::cout<<"\nRAW:";
    for(uint32_t i = 0; i < this->sizeofdata(); ++i){
            if(i > 0)std::cout<<"    ";
            this->printWord(i,true);
            std::cout<<"\n";
            }
*/


}

//------------------------------------------------------------------------------
// Size of the _Buffer in words
uint32_t RawDataBlock::sizeofdata()const { return _Header.getDataSize();}

void
RawDataBlock::tprintf( unsigned lvl, const char *fmt, ... )
{   if (lvl > 31) lvl=31;
    if (_tlvl & (1<<lvl))
    {   timeval 	tv;
	char            str[80];
	time_t          local;
	const char*     indent="                                ";//32 spaces
	va_list         ap;
	va_start( ap, fmt);
	gettimeofday( &tv, NULL );
	local = tv.tv_sec;
	strftime( str, sizeof(str), "%T", localtime(&local) );
	printf( "%s.%06ld%s", str, (long)tv.tv_usec, &indent[31-lvl] );
	vprintf( fmt, ap );
    }
}   // tprintf
