#include "DAQDataFormats/RawMicroBlock.h"
#include "DAQDataFormats/RawMilliSliceHeader.h"
#include "DAQDataFormats/DataFormatException.h"
#include <boost/lexical_cast.hpp>
#include <cstdio>

#ifndef RAWBUFF32
// Define a macros to make accessing the raw buffer easier
#define RAWBUFF32 ((uint32_t*)this->_Buffer)
#define RAWBUFF16 ((uint16_t*)this->_Buffer)
#endif

using namespace daqdataformats;

//------------------------------------------------------------------------------
// Default constructor
RawMicroBlock::RawMicroBlock(uint32_t size){

    if (size < RAWMICROBLOCKHEADER_SIZE) {
            std::string msg("An external input buffer was given to the ");
            msg.append("RawMicroBlock constructor that was too small (size=");
            msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
            msg.append(boost::lexical_cast<std::string>(RAWMICROBLOCKHEADER_SIZE) + ").");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
        }

    // Set the maximum size
    _max_buffer_size = size;

    // Allocate the internal Buffer
    _InternalBuffer.assign(_max_buffer_size,0);

    // Set pointers to point to the internal buffer
    this->resetBufferSource();

    this->init();
}
//------------------------------------------------------------------------------
RawMicroBlock::RawMicroBlock(const RawMicroBlock& copy_in){
   this->copy(copy_in);
}

//------------------------------------------------------------------------------
RawMicroBlock::RawMicroBlock(void* microblock_buffer, uint32_t size){

    if (size < RAWMICROBLOCKHEADER_SIZE) {
            std::string msg("An external input buffer was given to the ");
            msg.append("RawMicroBlock constructor that was too small (size=");
            msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
            msg.append(boost::lexical_cast<std::string>(RAWMICROBLOCKHEADER_SIZE) + ").");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
        }
    // Set the maximum size
    _max_buffer_size = size;

    // Allocate the internal Buffer
    _InternalBuffer.reserve(_max_buffer_size);
    // Zero it out
    for(uint32_t i = 0; i < _max_buffer_size; ++i)
        _InternalBuffer[i] = 0;


    // First, initialize the Internal Buffer
            // Set pointers to point to the internal buffer
            this->resetBufferSource();

            // Initialize it
            this->init();

    // Now set the pointer to the external buffer
    this->setBufferSource(microblock_buffer);

}
//------------------------------------------------------------------------------
// Destructor
RawMicroBlock::~RawMicroBlock() { }

//------------------------------------------------------------------------------
// Initialization of Data members
void RawMicroBlock::init(){

    // Initialize the header
    _Header.init();

    //Don't need to initialize the MicroSlice. We'll just copy it straight to the buffer
    
}

//------------------------------------------------------------------------------
void RawMicroBlock::copy(const RawMicroBlock& copy_in){

        // Set the maximum size
    _max_buffer_size = copy_in._max_buffer_size;

    // Setup the internal buffer
    _InternalBuffer.reserve(_max_buffer_size);

      // Copy the data from one buffer to the other
    for(uint32_t i = 0; i < _max_buffer_size; ++i)
        _InternalBuffer[i] = copy_in._InternalBuffer[i];

    // Setup the main buffer pointer
    // If the copy_in's buffer is external, point to that external memory
    if(copy_in.isInternalBuffer()) this->resetBufferSource();
    else                           this->setBufferSource(copy_in._Buffer);

}

//------------------------------------------------------------------------------
// This is a general method to read the microslice from microblock_buffer to _Buffer
void* RawMicroBlock::readData(void* microblock_buffer){
    //Read the header and the MicroSlice
    microblock_buffer = _Header.readData(microblock_buffer);
    return _MicroSlice.readData(microblock_buffer);
}

//------------------------------------------------------------------------------
void* RawMicroBlock::readData(void* void_buffer, uint32_t size){

  uint32_t* buffer = (uint32_t*)void_buffer;

  if(buffer == 0){return 0;}

  for(uint32_t i = 0; i < size; ++i )
    RAWBUFF32[i] = buffer[i];

  // Return the address of the buffer advanced by 1 word;
  return (void*)(buffer + size);
}

//------------------------------------------------------------------------------
void* RawMicroBlock::readData(void* buffer_void, uint32_t size, uint32_t _Buffer_start_position){

  uint32_t* buffer = (uint32_t*)buffer_void;

  if(buffer == 0){return 0;}

  for(uint32_t i = 0; i < size; ++i )
    RAWBUFF32[i+_Buffer_start_position] = buffer[i];

  // Return the address of the buffer advanced by 1 word;
  return (void*)(buffer+size);
}

//------------------------------------------------------------------------------
//Getters
    RawMicroBlockHeader* RawMicroBlock::getHeader()      { return &_Header; }
    RawMicroSlice*       RawMicroBlock::getMicroSlice()  { return &_MicroSlice; }


//------------------------------------------------------------------------------
//Setters
    bool RawMicroBlock::setMicroBlockHeader(void* millislice_buffer){
        
        // Setup the correct pointer to the word of the MilliSlice Header
        // which contains the DCMChan
        uint32_t* pp = ((uint32_t*) millislice_buffer ) + rawmillihdr::WORD_DCMChan;

        // Now read that word. This is the MicroBlock Header
        _Header.readData(pp);

        return true;
    }

    bool RawMicroBlock::setMicroSlice(void* microslice_buffer){
 
        // Just read it straight to the buffer
        _MicroSlice.readData(microslice_buffer);

        return true;
    }


    bool RawMicroBlock::setBufferSource(void* pointer){
        _Buffer = pointer;
        _Header.setBufferSource(pointer);

        uint32_t* pp = (uint32_t*)pointer + _Header.sizeofdata();
        _MicroSlice.setBufferSource(pp);

        return true;
    }
    bool RawMicroBlock::resetBufferSource(){
        this->setBufferSource(&_InternalBuffer[0]);
        return true;
    }

// clears all data in the buffer
    bool RawMicroBlock::clear(){
        uint32_t max_data = this->sizeofdata();

        // If it exceeds, then it means that External buffer was used.
        if(this->isInternalBuffer() && max_data > _max_buffer_size)
            max_data = _max_buffer_size;

        // Loop over the buffer and zero it
        for(uint32_t i = 0; i< max_data; ++i)
            RAWBUFF32[i] = 0;

        return true;
    }
    bool RawMicroBlock::reset(){
        this->resetBufferSource();
        this->clear();
        return true;
    }
    bool RawMicroBlock::isInternalBuffer() const {

        if((this->_Buffer) == (void*)(&_InternalBuffer[0]))
            return true;
        return false;
  }


//------------------------------------------------------------------------------
// Printing the MicroBlock structure
void RawMicroBlock::print() {
    printf("\n---MicroBlock---------------------------------------------size = %u words\n",this->sizeofdata());
    _Header.print();
    _MicroSlice.print();

    /*
    std::cout<<"\nMicroBlock Data RAW 32bit WORDs (converted, if required, to native endian format):";
    for(uint32_t i = 0; i < this->sizeofdata(); ++i){
        this->printWord(i,false);
        printf(" ");
    }

    std::cout<<"\nMicroBlock Data RAW 32bit Bits (converted, if required, to native endian format):";
    for(uint32_t i = 0; i < this->sizeofdata(); ++i){
            std::cout<<"\n    ";
            this->printWord(i,true);
            }
*/
}

//------------------------------------------------------------------------------
// Size of the MicroBlock _Buffer in words
uint32_t RawMicroBlock::sizeofdata()const { return _Header.sizeofdata() + _MicroSlice.sizeofdata(); }
