#include "DAQDataFormats/RawMilliSlice.h"
#include "DAQDataFormats/DataFormatException.h"
#include <sys/time.h>
#include <boost/lexical_cast.hpp>
#include <cstdio>

#ifndef RAWBUFF32
// Define a macros to make accessing the raw buffer easier
#define RAWBUFF32 ((uint32_t*)this->_Buffer)
#define RAWBUFF16 ((uint16_t*)this->_Buffer)
#endif

using namespace daqdataformats;

//------------------------------------------------------------------------------
// Default constructor
RawMilliSlice::RawMilliSlice(uint32_t size, uint32_t tlvl) : _tlvl(tlvl)
{
    tprintf( 26,"RawMilliSlice(uint32_t size) construct (this=%p)\n",this );

    if (size < MILLISLICEHEADER_SIZE) {
            std::string msg("An external input buffer was given to the ");
            msg.append("RawMilliSlice constructor that was too small (size=");
            msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
            msg.append(boost::lexical_cast<std::string>(MILLISLICEHEADER_SIZE) + ").");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
        }

    // Set the maximum size
    _max_buffer_size = size;

    // Allocate the internal Buffer
    _InternalBuffer.assign(_max_buffer_size,0);
    //printf("RawMilliSlice default construct size=%d\n", size );

    // Set pointers to point to the internal buffer
    this->resetBufferSource();

    this->init();
}
//------------------------------------------------------------------------------
RawMilliSlice::RawMilliSlice(const RawMilliSlice& copy_in)
# if 1
    : RawDAQData()
    , _InternalBuffer(copy_in._InternalBuffer)
    , _Header(_Buffer)
    , _Index(copy_in._Index)
    , _FloatingMicroSlice(copy_in._FloatingMicroSlice)
    , _FloatingMicroSliceNumber(copy_in._FloatingMicroSliceNumber)
    , _max_buffer_size(copy_in._max_buffer_size)
    , _tlvl(copy_in._tlvl)
{
    if(copy_in.isInternalBuffer()) this->resetBufferSource();
    else                           this->setBufferSource(copy_in._Buffer);    
# else
    {
        // Set the maximum size
    _max_buffer_size = copy_in._max_buffer_size;

    // Setup the internal buffer
    _InternalBuffer.reserve(_max_buffer_size);

      // Copy the data from one buffer to the other
    for(uint32_t i = 0; i < _max_buffer_size; ++i)
        _InternalBuffer[i] = copy_in._InternalBuffer[i];

    // Setup the main buffer pointer
    // If the copy_in's buffer is external, point to that external memory
    if(copy_in.isInternalBuffer()) this->resetBufferSource();
    else                           this->setBufferSource(copy_in._Buffer);
    
    // Setup the FloatingMicroSlice
    _FloatingMicroSliceNumber = copy_in._FloatingMicroSliceNumber;
    if(_FloatingMicroSliceNumber > -1)this->setFloatingMicroSlice(copy_in._FloatingMicroSliceNumber);

    // Finally, copy the MilliSlice Index
    _Index.copy(copy_in._Index);
# endif
}

//------------------------------------------------------------------------------
RawMilliSlice::RawMilliSlice(void* millislice_buffer, uint32_t size)
    : _tlvl(0)
{

    if (size < MILLISLICEHEADER_SIZE) {
            std::string msg("An external input buffer was given to the ");
            msg.append("RawMilliSlice constructor that was too small (size=");
            msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
            msg.append(boost::lexical_cast<std::string>(MILLISLICEHEADER_SIZE) + ").");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
    }
    // Set the maximum size
    _max_buffer_size = size;

    // Allocate the internal Buffer
    _InternalBuffer.reserve(_max_buffer_size);
    // Zero it out
    for(uint32_t i = 0; i < _max_buffer_size; ++i)
        _InternalBuffer[i] = 0;

    
    // First, initialize the Internal Buffer
    // Set pointers to point to the internal buffer
    this->resetBufferSource();

    // Initialize it
    this->init();

    // Now set the pointer to the external buffer
    this->setBufferSource(millislice_buffer);

}
//------------------------------------------------------------------------------
// Destructor
 RawMilliSlice::~RawMilliSlice() { tprintf(26,"~RawMilliSlice (this=%p) tlvl=0x%x\n",this,_tlvl);
}

//------------------------------------------------------------------------------
void RawMilliSlice::copy(const RawMilliSlice& copy_in){

        // Set the maximum size
    _max_buffer_size = copy_in._max_buffer_size;

    // Setup the internal buffer
    _InternalBuffer.reserve(_max_buffer_size);

      // Copy the data from one buffer to the other
    for(uint32_t i = 0; i < _max_buffer_size; ++i)
        _InternalBuffer[i] = copy_in._InternalBuffer[i];

    // Setup the main buffer pointer
    // If the copy_in's buffer is external, point to that external memory
    if(copy_in.isInternalBuffer()) this->resetBufferSource();
    else                           this->setBufferSource(copy_in._Buffer);
    
    // Setup the FloatingMicroSlice
    _FloatingMicroSliceNumber = copy_in._FloatingMicroSliceNumber;
    if(_FloatingMicroSliceNumber > -1)this->setFloatingMicroSlice(copy_in._FloatingMicroSliceNumber);

    // Finally, copy the MilliSlice Index
    _Index.copy(copy_in._Index);
}


//------------------------------------------------------------------------------
// Initialization of Data members
void RawMilliSlice::init(){

    // Initialize the header
    _Header.init();

    // Initialize the index
    _Index.init();

    // Initialize the _FloatingMicroSliceNumber
    _FloatingMicroSliceNumber = -1;

    _fd_read_bytes = 0;
}

//------------------------------------------------------------------------------
// Add a MicroSlice into the MilliSlice changing the Header appropriately
void RawMilliSlice::addMicroSlice(void* microslice_buffer, bool indexate, bool check_mc_flag){

// Check if the size of the MicroSlice would exceed the FreeMemory. Check only if the MilliSlice buffer is Internal.
   if(this->isInternalBuffer()){
                RawMicroSliceHeader header_check(microslice_buffer);
                uint32_t slicesize = header_check.getByteCount()/4 + header_check.sizeofdata();
                uint32_t free = this->getFreeMemory();
                if (free < slicesize) {
                    std::string msg("Cannot add a MicroSlice. ");
                    msg.append("MilliSlice buffer is full (total MilliSlice buffer size=");
                    msg.append(boost::lexical_cast<std::string>(_max_buffer_size) + ", current data size=");
                    msg.append(boost::lexical_cast<std::string>(this->sizeofdata()) + ", MicroSlice size = ");
                    msg.append(boost::lexical_cast<std::string>(slicesize) + ")");
                    GENERATE_DATAFORMAT_EXCEPTION(msg);
                }
    }

    // Set the floating microslice in the end of the current Buffer
    _FloatingMicroSlice.setBufferSource(&RAWBUFF32[this->sizeofdata()]);
    // Read the MicroSlice
    _FloatingMicroSlice.readData(microslice_buffer);

    // Add an index if told so.
    if(indexate){
        // Get Time length and offset
        uint32_t time = (uint32_t) (_FloatingMicroSlice.getTime() - this->getTimeStart());
        uint32_t offset = this->sizeofdata() - _Header.sizeofdata();
        // Add the index
        _Index.addIndex(time, offset);
        // Update the number of indices in the MilliSlice header
        _Header.setMilliSliceIndexNumber(_Index.getIndexNumber());
    }

    // Setup the MilliSliceHeader Header
    // Advance the MicroSlice number in the MilliSlice
    _Header.advanceMicrosliceNumber();
    // Change the MilliSlice size

    _Header.addSliceSize(_FloatingMicroSlice.sizeofdata());

    // Check if the MicroSlice is MC generated
    // If it is, set the appropriate MilliSlice MC flag
    if(!check_mc_flag)return;
    this->setMonteCarloFlag(_FloatingMicroSlice.getMonteCarloFlag());
}

//------------------------------------------------------------------------------
// Add a chunk of MicroSlices into the MilliSlice changing the Header appropriately
void RawMilliSlice::addData(void* buffer, uint32_t size, uint32_t NumberofMicroslices, 
                            bool check_mc_flag, bool check_on_space){

// Check if the size of the MicroSlice would exceed the FreeMemory

  if ( check_on_space ) {
    if (this->isInternalBuffer() && this->getFreeMemory() < size) {
            std::string msg("Cannot add a MicroSlice. ");
            msg.append("MilliSlice buffer is full (total MilliSlice buffer size=");
            msg.append(boost::lexical_cast<std::string>(_max_buffer_size) + ", current data size=");
            msg.append(boost::lexical_cast<std::string>(this->sizeofdata()) + ", MicroSlice size = ");
            msg.append(boost::lexical_cast<std::string>(size) + ")");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
    }
  }

    // Just read the chunk of Microslices straight from the buffer into the end of the (internal) Buffer
    this->readData(buffer, size, this->sizeofdata());

    // Advance the MilliSlice frame number (MicroSlice number in the MilliSlice)
    _Header.setMicrosliceNumber(_Header.getMicrosliceNumber() + NumberofMicroslices);
    // Change the MilliSlice size
    _Header.addSliceSize(size);

    // Check if the first MicroSlice is MC generated
    // If it is, set the appropriate MilliSlice MC flag
    if(!check_mc_flag)return;
    RawMicroSliceHeader micro_header(buffer);
    this->setMonteCarloFlag(micro_header.getMonteCarloFlag());
}

// Closing the MilliSlice by adding the Index (if available), tail and the CRC
bool RawMilliSlice::close(bool produce_crc){

    // If CRC is calculated, set the isCRCCalculationUsed flag
    if(produce_crc)_Header.setCRCCalculationUsed(true);

    // First add the index if available
    if(_Index.sizeofdata() > 0){
        // Just copy the index to the end
        this->readData(_Index.getBuffer(), _Index.sizeofdata(), this->sizeofdata());
        // Adjust the MilliSlice size
        _Header.addSliceSize(_Index.sizeofdata());
    }

    void* current_buffer = &RAWBUFF32[this->sizeofdata()]; //Remember the current Buffer position

    // Add MILLISLICEHEADER_SIZE + 1 words to the buffer as we
    // will add the tail (another MilliSlice header) and the CRC-word in the end of the MilliSlice
    _Header.addSliceSize(MILLISLICEHEADER_SIZE + 1);

    // Adding a tail
        // Create a new Tail and assigning its pointer

       RawMilliSliceHeader tail(current_buffer);
        // Copy the content of the _Header to the tail
        tail.readData(_Header.getBuffer());
        // Set an appropriate Marker for the tail
        tail.setMarker(rawmillihdr::Marker_value[1]);
        // tail is ready

    // Finally, add CRC at the very end of _Buffer
    if(produce_crc)this->addCheckSum();
    return true;
}   // close

//------------------------------------------------------------------------------
// This is a general method to read the microslice from millislice_buffer to _Buffer
void* RawMilliSlice::readData(void* millislice_buffer)
{   //Read the header and unpack it to find out the size of the millislice
    millislice_buffer = _Header.readData(millislice_buffer);

    //Now, read the entire millislice_buffer to the _Buffer
    return this->readData(millislice_buffer, this->sizeofdata() - _Header.sizeofdata(), _Header.sizeofdata());
 }   // readData( void* )

//------------------------------------------------------------------------------
void* RawMilliSlice::readData(void* void_buffer, uint32_t size)
{
  uint32_t* buffer = (uint32_t*)void_buffer;

  if(buffer == 0){return 0;}

  for(uint32_t i = 0; i < size; ++i )
    RAWBUFF32[i] = buffer[i];

  // Return the address of the buffer advanced by 1 word;
  return (void*)(buffer + size);
 }   // readData( void*, uint32_t )

//------------------------------------------------------------------------------
void* RawMilliSlice::readData(void* buffer_void, uint32_t size, uint32_t _Buffer_start_position)
{
    uint32_t* buffer = (uint32_t*)buffer_void;

    if (buffer == 0) {return 0;}

    if (this->isInternalBuffer() && (_InternalBuffer.capacity() < (size+_Buffer_start_position)))
    {   _InternalBuffer.resize( size+_Buffer_start_position );
	_max_buffer_size = size+_Buffer_start_position;
	this->resetBufferSource();
    }

    for(uint32_t i = 0; i < size; ++i )
	RAWBUFF32[i+_Buffer_start_position] = buffer[i];

    // Return the address of the buffer advanced by 1 word;
    return (void*)(buffer+size);
}   // readData(void* buffer_void, uint32_t size, uint32_t _Buffer_start_position)

// Method to read data in to a file descriptor
// The assumption is that the read will not block because a poll or select
// (in the case that the fd is a socket) has determined that data is ready.
// (Of course, the fd is always "ready" if the fd is associate with a file
// on disk).  But no assumption is made that a 2nd read will not block, so
// only one read (within this readData routine) is every done. The caller
// is responsible for doing the necessary number of readData calls to get a
// "complete" millislice. 
readDataStatus RawMilliSlice::readData(int fd)
{
    if (fd >= 0)
    {
        if (isInternalBuffer())
        {   ssize_t ss;
	    size_t bytes_to_read;
            if (_fd_read_bytes < MILLISLICEHEADER_SIZE*sizeof(uint32_t))
            {   size_t min_words = MILLISLICEHEADER_SIZE;
                size_t min_bytes =min_words*sizeof (uint32_t);
                if (_InternalBuffer.capacity() < min_words)
                {
		    _InternalBuffer.resize(min_words);
		    _max_buffer_size = min_words;
                    this->resetBufferSource();
                }

		bytes_to_read = min_bytes - _fd_read_bytes;
            }
            else
            {   uint32_t ss=sizeofdata();
		// check before using size
		if (   (getMarker() != rawmillihdr::Marker_value[0])
		    || (ss           > 0x400000/*arbitrary val*/) )
		{   tprintf(30,"no marker or size too big\n");
		    return rds_error;
		}
                if (_InternalBuffer.capacity() < (size_t)ss)
                {   _InternalBuffer.resize( ss );
		    _max_buffer_size = ss;
                    this->resetBufferSource();
                }
                bytes_to_read = (size_t)ss*sizeof(uint32_t) - _fd_read_bytes;
            }

	    ss = read(fd, (char*)_Buffer+_fd_read_bytes, bytes_to_read );
	    if (ss==-1 || ss==0)
	    {   tprintf(30,"read returned %ld\n", ss);
		return rds_error;
	    }
	    _fd_read_bytes += ss;
	    if (ss != (int)bytes_to_read)
	    {   return rds_incomplete;
	    }

	    // at this point, the header should have been read in and the
	    // size from within can be examined.
	    // check for resonableness
	    if (   (getMarker() != rawmillihdr::Marker_value[0])
		|| (sizeofdata() > 0x400000/*arbitrary val*/) )
	    {   tprintf(30,"no marker or size too big (2)\n");
		return rds_error;
	    }

	    if (((size_t)this->sizeofdata()*sizeof (uint32_t)) == _fd_read_bytes)
		return rds_complete;
	    return rds_incomplete;
        }
        else
        {
	    printf("external buffer\n");
            // assume there is enough space???
	    return RawDAQData::readData(fd);
        }
    }
    return rds_error;
}   // readData( int )

//------------------------------------------------------------------------------
//Getters
RawMilliSliceHeader* RawMilliSlice::getHeader()                 { return &_Header; }
RawMilliSliceIndex*  RawMilliSlice::getMilliSliceIndex()        { return &_Index;}
RawMicroSlice*       RawMilliSlice::getFloatingMicroSlice()     { return &_FloatingMicroSlice; }
uint32_t             RawMilliSlice::getVersion()                 const { return _Header.getVersion();}
uint32_t             RawMilliSlice::getMarker()                  const { return _Header.getMarker();}
uint32_t             RawMilliSlice::getDestBuffID()              const { return _Header.getDestBuffID();}
uint32_t             RawMilliSlice::getPartition()               const { return _Header.getPartition();}
uint64_t             RawMilliSlice::getTimeStart()               const { return _Header.getTimeStart();}
uint32_t             RawMilliSlice::getTimeStartLow()            const { return _Header.getTimeStartLow();}
uint32_t             RawMilliSlice::getTimeStartHigh()           const { return _Header.getTimeStartHigh();}
uint32_t             RawMilliSlice::getTimeLength()              const { return _Header.getTimeLength();}
bool                 RawMilliSlice::isComplete()                 const { return _Header.isComplete();}
rawmillihdr::Mode_t  RawMilliSlice::getMode()                    const { return _Header.getMode();}
bool                 RawMilliSlice::isConnectionInitialization() const { return _Header.isConnectionInitialization();}
bool                 RawMilliSlice::isCloseConnection()          const { return _Header.isCloseConnection();}
uint32_t             RawMilliSlice::getMilliSliceSEQ()           const { return _Header.getMilliSliceSEQ();}
uint32_t             RawMilliSlice::getFrameSEQ()                const { return _Header.getFrameSEQ();}
uint32_t             RawMilliSlice::getSliceSize()               const { return _Header.getSliceSize();}
uint32_t             RawMilliSlice::getMicrosliceNumber()        const { return _Header.getMicrosliceNumber();}
uint32_t             RawMilliSlice::getDiblockId()               const { return _Header.getDiblockId();}
uint32_t             RawMilliSlice::getDCMId()                   const { return _Header.getDCMId();}
uint32_t             RawMilliSlice::getDetId()                   const { return _Header.getDetId();}
uint32_t             RawMilliSlice::getGlobalDCMID()             const { return _Header.getGlobalDCMID();}
uint32_t             RawMilliSlice::getMilliSliceIndexNumber()   const { return _Header.getMilliSliceIndexNumber();}
uint32_t             RawMilliSlice::getMicroSliceTimeLength_usec() const { return _Header.getMicroSliceTimeLength_usec();}
bool                 RawMilliSlice::getMonteCarloFlag()          const { return _Header.getMonteCarloFlag();}
bool                 RawMilliSlice::isCRCCalculationUsed()       const { return _Header.isCRCCalculationUsed();}
bool                 RawMilliSlice::isMilliSliceIndexUsed()      const { return _Header.isMilliSliceIndexUsed();}
uint32_t             RawMilliSlice::getFreeMemory() const {
        // Allocated size minus (current buffer size + Tail size + CRC + Index size + 1)
        if(this->sizeofdata() + _Header.sizeofdata() + _Index.sizeofdata() + 1 > _max_buffer_size) return 0;
        return _max_buffer_size - this->sizeofdata() - _Header.sizeofdata() - _Index.sizeofdata() - 1;
}

//------------------------------------------------------------------------------
//Setters
bool RawMilliSlice::setDiblockId                (uint32_t id)   { return _Header.setDiblockId(id);}
bool RawMilliSlice::setDCMId                    (uint32_t id)   { return _Header.setDCMId(id);}
bool RawMilliSlice::setDetId                    (uint32_t id)   { return _Header.setDetId(id);}
bool RawMilliSlice::setGlobalDCMID              (uint32_t id)   { return _Header.setGlobalDCMID(id);}
bool RawMilliSlice::setVersion                  (uint32_t v)    { return _Header.setVersion(v);}
bool RawMilliSlice::setDestBuffID               (uint32_t b)    { return _Header.setDestBuffID(b);}
bool RawMilliSlice::setPartition                (uint32_t p)    { return _Header.setPartition(p);}
bool RawMilliSlice::setTimeStart                (uint64_t ts)   { return _Header.setTimeStart(ts);}
bool RawMilliSlice::setTimeStartLow             (uint32_t ts)   { return _Header.setTimeStartLow(ts);}
bool RawMilliSlice::setTimeStartHigh            (uint32_t ts)   { return _Header.setTimeStartHigh(ts);}
bool RawMilliSlice::setTimeLength               (uint32_t tl)   { return _Header.setTimeLength(tl);}
bool RawMilliSlice::setComplete                 (bool ft)       { return _Header.setComplete(ft);}
bool RawMilliSlice::setMode(rawmillihdr::Mode_t mode)           { return _Header.setMode(mode);}
bool RawMilliSlice::setConnectionInitialization (bool ft)       { return _Header.setConnectionInitialization(ft);}
bool RawMilliSlice::setCloseConnection          (bool ft)       { return _Header.setCloseConnection(ft);}
bool RawMilliSlice::setMilliSliceSEQ            (uint32_t seq)  { return _Header.setMilliSliceSEQ(seq);}
bool RawMilliSlice::setMicroSliceTimeLength_usec(uint32_t usec) { return _Header.setMicroSliceTimeLength_usec(usec);}
bool RawMilliSlice::setMonteCarloFlag           (bool mc_flag)  { return _Header.setMonteCarloFlag (mc_flag);}
bool RawMilliSlice::setMilliSliceIndexUsed      (bool is_used)  { return _Header.setMilliSliceIndexUsed (is_used);}
bool RawMilliSlice::setFrameSEQ                 (uint32_t fseq) { return _Header.setFrameSEQ(fseq);}

bool RawMilliSlice::setBufferSource(void* pointer){
        _Buffer = pointer;

        _Header.setBufferSource(pointer);
         
        // reset the Floating MicroSlice number
        _FloatingMicroSliceNumber = -1;
        return true;
}
bool RawMilliSlice::resetBufferSource(){
        this->setBufferSource(&_InternalBuffer[0]);
        return true;
}

// clears all data in the buffer
bool RawMilliSlice::clear(){
        uint32_t max_data = this->sizeofdata();

        // If it exceeds, then it means that External buffer was used.
        if(this->isInternalBuffer() && max_data > _max_buffer_size)
            max_data = _max_buffer_size;

        // Loop over the buffer and zero it
        for(uint32_t i = 0; i < max_data; ++i)
            RAWBUFF32[i] = 0;

        _fd_read_bytes = 0;

        return true;
}
bool RawMilliSlice::reset(){
        this->resetBufferSource();
        this->clear();
        return true;
}
bool RawMilliSlice::isInternalBuffer() const {

        if((this->_Buffer) == (void*)(&_InternalBuffer[0]))
            return true;
        return false;
}


//------------------------------------------------------------------------------
bool RawMilliSlice::setFloatingMicroSlice(uint32_t imicro){
        if(imicro == 0){
            _FloatingMicroSliceNumber = imicro;
            return _FloatingMicroSlice.setBufferSource(&RAWBUFF32[_Header.sizeofdata()]);
        }
        if((int)imicro == _FloatingMicroSliceNumber)return true;
        if (imicro >= getMicrosliceNumber()) return false;
        if((int)imicro > _FloatingMicroSliceNumber && _FloatingMicroSliceNumber > -1){
            uint32_t* pointer = (uint32_t*)_FloatingMicroSlice.getBuffer();
            for(; _FloatingMicroSliceNumber < (int)imicro; _FloatingMicroSliceNumber++){
                 pointer += _FloatingMicroSlice.sizeofdata();
                 uint32_t sod = this->sizeofdata();
                 if ((pointer-(uint32_t*)this->_Buffer) >= sod) return false;
                _FloatingMicroSlice.setBufferSource(pointer);
            }
            return true;
        }

        // Firrst point to the first MicroSlice
        uint32_t* pointer = (uint32_t*)&RAWBUFF32[_Header.sizeofdata()];
        _FloatingMicroSlice.setBufferSource(pointer);

        for(_FloatingMicroSliceNumber = 0; _FloatingMicroSliceNumber < (int)imicro; _FloatingMicroSliceNumber++){
            pointer += _FloatingMicroSlice.sizeofdata();
            if ((pointer - (uint32_t*)this->_Buffer) >= this->sizeofdata()) return false;
            _FloatingMicroSlice.setBufferSource(pointer);
        }
        return true;
}

//------------------------------------------------------------------------------
// Printing the MilliSlice structure
void RawMilliSlice::print() const{
    printf("\n---MilliSlice---------------------------------------------size = %u words\n",this->sizeofdata());
    _Header.print();

    std::cout<<"\nMilliSlice Data RAW 32bit WORDs (converted, if required, to native endian format):";
    for(uint32_t i = 0; i < this->sizeofdata(); ++i){
        this->printWord(i,false);
        printf(" ");
    }

    std::cout<<"\nMilliSlice Data RAW 32bit Bits (converted, if required, to native endian format):";
    for(uint32_t i = 0; i < this->sizeofdata(); ++i){
            std::cout<<"\n    ";
            this->printWord(i,true);
    }

}

//------------------------------------------------------------------------------
// Size of the MilliSlice _Buffer in words
uint32_t RawMilliSlice::sizeofdata()const { return _Header.getSliceSize(); }

void
RawMilliSlice::tprintf( unsigned lvl, const char *fmt, ... )
{   if (lvl > 31) lvl=31;
    if (_tlvl & (1<<lvl))
    {   timeval 	tv;
	char            str[80];
	time_t          local;
	const char*     indent="                                ";//32 spaces
	va_list         ap;
	va_start( ap, fmt);
	gettimeofday( &tv, NULL );
	local = tv.tv_sec;
	strftime( str, sizeof(str), "%T", localtime(&local) );
	printf( "%s.%06ld%s", str, (long)tv.tv_usec, &indent[31-lvl] );
	vprintf( fmt, ap );
    }
}   // tprintf
