#include "DAQDataFormats/RawRun.h"
#include "DAQDataFormats/DataFormatException.h"
#include <boost/lexical_cast.hpp>
#include <cstdio>

#ifndef RAWBUFF32
// Define a macros to make accessing the raw buffer easier
#define RAWBUFF32 ((uint32_t*)this->_Buffer)
#define RAWBUFF16 ((uint16_t*)this->_Buffer)
#endif

using namespace daqdataformats;

//------------------------------------------------------------------------------
// Default constructor
RawRun::RawRun(uint32_t size){

    if (size < 2 * (rawrunhdr::WORD_CRC + 1)) {
            std::string msg("An external input buffer was given to the ");
            msg.append("RawRun constructor that was too small (size=");
            msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
            msg.append(boost::lexical_cast<std::string>(2 * (rawrunhdr::WORD_CRC + 1)) + ").");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
        }

    // Set the maximum size
    _max_buffer_size = size;

    // Allocate the internal Buffer
    _InternalBuffer.assign(_max_buffer_size,0);

    // Set pointers to point to the internal buffer
    this->resetBufferSource();

    this->init();
}
//------------------------------------------------------------------------------
RawRun::RawRun(const RawRun& copy_in){
   this->copy(copy_in);
}

//------------------------------------------------------------------------------
RawRun::RawRun(void* millislice_buffer, uint32_t size){

    if (size < 2 * (rawrunhdr::WORD_CRC + 1)) {
            std::string msg("An external input buffer was given to the ");
            msg.append("RawRun constructor that was too small (size=");
            msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
            msg.append(boost::lexical_cast<std::string>(2 * (rawrunhdr::WORD_CRC + 1)) + ").");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
        }
    // Set the maximum size
    _max_buffer_size = size;

    // Allocate the internal Buffer
    _InternalBuffer.reserve(_max_buffer_size);
    // Zero it out
    for(uint32_t i = 0; i < _max_buffer_size; ++i)
        _InternalBuffer[i] = 0;

    
    // First, initialize the Internal Buffer
            // Set pointers to point to the internal buffer
            this->resetBufferSource();

            // Initialize it
            this->init();

    // Now set the pointer to the external buffer
    this->setBufferSource(millislice_buffer);

}
//------------------------------------------------------------------------------
// Destructor
RawRun::~RawRun() { }

//------------------------------------------------------------------------------
// Initialization of Data members
void RawRun::init(){

    // Initialize the header
    _Header.init();

    // Initialize the _FloatingEventNumber
    _FloatingEventNumber = -1;

}

//------------------------------------------------------------------------------
void RawRun::copy(const RawRun& copy_in){

        // Set the maximum size
    _max_buffer_size = copy_in._max_buffer_size;

    // Setup the internal buffer
    _InternalBuffer.reserve(_max_buffer_size);

      // Copy the data from one buffer to the other
    for(uint32_t i = 0; i < _max_buffer_size; ++i)
        _InternalBuffer[i] = copy_in._InternalBuffer[i];

    // Setup the main buffer pointer
    // If the copy_in's buffer is external, point to that external memory
    if(copy_in.isInternalBuffer()) this->resetBufferSource();
    else                           this->setBufferSource(copy_in._Buffer);
    
    // Setup the FloatingEvent
    _FloatingEventNumber = copy_in._FloatingEventNumber;
    if(_FloatingEventNumber > -1)this->setFloatingEvent(copy_in._FloatingEventNumber);

}


//------------------------------------------------------------------------------
// Add a Event into the Run changing the Header appropriately
void RawRun::addEvent(void* event_buffer){

// Check if the size of the Event would exceed the FreeMemory. Check only if the Run buffer is Internal.
   if(this->isInternalBuffer()){
                RawEventHeader header_check(event_buffer);
                uint32_t slicesize = header_check.getEventSize();
                uint32_t free = this->getFreeMemory();
                if (free < slicesize) {
                    std::string msg("Cannot add a Event. ");
                    msg.append("Run buffer is full (total Run buffer size=");
                    msg.append(boost::lexical_cast<std::string>(_max_buffer_size) + ", current data size=");
                    msg.append(boost::lexical_cast<std::string>(this->sizeofdata()) + ", Event size = ");
                    msg.append(boost::lexical_cast<std::string>(slicesize) + ")");
                    GENERATE_DATAFORMAT_EXCEPTION(msg);
                }
    }

    // Set the floating event in the end of the current Buffer
    _FloatingEvent.setBufferSource(&RAWBUFF32[this->sizeofdata()]);
    // Read the Event
    _FloatingEvent.readData(event_buffer);

    // Setup the RunHeader Header
    // Advance the Event number in the Run
    _Header.advanceEventNumber();
    // Change the Run size

    _Header.addRunSize(_FloatingEvent.sizeofdata());

    // Adding Trigger information
    RawTrigger* trigger = _FloatingEvent.getTrigger();
    RawTriggerHeader* trigger_header = trigger->getTriggerHeader();
    RawTriggerMask*   trigger_mask = trigger->getTriggerMask();
    std::cout << "Adding event; trigger SIM = " << trigger_mask->getMonteCarloFlag() << std::endl;
    _Header.setTriggerVersion(trigger_header->getVersion());
    _Header.setTriggerMC(trigger_mask->getMonteCarloFlag());
    _Header.setVTMod(trigger_mask->getTrigMod());
    _Header.addValidTriggerTypesLow(trigger_mask->getTrigMask_Low());
    _Header.addValidTriggerTypesHigh(trigger_mask->getTrigMask_High());
    _Header.addValidTriggerTypesHigh2(trigger_mask->getTrigMask_High2());





}

// Closing the Run by adding the Index (if available), tail and the CRC
bool RawRun::close(){


    void* current_buffer = &RAWBUFF32[this->sizeofdata()]; //Remember the current Buffer position
    if (getTotalEvents() != 0) {
      setFloatingEvent(0);
      _Header.setUniversalEventStart(getFloatingEvent()->getEventNumber());
      
      setFloatingEvent(getTotalEvents() - 1);
      _Header.setUniversalEventEnd(getFloatingEvent()->getEventNumber());
    }

   // Add Run Header size + 1 words to the buffer as we
   // will add the tail (another Run header) and the CRC-word in the end of the Run
   _Header.addRunSize(_Header.sizeofdata() + 1);
   // Close the Run Header. Produces Header CRC
   _Header.close();

    // Adding a tail

       // Create a new Tail and assign its pointer
       RawRunTail tail(current_buffer);
   
       // Copy the content of the _Header into the tail
       tail.readData(_Header.getBuffer());
       // Set an appropriate Markers for the tail
       tail.setTail();
       // Close the Run Tail. Produces Tail CRC
       tail.close();
       //tail is ready

   // Finally, add Run CRC at the very end of _Buffer
   this->addCheckSum();
   return true;
}

//------------------------------------------------------------------------------
// This is a general method to read the event from millislice_buffer to _Buffer
void* RawRun::readData(void* millislice_buffer){
    //Read the header and unpack it to find out the size of the millislice
    millislice_buffer = _Header.readData(millislice_buffer);

    //Now, read the entire millislice_buffer to the _Buffer
    return this->readData(millislice_buffer, this->sizeofdata() - _Header.sizeofdata(), _Header.sizeofdata());
}

//------------------------------------------------------------------------------
void* RawRun::readData(void* void_buffer, uint32_t size){

  uint32_t* buffer = (uint32_t*)void_buffer;

  if(buffer == 0){return 0;}

  for(uint32_t i = 0; i < size; ++i )
    RAWBUFF32[i] = buffer[i];

  // Return the address of the buffer advanced by 1 word;
  return (void*)(buffer + size);
}


//------------------------------------------------------------------------------
void* RawRun::readData(void* buffer_void, uint32_t size, uint32_t _Buffer_start_position){

  uint32_t* buffer = (uint32_t*)buffer_void;

  if(buffer == 0){return 0;}

  for(uint32_t i = 0; i < size; ++i )
    RAWBUFF32[i+_Buffer_start_position] = buffer[i];

  // Return the address of the buffer advanced by 1 word;
  return (void*)(buffer+size);
}

//------------------------------------------------------------------------------
//Getters
RawRunHeader* RawRun::getHeader()                 { return &_Header; }

// Assuming the run is closed. Don't use it on non-closed runs!
RawRunTail* RawRun::getTail(){

    uint32_t* pointer = &RAWBUFF32[0];
    pointer += this->sizeofdata() - _Header.sizeofdata() - 1;

    _Tail.setBufferSource(pointer);
    return &_Tail;
}

RawEvent*     RawRun::getFloatingEvent()          { return &_FloatingEvent; }
RawConfigurationBlock* RawRun::getConfigurationBlock() { 
  return &_ConfigurationBlock; 
}
    /// Getters
uint32_t RawRun::getMarkerHi() const { return _Header.getMarkerHi();}
uint32_t RawRun::getMarkerLo() const { return _Header.getMarkerLo();}
uint32_t RawRun::getRunNumber() const { return _Header.getRunNumber();}
uint32_t RawRun::getRunType() const { return _Header.getRunType();}
uint32_t RawRun::getConfigIDX() const { return _Header.getConfigIDX();}
uint32_t RawRun::getPartition() const { return _Header.getPartition();}
uint32_t RawRun::getSubrun() const { return _Header.getSubrun();}
uint32_t RawRun::getConfigurationBlockSet() const { return _Header.getConfigurationBlockSet();}
uint32_t RawRun::getRunControlVersion() const { return _Header.getRunControlVersion();}
uint32_t RawRun::getRunControlID() const { return _Header.getRunControlID();}
uint32_t RawRun::getDataLoggerVersion() const { return _Header.getDataLoggerVersion();}
uint32_t RawRun::getDataLoggerID() const { return _Header.getDataLoggerID();}
uint32_t RawRun::getRunStartTime() const { return _Header.getRunStartTime();}
uint32_t RawRun::getSubRunStartTime() const { return _Header.getSubRunStartTime();}
uint32_t RawRun::getSubRunEndTime () const { return _Header.getSubRunEndTime ();}
uint32_t RawRun::getRunEndTime() const { return _Header.getRunEndTime();}
uint32_t RawRun::getTotalEvents() const { return _Header.getTotalEvents();}
uint32_t RawRun::getUniversalEventStartLo() const { return _Header.getUniversalEventStartLo();}
uint32_t RawRun::getUniversalEventStartHi() const { return _Header.getUniversalEventStartHi();}
uint64_t RawRun::getUniversalEventStart() const { return _Header.getUniversalEventStartHi();}
uint32_t RawRun::getUniversalEventEndLo() const { return _Header.getUniversalEventEndLo();}
uint32_t RawRun::getUniversalEventEndHi() const { return _Header.getUniversalEventEndHi();}
uint64_t RawRun::getUniversalEventEnd() const { return _Header.getUniversalEventEnd();}
uint32_t RawRun::getRunSize() const { return _Header.getRunSize();}
uint32_t RawRun::getTriggerVersion() const { return _Header.getTriggerVersion();}
uint32_t RawRun::getTriggerCtrlID() const { return _Header.getTriggerCtrlID();}
uint32_t RawRun::getTriggerMC() const { return _Header.getTriggerMC();}
uint32_t RawRun::getVTMod() const { return _Header.getVTMod();}
uint32_t RawRun::getValidTriggerTypesLow() const { return _Header.getValidTriggerTypesLow();}
uint32_t RawRun::getValidTriggerTypesHigh() const { return _Header.getValidTriggerTypesHigh();}
uint32_t RawRun::getValidTriggerTypesHigh2() const { return _Header.getValidTriggerTypesHigh2();}
uint32_t RawRun::getTriggerListIDX() const { return _Header.getTriggerListIDX();}
uint32_t RawRun::getTriggerPrescaleListIDX() const { return _Header.getTriggerPrescaleListIDX();}


uint32_t             RawRun::getFreeMemory() const {
        // Allocated size minus (current buffer size + Tail size + CRC + 1)
        if(this->sizeofdata() + _Header.sizeofdata() + 1 > _max_buffer_size) return 0;
        return _max_buffer_size - this->sizeofdata() - _Header.sizeofdata() - 1;
}

//------------------------------------------------------------------------------
//Setters
    bool RawRun::setConfigurationBlock  (RawConfigurationBlock&  copyin){
        _ConfigurationBlock.readData(copyin.getBuffer());

        _Header.addRunSize(_ConfigurationBlock.sizeofdata());
	_Header.setConfigurationBlockSet(1);

        return true;
    }
    bool RawRun::setConfigurationBlock   (void*  pointer){
        _ConfigurationBlock.readData(pointer);
	_Header.setConfigurationBlockSet(1);

        _Header.addRunSize(_ConfigurationBlock.sizeofdata());

        return true;
    }





bool RawRun::setRunNumber             (uint32_t w) { return _Header.setRunNumber(w);}
bool RawRun::setRunType               (uint32_t w) { return _Header.setRunType(w);}
bool RawRun::setConfigIDX             (uint32_t w) { return _Header.setConfigIDX(w);}
bool RawRun::setPartition             (uint32_t w) { return _Header.setPartition(w);}
bool RawRun::setSubrun                (uint32_t w) { return _Header.setSubrun(w);}
bool RawRun::setRunControlVersion     (uint32_t w) { return _Header.setRunControlVersion(w);}
bool RawRun::setRunControlID          (uint32_t w) { return _Header.setRunControlID(w);}
bool RawRun::setDataLoggerVersion     (uint32_t w) { return _Header.setDataLoggerVersion(w);}
bool RawRun::setDataLoggerID          (uint32_t w) { return _Header.setDataLoggerID(w);}
bool RawRun::setRunStartTime          (uint32_t w) { return _Header.setRunStartTime(w);}
bool RawRun::setSubRunStartTime       (uint32_t w) { return _Header.setSubRunStartTime(w);}
bool RawRun::setSubRunEndTime         (uint32_t w) { return _Header.setSubRunEndTime(w);}
bool RawRun::setRunEndTime            (uint32_t w) { return _Header.setRunEndTime(w);}
bool RawRun::setTriggerCtrlID         (uint32_t w) { return _Header.setTriggerCtrlID(w);}
bool RawRun::setTriggerVersion        (uint32_t w) { return _Header.setTriggerVersion(w);}
bool RawRun::setTriggerListIDX        (uint32_t w) { return _Header.setTriggerListIDX(w);}
bool RawRun::setTriggerPrescaleListIDX(uint32_t w) { return _Header.setTriggerPrescaleListIDX(w);}


bool RawRun::setBufferSource(void* pointer){
        _Buffer = pointer;
        _Header.setBufferSource(pointer);

        uint32_t* pp = (uint32_t*)pointer + _Header.sizeofdata();
        _ConfigurationBlock.setBufferSource(pp);
 
        // reset the Floating Event number
        _FloatingEventNumber = -1;
        return true;
}
bool RawRun::resetBufferSource(){
        this->setBufferSource(&_InternalBuffer[0]);
        return true;
}

// clears all data in the buffer
bool RawRun::clear(){
        uint32_t max_data = this->sizeofdata();

        // If it exceeds, then it means that External buffer was used.
        if(this->isInternalBuffer() && max_data > _max_buffer_size)
            max_data = _max_buffer_size;

        // Loop over the buffer and zero it
        for(uint32_t i = 0; i < max_data; ++i)
            RAWBUFF32[i]=0;

        return true;
}
bool RawRun::reset(){
        this->resetBufferSource();
        this->clear();
        return true;
}
bool RawRun::isInternalBuffer() const {

        if( _Buffer == (void*)(&_InternalBuffer[0]))
            return true;
        return false;
}

//------------------------------------------------------------------------------
bool RawRun::setFloatingEvent(uint32_t ievent){
        if(ievent == 0){
            _FloatingEventNumber = ievent;
	    
            if (this->getConfigurationBlockSet())
                return _FloatingEvent.setBufferSource(&RAWBUFF32[_Header.sizeofdata() + _ConfigurationBlock.sizeofdata()]);
            return _FloatingEvent.setBufferSource(&RAWBUFF32[_Header.sizeofdata()]);
        }// end of if for the event 0

        if((int)ievent == _FloatingEventNumber)return true;

        if (ievent >= this->getTotalEvents()) return false;

        if((int)ievent > _FloatingEventNumber && _FloatingEventNumber > -1){
            uint32_t* pointer = (uint32_t*)_FloatingEvent.getBuffer();
            for(; _FloatingEventNumber < (int)ievent; ++_FloatingEventNumber){
                 pointer += _FloatingEvent.sizeofdata();
                 if ((pointer - RAWBUFF32) >= this->sizeofdata()) return false;
                _FloatingEvent.setBufferSource(pointer);
            }// end of loop
            return true;
        }// end of if

        // First point to the first Event
        uint32_t* pointer;
        if (this->getConfigurationBlockSet())
            pointer = (uint32_t*)&RAWBUFF32[_Header.sizeofdata() + _ConfigurationBlock.sizeofdata()];
	else
            pointer = (uint32_t*)&RAWBUFF32[_Header.sizeofdata()];

        _FloatingEvent.setBufferSource(pointer);

        for(_FloatingEventNumber = 0; _FloatingEventNumber < (int)ievent; _FloatingEventNumber++){
            pointer += _FloatingEvent.sizeofdata();
            if ((pointer - RAWBUFF32) >= this->sizeofdata()) return false;
            _FloatingEvent.setBufferSource(pointer);
        }
        return true;
}

//------------------------------------------------------------------------------
// Printing the Run structure
void RawRun::print() const{
    printf("\n---Run---------------------------------------------size = %u words\n",this->sizeofdata());
    _Header.print();

    std::cout<<"\nRun Data RAW 32bit WORDs (converted, if required, to native endian format):";
    for(uint32_t i = 0; i < this->sizeofdata(); ++i){
        this->printWord(i,false);
        printf(" ");
    }

    std::cout<<"\nRun Data RAW 32bit Bits (converted, if required, to native endian format):";
    for(uint32_t i = 0; i < this->sizeofdata(); ++i){
            std::cout<<"\n    ";
            this->printWord(i,true);
    }

}

//------------------------------------------------------------------------------
// Size of the Run _Buffer in words
uint32_t RawRun::sizeofdata() const { return _Header.getRunSize(); }
