#include "DAQDataFormats/RawTimingMarker.h"
#include "DAQDataFormats/TimeStampCounter.h"
#include <cstdio>

#ifndef RAWBUFF32
// Define a macros to make accessing the raw buffer easier
#define RAWBUFF32 ((uint32_t*)this->_Buffer)
#define RAWBUFF16 ((uint16_t*)this->_Buffer)
#endif

using namespace daqdataformats;
using namespace timmarker;

//------------------------------------------------------------------------------
// Default constructor
RawTimingMarker::RawTimingMarker(){
    // Set pointers to point to the internal buffer
    this->resetBufferSource();

    // Initialize it
    this->init();
}
//------------------------------------------------------------------------------
RawTimingMarker::RawTimingMarker(const RawTimingMarker& copy_in){
    this->copy(copy_in);
}

//------------------------------------------------------------------------------
RawTimingMarker::RawTimingMarker(void* timingmarker_buffer){
    // First, initialize the Internal Buffer
            // Set pointers to point to the internal buffer
            this->resetBufferSource();

            // Initialize it
            this->init();

    // Now set the pointer to the external buffer
    this->setBufferSource(timingmarker_buffer);
}

//------------------------------------------------------------------------------
// Destructor
RawTimingMarker::~RawTimingMarker() { }


//------------------------------------------------------------------------------
// Initialize data members
void RawTimingMarker::init(){
    for(uint32_t i = 0; i < RAWTIMINGMARKER_SIZE; i++)
            RAWBUFF32[i] = 0;
}
//------------------------------------------------------------------------------
// Deep Copy Method
void RawTimingMarker::copy(const RawTimingMarker& copy_in){
    // Copy the data from one buffer to the other
    for(int i = 0; i < RAWTIMINGMARKER_SIZE; ++i)
        _InternalBuffer[i] = copy_in._InternalBuffer[i];

    // Setup the main buffer pointer
    // If the copy_in's buffer is external, point to that external memory
    if(copy_in.isInternalBuffer()) this->resetBufferSource();
    else                           this->setBufferSource(copy_in._Buffer);
}

//------------------------------------------------------------------------------
// Getters
uint32_t RawTimingMarker::getLowWord() const  {
    return BITFIELDGET(RAWBUFF32[WORD_Lo], MASK_Lo, SHIFT_Lo);
}
uint32_t RawTimingMarker::getHighWord() const {
    return BITFIELDGET(RAWBUFF32[WORD_Hi], MASK_Hi, SHIFT_Hi);
}
/**
 * Return the timestamp as a single 64-bit variable
 *
 * @return full timestamp in clock ticks
 */
uint64_t RawTimingMarker::getTime() const {
      uint64_t timestamp;
      uint32_t low, high;

      low  = this->getLowWord();
      high = this->getHighWord();

      timestamp=(uint64_t)high<<32 | low;
      return timestamp;
}


//------------------------------------------------------------------------------
// Setters
// Set Method for Byte Count
bool RawTimingMarker::setLowWord (uint32_t word) {
        BITFIELDSET(RAWBUFF32[WORD_Lo], word , MASK_Lo, SHIFT_Lo);
        return true;
}
// Set Method for Byte Count
bool RawTimingMarker::setHighWord(uint32_t word) {
        BITFIELDSET(RAWBUFF32[WORD_Hi], word, MASK_Hi, SHIFT_Hi);
        return true;
}
/**
 * Set time using 64-bit argument
 *
 * @param timestamp as 64-bit
 */
bool RawTimingMarker::setTime(uint64_t timestamp) {
      uint32_t low, high;

     low  = timestamp & 0xffffffff;
     high = timestamp >> 32;

     this->setHighWord(high);
     this->setLowWord(low);

     return true;
}

bool RawTimingMarker::setBufferSource(void* pointer){
        _Buffer                = pointer;
        return true;
}
    bool RawTimingMarker::resetBufferSource(){
        this->setBufferSource(&_InternalBuffer[0]);
        return true;
    }

// clears all data in the buffer
    bool RawTimingMarker::clear(){
        int max_data = this->sizeofdata();

        // Loop over the buffer and zero it
        for(int i=0; i< max_data; ++i)
            RAWBUFF32[i]=0;

        return true;
    }
    bool RawTimingMarker::reset(){
        this->resetBufferSource();
        this->clear();
        return true;
    }
    bool RawTimingMarker::isInternalBuffer() const {

        if((this->_Buffer) == (void*)(&_InternalBuffer[0]))
            return true;
        return false;
  }


//------------------------------------------------------------------------------
// Generate a time stamp
void RawTimingMarker::generate(){
    cycle_t full_timestamp;

    full_timestamp.full_time = tsc();

    this->setLowWord(full_timestamp.time[0]);
    this->setLowWord(full_timestamp.time[1]);
}

//------------------------------------------------------------------------------
// Print the Timing Marker
void RawTimingMarker::print() const{
    printf("\nTiming Marker: HighWord=%u LowWord=%u",this->getHighWord(),
                                                     this->getLowWord());
    printf("  RAW=");
        this->printWord(0,false);
    printf(" ");
        this->printWord(1,false);
}

//------------------------------------------------------------------------------
// Size of the timing marker in words
uint32_t RawTimingMarker::sizeofdata()const{
    // Return the size of the data block in words
    return RAWTIMINGMARKER_SIZE;
}
