#include "DAQDataFormats/RawTrigger.h"
#include "DAQDataFormats/DataFormatException.h"
#include <boost/lexical_cast.hpp>


#ifndef RAWBUFF32
// Define a macros to make accessing the raw buffer easier
#define RAWBUFF32 ((uint32_t*)this->_Buffer)
#define RAWBUFF16 ((uint16_t*)this->_Buffer)
#endif
 
using namespace daqdataformats;

#define MASK_HWFLAG
 
RawTrigger::RawTrigger(){
  init();			// Now initialize everything
}

RawTrigger::~RawTrigger(){
  if(_trigBuffer){
    free(_trigBuffer);		// Free the memory allocated for the data block
  }
}


//
// Constructor Methods
//

// Constructor with Initialization
RawTrigger::RawTrigger(void* input_buffer, uint32_t size){
  init(input_buffer, size);
  return;
}

// Deep Copy Constructor
RawTrigger::RawTrigger(const RawTrigger& copy_in){
#ifdef COPY_COPIES_INTERNAL
  // Initialize the member data
  this->init();
  // Call the copy method to do the deep copy
  this->copy(copy_in);
#else
  // Check to see what type of buffer we have
  if(copy_in.isInternalBuffer()){
    // Internal to Internal initialization
    this->init();
    this->copy(copy_in);
  }else{
    // Copy_in uses an external buffer so we only 
    // copy the point to the external block and set the
    // size of that buffer
    this->init(copy_in._Buffer, copy_in._max_buffer_size);
  }
#endif
  return;
}
 
// General Initialization 
void RawTrigger::init(void* input_buffer, uint32_t size){

  // Zero out the general data buffer pointer
  _trigBuffer = NULL;

  // Determine how large a data buffer we need
  BLOCK_WORDS  = 0;
  BLOCK_WORDS += _theHeader.sizeofdata(); 
  BLOCK_WORDS += _theTime.sizeofdata(); 
  BLOCK_WORDS += _theMarker.sizeofdata(); 
  BLOCK_WORDS += _theRange.sizeofdata(); 
  BLOCK_WORDS += _theMask.sizeofdata(); 

  if(input_buffer){
    this->setBufferSource(input_buffer);
  }else{
    // Get a data block of the appropriate size
    _trigBuffer = (uint32_t*)calloc(BLOCK_WORDS, sizeof(uint32_t));

    assert(_trigBuffer);

    this->resetBufferSource();
    this->clear();

    _theHeader.setDelimiter(TriggerHeaderDelimiter);
  }

  if(size){
    if (size < sizeofdata()) {
      std::string msg("An external input buffer was given to the ");
      msg.append("RawTrigger constructor that was too small (size=");
      msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
      msg.append(boost::lexical_cast<std::string>(sizeofdata()) + ").");
      GENERATE_DATAFORMAT_EXCEPTION(msg);
    }
    _max_buffer_size = size;
  }else{
    _max_buffer_size = BLOCK_WORDS;
  }


  return;
}

// Shallow copy method
bool RawTrigger::copy(void* input_buffer, uint32_t size){
  if(input_buffer){
    this->setBufferSource(input_buffer);
  }else{
    return false;
  }

  if(size){
    if (size < sizeofdata()) {
      std::string msg("An external input buffer was given to the ");
      msg.append("RawTrigger constructor that was too small (size=");
      msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
      msg.append(boost::lexical_cast<std::string>(sizeofdata()) + ").");
      GENERATE_DATAFORMAT_EXCEPTION(msg);
    }
    _max_buffer_size = size;
  }else{
    _max_buffer_size = BLOCK_WORDS;
  }
  
  return true;
}

 
// Deep Copy Method
bool RawTrigger::copy(const RawTrigger& copy_in){

  // To do this properly, we need to assume that the copy_in
  // could have a different size data block than our 
  // currently allocated block.
  // (of course, this doesn't really support versioning since
  // we still assume a fixed location for the trigger components)

  if(copy_in._max_buffer_size < sizeofdata()) {
    std::string msg("The RawTrigger copy method was called with a ");
    msg.append("source instance that has a buffer that is too small (size=");
    msg.append(boost::lexical_cast<std::string>(copy_in._max_buffer_size));
    msg.append(", minimum size=");
    msg.append(boost::lexical_cast<std::string>(sizeofdata()) + ").");
    GENERATE_DATAFORMAT_EXCEPTION(msg);
  }

  // reallocate the internal buffer, if needed
  if(this->isInternalBuffer() && copy_in._max_buffer_size > _max_buffer_size) {
    free(_trigBuffer);
    _trigBuffer =
           (uint32_t*)calloc(copy_in._max_buffer_size,sizeof(uint32_t));
    assert(_trigBuffer);
    this->setBufferSource(_trigBuffer);
  }
  _max_buffer_size = copy_in._max_buffer_size;

  // Copy the raw packed block
  for(unsigned i=0; i<_max_buffer_size; ++i){
    RAWBUFF32[i] = ((uint32_t*)(copy_in._Buffer))[i];
  } // endfor
  
  return true;
}

 
//------------------------------
// GET METHODS
//

// NONE

//------------------------------
// SET METHODS
//
 
// NONE

//------------------------------
// RESET/CLEAR METHODS
//

bool RawTrigger::reset(){
  this->resetBufferSource();
  this->clear();
  return true;
}

bool RawTrigger::clear(){
  int max_data = this->sizeofdata();


  /// Loop over the buffer and zero it
  for(int i=0; i< max_data; ++i){
    RAWBUFF32[i]=0;
  }
  return true;
}

bool RawTrigger::resetBufferSource(){
  return this->setBufferSource(_trigBuffer);
}
// Method to return the size of the data block in words
bool RawTrigger::setBufferSource(void* input_buffer){

    if(input_buffer){
    this->_Buffer = input_buffer;

    // Now, setup the other pointers
        uint32_t* bufferptr = (uint32_t*)input_buffer;
        uint32_t offset = 0;

        _theHeader.setBufferSource(bufferptr + offset);
        offset += _theHeader.sizeofdata();

        _theTime.setBufferSource(bufferptr + offset);
        offset += _theTime.sizeofdata();

        _theMarker.setBufferSource(bufferptr + offset);
        offset += _theMarker.sizeofdata();

        _theRange.setBufferSource(bufferptr + offset);
        offset += _theRange.sizeofdata();

        _theMask.setBufferSource(bufferptr + offset);
        offset += _theMask.sizeofdata();
    return true;
  }


  return false;
}

//------------------------------
// INQUIRIY METHODS (CONST)
//

// Method to return the size of the data block in words
bool RawTrigger::isInternalBuffer() const {
  if((this->_Buffer) == (void*)(this->_trigBuffer)){
    return true;
  }else{
    return false;
  }

  return false;
}

//------------------------------
// WRITE METHODS
//
// Method to write the data out to a raw buffer
void* RawTrigger::writeData(void* buffer){
  uint32_t* write_buffer;

  if(buffer == 0){return 0;}

  write_buffer = (uint32_t*)buffer;

  for(unsigned i=0; i < this->sizeofdata(); ++i){
    *write_buffer = RAWBUFF32[i];
    write_buffer++;
  } // endfor

  return write_buffer;
}

// Method to write the data out to a file descriptor
ssize_t RawTrigger::writeData(int fd){
  if(fd>0){
    return write(fd,_Buffer,(size_t)this->sizeofdata()*sizeof(uint32_t));
  }else{
    return -1;
  }
}

//
// READ METHODS
//

// Method to read the data in from a raw buffer
void* RawTrigger::readData(void* buffer){
  uint32_t* read_buffer;

  if(buffer == 0){return 0;}

  read_buffer = (uint32_t*)buffer;

  for(unsigned i=0; i< this->sizeofdata(); ++i){
    RAWBUFF32[i] = *read_buffer;
    read_buffer++;
  } // endfor

  return read_buffer;
}

//
// SIZEOF METHODS
//

// Method to return the size of the data block in words
uint32_t RawTrigger::sizeofdata()const{
  return BLOCK_WORDS;
}

// Assignment Operator
bool RawTrigger::operator=(const RawTrigger& copy_in){
  return (this->copy(copy_in));
}


void RawTrigger::print() const{


    _theHeader.print();

    _theTime.print();

    _theMarker.print();

    _theRange.print();

    _theMask.print();

}
