#include "DAQDataFormats/RawTriggerHeader.h"
#include "DAQDataFormats/DataFormatException.h"
#include <boost/lexical_cast.hpp>
#include <cstdio>


#ifndef RAWBUFF32
// Define a macros to make accessing the raw buffer easier
#define RAWBUFF32 ((uint32_t*)this->_Buffer)
#define RAWBUFF16 ((uint16_t*)this->_Buffer)
#endif
 
using namespace daqdataformats;
 
RawTriggerHeader::RawTriggerHeader(){
  init();			// Initialize everything
}
 
RawTriggerHeader::~RawTriggerHeader(){
  // No memory is allocated in this class
}
 
 
//
// Constructor Methods
//
 
// Constructor with Initialization
RawTriggerHeader::RawTriggerHeader(void* input_buffer, uint32_t size){
  init(input_buffer, size);
  return;
}
 
// Deep Copy Constructor
RawTriggerHeader::RawTriggerHeader(const RawTriggerHeader& copy_in){
#ifdef COPY_COPIES_INTERNAL
  // Initialize the member data
  this->init();
  // Call the copy method to do the deep copy
  this->copy(copy_in);
#else
  // Check to see what type of buffer we have
  if(copy_in.isInternalBuffer()){
    // Internal to Internal initialization
    this->init();
    this->copy(copy_in);
  }else{
    // Copy_in uses an external buffer so we only 
    // copy the point to the external block and set the
    // size of that buffer
    this->init(copy_in._Buffer, copy_in._max_buffer_size);
  }
#endif
  return;
}
 
// General Initialization 
void RawTriggerHeader::init(void* input_buffer, uint32_t size){
  if(input_buffer){
    this->setBufferSource(input_buffer);
  }else{
    this->setBufferSource(_header);
    this->clear();
    
    //Set the TriggerHeader's delimiter
    this->setDelimiter(TriggerHeaderDelimiter);
  }

  if(size){
    if (size < sizeofdata()) {
      std::string msg("An external input buffer was given to the ");
      msg.append("RawTriggerHeader constructor that was too small (size=");
      msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
      msg.append(boost::lexical_cast<std::string>(sizeofdata()) + ").");
      GENERATE_DATAFORMAT_EXCEPTION(msg);
    }
    _max_buffer_size = size;
  }else{
    _max_buffer_size = HEADER_WORDS;
  }

  
  return;
}

// Shallow Copy Method
bool RawTriggerHeader::copy(void* input_buffer, uint32_t size){
  // This is a shallow copy, so we only set the pointer
  // to the data buffer and the size of the data if passed
  if(input_buffer){
    this->setBufferSource(input_buffer);
  }else{
    return false;
  }

  if(size){
    if (size < sizeofdata()) {
      std::string msg("An external input buffer was given to the ");
      msg.append("RawTriggerHeader copy method that was too small (size=");
      msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
      msg.append(boost::lexical_cast<std::string>(sizeofdata()) + ").");
      GENERATE_DATAFORMAT_EXCEPTION(msg);
    }
    _max_buffer_size = size;
  }else{
    _max_buffer_size = HEADER_WORDS;
  }

  return true;
}
 
// Deep Copy Method
bool RawTriggerHeader::copy(const RawTriggerHeader& copy_in){

  // Copy the max buffer size
  _max_buffer_size = copy_in._max_buffer_size;

  // Copy the data from one buffer to the other
  for(uint32_t i = 0; i < HEADER_WORDS; ++i){
    RAWBUFF32[i] = ((uint32_t*)(copy_in._Buffer))[i];
  } // endfor i
  
  return true;
}
 
//
// GET METHODS
//

uint32_t RawTriggerHeader::getDelimiter() const{
  return BITFIELDGET(RAWBUFF32[WORD_DELIMITER], MASK_DELIMITER, SHIFT_DELIMITER);
}
// get Method for Hardware Flag
uint32_t RawTriggerHeader::getHWFlag() const {
  return BITFIELDGET(RAWBUFF32[WORD_HWFLAG], MASK_HWFLAG, SHIFT_HWFLAG);
}

// get Method for Version
uint32_t RawTriggerHeader::getVersion() const {
  return BITFIELDGET(RAWBUFF32[WORD_VERSION], MASK_VERSION, SHIFT_VERSION);
}

// get Method for Source
uint32_t RawTriggerHeader::getSource() const {
  return BITFIELDGET(RAWBUFF32[WORD_SOURCE], MASK_SOURCE, SHIFT_SOURCE);
}

// get Method for Source SubID
uint32_t RawTriggerHeader::getSourceSubID() const {
  return BITFIELDGET(RAWBUFF32[WORD_SOURCE_SUBID], MASK_SOURCE_SUBID, SHIFT_SOURCE_SUBID);
}

// get Method for Source ID
uint32_t RawTriggerHeader::getSourceID() const {
  return BITFIELDGET(RAWBUFF32[WORD_SOURCE_ID], MASK_SOURCE_ID, SHIFT_SOURCE_ID);
}

// get Method for Trigger Low Bits
uint32_t RawTriggerHeader::getTriggerNumberLow() const {
  return BITFIELDGET(RAWBUFF32[WORD_TRIG_LOW], MASK_TRIG_LOW, SHIFT_TRIG_LOW);
}

// get Method for Trigger High Bits
uint32_t RawTriggerHeader::getTriggerNumberHigh() const {
  return BITFIELDGET(RAWBUFF32[WORD_TRIG_HIGH], MASK_TRIG_HIGH, SHIFT_TRIG_HIGH);
}

// get Method for a full (64bit) wide trigger number (40bits in use)
uint64_t RawTriggerHeader::getTriggerNumber() const {
  uint64_t triggerNumber;
  register uint32_t low, high;


  low  = this->getTriggerNumberLow(); 
  high = this->getTriggerNumberHigh(); 
  triggerNumber=(uint64_t)high<<32 | low;

  return triggerNumber;
}
// get Method for Trigger Low Bits
uint32_t RawTriggerHeader::getMasterTriggerNumberLow() const {
  return BITFIELDGET(RAWBUFF32[WORD_MASTERTRIGGER_LOW], MASK_MASTERTRIGGER_LOW, SHIFT_MASTERTRIGGER_LOW);
}

// get Method for Trigger High Bits
uint32_t RawTriggerHeader::getMasterTriggerNumberHigh() const {
  return BITFIELDGET(RAWBUFF32[WORD_MASTERTRIGGER_HIGH], MASK_MASTERTRIGGER_HIGH, SHIFT_MASTERTRIGGER_HIGH);
}

// get Method for a full (64bit) wide trigger number (40bits in use)
uint64_t RawTriggerHeader::getMasterTriggerNumber() const {
  uint64_t mastertriggerNumber;
  register uint32_t low, high;


  low  = this->getMasterTriggerNumberLow();
  high = this->getMasterTriggerNumberHigh();
  mastertriggerNumber=(uint64_t)high<<32 | low;

  return mastertriggerNumber;
}

//
// SET METHODS
//

// set Method for Delimiter
bool RawTriggerHeader::setDelimiter(uint32_t delim){
  BITFIELDSET( RAWBUFF32[WORD_DELIMITER], delim, MASK_DELIMITER , SHIFT_DELIMITER);
  return true;
}

// set Method for Version
bool RawTriggerHeader::setVersion(uint32_t version){
  BITFIELDSET( RAWBUFF32[WORD_VERSION], version, MASK_VERSION , SHIFT_VERSION );
  return true;
}

// set Method for Hardware Flag
bool RawTriggerHeader::setHWFlag(uint32_t flag){
  BITFIELDSET( RAWBUFF32[WORD_HWFLAG], flag, MASK_HWFLAG , SHIFT_HWFLAG );
  return true;
}

// set Method for Source
bool RawTriggerHeader::setSource(trigSource source){
  BITFIELDSET( RAWBUFF32[WORD_SOURCE], source, MASK_SOURCE , SHIFT_SOURCE );
  return true;
}

// set Method for Source ID
bool RawTriggerHeader::setSourceID(trigSourceID source){
  BITFIELDSET( RAWBUFF32[WORD_SOURCE_ID], source, MASK_SOURCE_ID , SHIFT_SOURCE_ID );
  return true;
}

// set Method for Source ID
bool RawTriggerHeader::setSourceSubID(trigSourceSubID source){
  BITFIELDSET( RAWBUFF32[WORD_SOURCE_SUBID], source, MASK_SOURCE_SUBID , SHIFT_SOURCE_SUBID );
  return true;
}

// set Method for Trigger Number Low
bool RawTriggerHeader::setTriggerNumberLow(uint32_t trig_low){
  BITFIELDSET( RAWBUFF32[WORD_TRIG_LOW], trig_low, MASK_TRIG_LOW , SHIFT_TRIG_LOW );
  return true;
}

// set Method for Trigger Number High
bool RawTriggerHeader::setTriggerNumberHigh(uint32_t trig_high){
  BITFIELDSET( RAWBUFF32[WORD_TRIG_HIGH], trig_high, MASK_TRIG_HIGH , SHIFT_TRIG_HIGH );
  return true;
}

// set Method for Trigger Number High
bool RawTriggerHeader::setTriggerNumber(uint64_t trig){
  register uint32_t low, high;

  low  = trig &  0xffffffff;
  high = trig >> 32;

  this->setTriggerNumberLow (low);
  this->setTriggerNumberHigh(high);
    
  return true;
}
// set Method for Trigger Number Low
bool RawTriggerHeader::setMasterTriggerNumberLow(uint32_t trig_low){
  BITFIELDSET( RAWBUFF32[WORD_MASTERTRIGGER_LOW], trig_low, MASK_MASTERTRIGGER_LOW , SHIFT_MASTERTRIGGER_LOW );
  return true;
}

// set Method for Trigger Number High
bool RawTriggerHeader::setMasterTriggerNumberHigh(uint32_t trig_high){
  BITFIELDSET( RAWBUFF32[WORD_MASTERTRIGGER_HIGH], trig_high, MASK_MASTERTRIGGER_HIGH , SHIFT_MASTERTRIGGER_HIGH );
  return true;
}

// set Method for Trigger Number High
bool RawTriggerHeader::setMasterTriggerNumber(uint64_t trig){
  register uint32_t low, high;

  low  = trig &  0xffffffff;
  high = trig >> 32;

  this->setMasterTriggerNumberLow  (low);
  this->setMasterTriggerNumberHigh (high);

  return true;
}


// set Method for Trigger Number High
bool RawTriggerHeader::reset(){
  this->resetBufferSource();
  this->clear();
  return true;
}

// clears all data in the buffer
bool RawTriggerHeader::clear(){
  int max_data = this->sizeofdata();
  
  // Loop over the buffer and zero it
  for(int i=0; i< max_data; ++i){
    RAWBUFF32[i]=0;
  }
  return true;
}

// set Method for Trigger Number High
bool RawTriggerHeader::resetBufferSource(){
  this->_Buffer = (void*)_header;
  return true;
}
// set Method for Trigger Number High
bool RawTriggerHeader::setBufferSource(void* input_buffer){
  if(input_buffer){
    this->_Buffer = input_buffer;
    return true;
  }else{
    return false;
  }
  return false;
}

//------------------------------
// INQUIRIY METHODS (CONST)
//

// Method to return the size of the data block in words
bool RawTriggerHeader::isInternalBuffer() const {
  if((this->_Buffer) == (void*)(this->_header)){
    return true;
  }else{
    return false;
  }

  return false;
}


//
// WRITE METHODS
//
// Method to write the data out to a raw buffer
void* RawTriggerHeader::writeData(void* buffer){
  uint32_t* write_buffer;

  if(buffer == 0){
    return 0;
  }

  write_buffer = (uint32_t*)buffer;

  for(uint32_t i = 0; i < this->sizeofdata(); ++i){
    *write_buffer = RAWBUFF32[i];
    write_buffer++;
  } // endfor

  return write_buffer;
}

//
// READ METHODS
//

// Method to read the data in from a raw buffer
void* RawTriggerHeader::readData(void* buffer){
  uint32_t* read_buffer;

  if(buffer == 0){
    return 0;
  }

  read_buffer = (uint32_t*)buffer;

  for(uint32_t i = 0; i < this->sizeofdata(); ++i){
    RAWBUFF32[i] = *read_buffer;
    read_buffer++;
  } // endfor

  return read_buffer;
}

//
// SIZEOF METHODS
//

// Method to return the size of the data block in words
uint32_t RawTriggerHeader::sizeofdata()const{
  return HEADER_WORDS;
}

// Assignment Operator
bool RawTriggerHeader::operator=(const RawTriggerHeader& copy_in) {
  return (this->copy(copy_in));
}

void RawTriggerHeader::print() const{

    printf("++++ Trigger Header ++++\n");
    printf("Trigger Delimiter:\t0x%04x\n",this->getDelimiter() );
    printf("Trigger Version:\t%u\n",this->getVersion() );
    printf("Hardware Flag:\t\t%u\n",this->getHWFlag() );
    printf("Trigger Source:\t\t%u\n",this->getSource() );
    printf("Trigger Source ID:\t%u\n",this->getSourceID() );
    printf("Trigger Source Sub ID:\t%u\n",this->getSourceSubID() );
    printf("Trigger Number Low:\t%u\n",this->getTriggerNumberLow() );
    printf("Trigger Number High:\t%u\n",this->getTriggerNumberHigh() );
    printf("Master Trig Num Low:\t%u\n",this->getMasterTriggerNumberLow() );
    printf("Master Trig Num High:\t%u\n",this->getMasterTriggerNumberHigh() );

}
