#include "DAQDataFormats/RawTriggerMask.h"
#include "DAQDataFormats/DataFormatException.h"
#include <boost/lexical_cast.hpp>
#include <cstdio>


using namespace daqdataformats;

#ifndef RAWBUFF32
// Define a macros to make accessing the raw buffer easier
#define RAWBUFF32 ((uint32_t*)this->_Buffer)
#define RAWBUFF16 ((uint16_t*)this->_Buffer)
#endif
 
RawTriggerMask::RawTriggerMask(){
  init();
}
 
RawTriggerMask::~RawTriggerMask(){
}
 
 
//
// Constructor Methods
//
 
// Constructor with Initialization (non-deep copy)
RawTriggerMask::RawTriggerMask(void* input_buffer, uint32_t size){
  init(input_buffer, size);
  return;
}

// Deep Copy Constructor
RawTriggerMask::RawTriggerMask(const RawTriggerMask& copy_in){
#ifdef COPY_COPIES_INTERNAL
  // Initialize the member data
  this->init();
  // Call the copy method to do the deep copy
  this->copy(copy_in);
#else
  // Check to see what type of buffer we have
  if(copy_in.isInternalBuffer()){
    // Internal to Internal initialization
    this->init();
    this->copy(copy_in);
  }else{
    // Copy_in uses an external buffer so we only 
    // copy the point to the external block and set the
    // size of that buffer
    this->init(copy_in._Buffer, copy_in._max_buffer_size);
  }
#endif
  return;
}
 
// General Initialization 
void RawTriggerMask::init(void* input_buffer, uint32_t size){
  if(input_buffer){
    this->setBufferSource(input_buffer);
  }else{
    this->setBufferSource(_trigMaskblock);
    this->clear();
  }

  if(size){
    if (size < sizeofdata()) {
      std::string msg("An external input buffer was given to the ");
      msg.append("RawTriggerMask constructor that was too small (size=");
      msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
      msg.append(boost::lexical_cast<std::string>(sizeofdata()) + ").");
      GENERATE_DATAFORMAT_EXCEPTION(msg);
    }
    _max_buffer_size = size;
  }else{
    _max_buffer_size = TRIGMASK_WORDS;
  }

  return;
}

// Shallow Copy Method
bool RawTriggerMask::copy(void* input_buffer, uint32_t size){
  // This is a shallow copy, so we only set the pointer
  // to the data buffer and the size of the data if passed
  if(input_buffer){
    this->setBufferSource(input_buffer);
  }else{
    return false;
  }
    
  if(size){
    if (size < sizeofdata()) {
      std::string msg("An external input buffer was given to the ");
      msg.append("RawTriggerMask copy method that was too small (size=");
      msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
      msg.append(boost::lexical_cast<std::string>(sizeofdata()) + ").");
      GENERATE_DATAFORMAT_EXCEPTION(msg);
    }
    _max_buffer_size = size;
  }else{
    _max_buffer_size = TRIGMASK_WORDS;
  }

  return true;
}
 
// Deep Copy Method
bool RawTriggerMask::copy(const RawTriggerMask& copy_in){

  // Copy the max buffer size
  _max_buffer_size = copy_in._max_buffer_size;

  // Copy the data from one buffer to the other
  for(uint32_t i = 0; i < TRIGMASK_WORDS; ++i){
    RAWBUFF32[i] = ((uint32_t*)(copy_in._Buffer))[i];
  } // endfor i
  return true;
}
 
//
// GET METHODS
//
 
// get Method for 
uint32_t RawTriggerMask::getMonteCarloFlag() const {
  return BITFIELDGET(RAWBUFF32[0], MASK_MONTECARLO_FLAG, SHIFT_MONTECARLO_FLAG);
}

// get Method for 
uint32_t RawTriggerMask::getTrigMod() const {
  return BITFIELDGET(RAWBUFF32[0], MASK_TRIG_MOD, SHIFT_TRIG_MOD);
}


// get Method for 
uint32_t RawTriggerMask::getTrigMask_Low() const {
  return BITFIELDGET(RAWBUFF32[0], MASK_TRIG_TYPE_LOW , SHIFT_TRIG_TYPE_LOW);
}

// get Method for 
uint32_t RawTriggerMask::getTrigMask_High() const {
  return BITFIELDGET(RAWBUFF32[1], MASK_TRIG_TYPE_HIGH , SHIFT_TRIG_TYPE_HIGH);
}

// get Method for 
uint32_t RawTriggerMask::getTrigMask_High2() const {
  return BITFIELDGET(RAWBUFF32[2], MASK_TRIG_TYPE_HIGH , SHIFT_TRIG_TYPE_HIGH);
}


uint32_t RawTriggerMask::getTriggerMask() const{

    if     (this->getTrigMask_Low()){
        uint32_t number = this->getTrigMask_Low();
        for(uint32_t ishift = 0; ishift < 32; ++ishift){
            if(number & 0x1 )return ishift;
            number = number >> 1;
        }
        std::string msg("Something is wrong in RawTriggerMask::getTriggerMask()");
        GENERATE_DATAFORMAT_EXCEPTION(msg);
        return 0;
    }
    else if(this->getTrigMask_High()){
        uint32_t number = this->getTrigMask_High();
        for(uint32_t ishift = 0; ishift < 32; ++ishift){
            if(number & 0x1)return ishift + 32;
            number = number >> 1;
        }
        std::string msg("Something is wrong in RawTriggerMask::getTriggerMask()");
        GENERATE_DATAFORMAT_EXCEPTION(msg);
        return 0;
    }
    else if(this->getTrigMask_High2()){
        uint32_t number = this->getTrigMask_High2();
        for(uint32_t ishift = 0; ishift < 32; ++ishift){
            if(number & 0x1)return ishift + 64;
            number = number >> 1;
        }
        std::string msg("Something is wrong in RawTriggerMask::getTriggerMask()");
        GENERATE_DATAFORMAT_EXCEPTION(msg);
        return 0;
    }

    //std::string msg("From RawTriggerMask::getTriggerMask(). TriggerMask is not set");
    //GENERATE_DATAFORMAT_EXCEPTION(msg);
    return 0;
}

// get Method for 
uint32_t RawTriggerMask::getPrescale() const {
  return BITFIELDGET(RAWBUFF32[3], MASK_PRESCALE , SHIFT_PRESCALE);
}

//
// SET METHODS
//
 

// set Method for continous readout state
bool RawTriggerMask::setMonteCarloFlag(uint32_t state){
  BITFIELDSET(RAWBUFF32[0],state,MASK_MONTECARLO_FLAG,SHIFT_MONTECARLO_FLAG);
  return true;
}

// set Method for trigger length field
bool RawTriggerMask::setTrigMod(uint32_t tmod){
  BITFIELDSET(RAWBUFF32[0],tmod,MASK_TRIG_MOD,SHIFT_TRIG_MOD);
  return true;
}

// set Method for trigger mask (low bytes)
bool RawTriggerMask::setTrigMask_Low(uint32_t mask){
  BITFIELDSET(RAWBUFF32[0], mask, MASK_TRIG_TYPE_LOW, SHIFT_TRIG_TYPE_LOW);
  return true;
}

// set Method for trigger mask (high bytes)
bool RawTriggerMask::setTrigMask_High(uint32_t mask){
  BITFIELDSET(RAWBUFF32[1], mask, MASK_TRIG_TYPE_HIGH, SHIFT_TRIG_TYPE_HIGH);
  return true;
}

// set Method for trigger mask (high bytes)
bool RawTriggerMask::setTrigMask_High2(uint32_t mask){
  BITFIELDSET(RAWBUFF32[2], mask, MASK_TRIG_TYPE_HIGH, SHIFT_TRIG_TYPE_HIGH);
  return true;
}


bool RawTriggerMask::setTriggerMask(uint32_t mask){


    if (mask > 88){
        std::string msg("The mask is too high");
        GENERATE_DATAFORMAT_EXCEPTION(msg);
        return false;
    }

    if(mask < 32){
        uint32_t number = 0x1;
        for(uint32_t ishift = 0; ishift < mask; ++ishift){
            number = number << 1;
        }
        this->setTrigMask_Low(number);
        return true;
    }
    else if(mask < 64){
        mask -= 32;
        uint32_t number = 0x1;
        for(uint32_t ishift = 0; ishift < mask; ++ishift){
            number = number << 1;
        }
        this->setTrigMask_High(number);
        return true;
    }

        mask -= 64;
        uint32_t number = 0x1;
        for(uint32_t ishift = 0; ishift < mask; ++ishift){
            number = number << 1;
        }
        this->setTrigMask_High2(number);
        return true;
}

// set Method for trigger mask (high bytes)
bool RawTriggerMask::setPrescale(uint32_t prescale){
  BITFIELDSET(RAWBUFF32[3], prescale, MASK_PRESCALE, SHIFT_PRESCALE);
  return true;
}



// set Method for Trigger Number High
bool RawTriggerMask::reset(){
  this->resetBufferSource();
  this->clear();
  return true;
}

// clears all data in the buffer
bool RawTriggerMask::clear(){
  int max_data = this->sizeofdata();
  
  // Loop over the buffer and zero it
  for(int i=0; i< max_data; ++i){
    RAWBUFF32[i]=0;
  }
  return true;
}

// set Method for Trigger Number High
bool RawTriggerMask::resetBufferSource(){
  this->_Buffer = (void*)_trigMaskblock;
  return true;
}
// set Method for Trigger Number High
bool RawTriggerMask::setBufferSource(void* input_buffer){
  if(input_buffer){
    this->_Buffer = input_buffer;
    return true;
  }else{
    return false;
  }
  return false;
}

//------------------------------
// INQUIRIY METHODS (CONST)
//

// Method to return the size of the data block in words
bool RawTriggerMask::isInternalBuffer() const {
  if((this->_Buffer) == (void*)(this->_trigMaskblock)){
    return true;
  }else{
    return false;
  }

  return false;
}


//
// WRITE METHODS
//
// Method to write the data out to a raw buffer
void* RawTriggerMask::writeData(void* buffer){
  uint32_t* write_buffer;

  if(buffer == 0){
    return 0;
  }

  write_buffer = (uint32_t*)buffer;

  for(uint32_t i = 0; i < this->sizeofdata(); ++i){
    *write_buffer = RAWBUFF32[i];
    write_buffer++;
  } // endfor

  return write_buffer;
}

//
// READ METHODS
//

// Method to read the data in from a raw buffer
void* RawTriggerMask::readData(void* buffer){
  uint32_t* read_buffer;

  if(buffer == 0){
    return 0;
  }

  read_buffer = (uint32_t*)buffer;

  for(uint32_t i = 0; i < this->sizeofdata(); ++i){
    RAWBUFF32[i] = *read_buffer;
    read_buffer++;
  } // endfor

  return read_buffer;
}

//
// SIZEOF METHODS
//

// Method to return the size of the data block in words
uint32_t RawTriggerMask::sizeofdata()const{
  return TRIGMASK_WORDS;
}


// Assignment Operator
bool RawTriggerMask::operator=(const RawTriggerMask& copy_in){
  return (this->copy(copy_in));
}

void RawTriggerMask::print() const{

    printf("++++ Trigger Mask ++++\n");
    printf("Monte Carlo:\t\t%s\n",this->getMonteCarloFlag()? "True" : "False");
    printf("Trigger Modification:\t%u\n",this->getTrigMod() );
    printf("Trigger Types low:\t%u\n",this->getTrigMask_Low() );
    printf("Trigger Types mid:\t%u\n",this->getTrigMask_High() );
    printf("Trigger Types high:\t%u\n",this->getTrigMask_High2() );
    printf("Trigger Type:\t%u\n\n", this->getTriggerMask());
    printf("Trigger prescale:\t%u\n",this->getPrescale() );


}
