#include "DAQDataFormats/RawTriggerRange.h"
#include "DAQDataFormats/DataFormatException.h"
#include <boost/lexical_cast.hpp>
#include <cstdio>
 
using namespace daqdataformats;

#ifndef RAWBUFF32
// Define a macros to make accessing the raw buffer easier
#define RAWBUFF32 ((uint32_t*)this->_Buffer)
#define RAWBUFF16 ((uint16_t*)this->_Buffer)
#endif
 
RawTriggerRange::RawTriggerRange(){
  init();
}
 
RawTriggerRange::~RawTriggerRange() { }
 
 
//
// Constructor Methods
//

// Constructor with Initialization of data buffer to external source
RawTriggerRange::RawTriggerRange(void* input_buffer, uint32_t size){
  init(input_buffer, size);
  return;
}
  
// Deep Copy Constructor
RawTriggerRange::RawTriggerRange(const RawTriggerRange& copy_in){
#ifdef COPY_COPIES_INTERNAL
  // Initialize the member data
  this->init();
  // Call the copy method to do the deep copy
  this->copy(copy_in);
#else
  // Check to see what type of buffer we have
  if(copy_in.isInternalBuffer()){
    // Internal to Internal initialization
    this->init();
    this->copy(copy_in);
  }else{
    // Copy_in uses an external buffer so we only 
    // copy the point to the external block and set the
    // size of that buffer
    init(copy_in._Buffer, copy_in._max_buffer_size);
  }
#endif
  return;
}
 
// General Initialization 
void RawTriggerRange::init(void* input_buffer, uint32_t size){
  if(input_buffer){
    this->setBufferSource(input_buffer);
  }else{
    this->setBufferSource(_trigRangeblock);
    this->clear();
  }

  if(size){
    if (size < sizeofdata()) {
      std::string msg("An external input buffer was given to the ");
      msg.append("RawTriggerRange constructor that was too small (size=");
      msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
      msg.append(boost::lexical_cast<std::string>(sizeofdata()) + ").");
      GENERATE_DATAFORMAT_EXCEPTION(msg);
    }
    _max_buffer_size = size;
  }else{
    _max_buffer_size = TRIGRANGE_WORDS;
  }

  return;
}

// Shallow Copy Method
bool RawTriggerRange::copy(void* input_buffer, uint32_t size){
  // This is a shallow copy, so we only set the pointer
  // to the data buffer and the size of the data if passed
  if(input_buffer){
    this->setBufferSource(input_buffer);
  }else{
    return false;
  }

  if(size){
    if (size < sizeofdata()) {
      std::string msg("An external input buffer was given to the ");
      msg.append("RawTriggerRange copy method that was too small (size=");
      msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
      msg.append(boost::lexical_cast<std::string>(sizeofdata()) + ").");
      GENERATE_DATAFORMAT_EXCEPTION(msg);
    }
    _max_buffer_size = size;
  }else{
    _max_buffer_size = TRIGRANGE_WORDS;
  }
  
  return true;
}
  
// Deep Copy Method
bool RawTriggerRange::copy(const RawTriggerRange& copy_in){

  // Copy the max buffer size
  _max_buffer_size = copy_in._max_buffer_size;

  // Copy the data from one buffer to the other
  for(uint32_t i = 0; i < TRIGRANGE_WORDS; ++i){
    RAWBUFF32[i] = ((uint32_t*)(copy_in._Buffer))[i];
  } // endfor i

  return true;
}

 
//
// GET METHODS
//
 

// get Method for trigger length field
uint32_t RawTriggerRange::getTriggerLength() const {
  return BITFIELDGET(RAWBUFF32[0], MASK_TRIGGER_LENGTH, SHIFT_TRIGGER_LENGTH);
}


//
// SET METHODS
//
 

// set Method for trigger length field
bool RawTriggerRange::setTriggerLength(uint32_t length){
  BITFIELDSET(RAWBUFF32[0], length, MASK_TRIGGER_LENGTH, SHIFT_TRIGGER_LENGTH);
  return true;
}


// set Method for Trigger Number High
bool RawTriggerRange::reset(){
  this->resetBufferSource();
  this->clear();
  return true;
}

// clears all data in the buffer
bool RawTriggerRange::clear(){
  int max_data = this->sizeofdata();
  
  // Loop over the buffer and zero it
  for(int i=0; i< max_data; ++i){
    RAWBUFF32[i]=0;
  }
  return true;
}

// set Method for Trigger Number High
bool RawTriggerRange::resetBufferSource(){
  this->_Buffer = (void*)_trigRangeblock;
  return true;
}
// set Method for Trigger Number High
bool RawTriggerRange::setBufferSource(void* input_buffer){
  if(input_buffer){
    this->_Buffer = input_buffer;
    return true;
  }else{
    return false;
  }
  return false;
}

//------------------------------
// INQUIRIY METHODS (CONST)
//

// Method to return the size of the data block in words
bool RawTriggerRange::isInternalBuffer() const {
  if((this->_Buffer) == (void*)(this->_trigRangeblock)){
    return true;
  }else{
    return false;
  }

  return false;
}


//
// WRITE METHODS
//
// Method to write the data out to a raw buffer
void* RawTriggerRange::writeData(void* buffer){
  uint32_t* write_buffer;

  if(buffer == 0){
    return 0;
  }

  write_buffer = (uint32_t*)buffer;

  for(uint32_t i = 0; i < this->sizeofdata(); ++i){
    *write_buffer = RAWBUFF32[i];
    write_buffer++;
  } // endfor

  return write_buffer;
}

//
// READ METHODS
//

// Method to read the data in from a raw buffer
void* RawTriggerRange::readData(void* buffer){
  uint32_t* read_buffer;

  if(buffer == 0){
    return 0;
  }

  read_buffer = (uint32_t*)buffer;

  for(uint32_t i = 0; i < this->sizeofdata(); ++i){
    RAWBUFF32[i] = *read_buffer;
    read_buffer++;
  } // endfor

  return read_buffer;
}

//
// SIZEOF METHODS
//

// Method to return the size of the data block in words
uint32_t RawTriggerRange::sizeofdata()const{
  return TRIGRANGE_WORDS;
}


// Assignment Operator
bool RawTriggerRange::operator=(const RawTriggerRange& copy_in){
  return (this->copy(copy_in));
}

void RawTriggerRange::print() const{

    printf("++++ Trigger Range ++++\n");
    printf("Trigger range:\t\t%u\n",this->getTriggerLength() );
}
