#include "DAQDataFormats/RawTriggerTime.h"
#include "DAQDataFormats/DataFormatException.h"
#include <boost/lexical_cast.hpp>
#include <cstdio>

#ifndef RAWBUFF32
// Define a macros to make accessing the raw buffer easier
#define RAWBUFF32 ((uint32_t*)this->_Buffer)
#define RAWBUFF16 ((uint16_t*)this->_Buffer)
#endif
 
using namespace daqdataformats;
 
RawTriggerTime::RawTriggerTime(){
  init();
}

RawTriggerTime::~RawTriggerTime(){
}

 
//
// Constructor Methods
//

// Constructor with Initialization of data buffer to external source
RawTriggerTime::RawTriggerTime(void* input_buffer, uint32_t size){
  init(input_buffer, size);
  return;
}
 
// Deep Copy Constructor
RawTriggerTime::RawTriggerTime(const RawTriggerTime& copy_in){

#ifdef COPY_COPIES_INTERNAL
  // Initialize the member data
  this->init();
  // Call the copy method to do the deep copy
  this->copy(copy_in);
#else
  // Check to see what type of buffer we have
  if(copy_in.isInternalBuffer()){
    // Internal to Internal initialization
    this->init();
    this->copy(copy_in);
  }else{
    // Copy_in uses an external buffer so we only 
    // copy the point to the external block and set the
    // size of that buffer
    init(copy_in._Buffer, copy_in._max_buffer_size);
  }
#endif

  return;
}

 
// General Initialization 
void RawTriggerTime::init(void* input_buffer, uint32_t size){
  if(input_buffer){
    this->setBufferSource(input_buffer);
  }else{
    this->setBufferSource(_trigTimeblock);
    this->clear();
  }

  if(size){
    if (size < sizeofdata()) {
      std::string msg("An external input buffer was given to the ");
      msg.append("RawTriggerTime constructor that was too small (size=");
      msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
      msg.append(boost::lexical_cast<std::string>(sizeofdata()) + ").");
      GENERATE_DATAFORMAT_EXCEPTION(msg);
    }
    _max_buffer_size = size;
  }else{
    _max_buffer_size = TIMEBLOCK_WORDS;
  }

  return;
}

// Copy Method (non-deep copy)
bool RawTriggerTime::copy(void* input_buffer, uint32_t size ){
  // Set the general pointer to the buffer only.  Don't copy the data
  // at that location to any local buffers
  if(input_buffer){
    this->setBufferSource(input_buffer);
  }else{
    return false;
  }

  if(size){
    if (size < sizeofdata()) {
      std::string msg("An external input buffer was given to the ");
      msg.append("RawTriggerTime copy method that was too small (size=");
      msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
      msg.append(boost::lexical_cast<std::string>(sizeofdata()) + ").");
      GENERATE_DATAFORMAT_EXCEPTION(msg);
    }
    _max_buffer_size = size;    
  }else{
    _max_buffer_size = TIMEBLOCK_WORDS;
  }

  return true;
}
 
// Deep Copy Method
bool RawTriggerTime::copy(const RawTriggerTime& copy_in){
 
  // Copy the max buffer size
  _max_buffer_size = copy_in._max_buffer_size;

  // Copy the data from one buffer to the other
  for(uint32_t i = 0; i <TIMEBLOCK_WORDS; ++i){
    RAWBUFF32[i] = ((uint32_t*)(copy_in._Buffer))[i];
  } // endfor i
  

  return true;
}
 
//
// GET METHODS
//
 

// get Method for low word of the time stamp
uint32_t RawTriggerTime::getTriggerTimeStampLow() const {
  return BITFIELDGET(RAWBUFF32[WORD_TIMESTAMP_LOW], MASK_TIMESTAMP_LOW, SHIFT_TIMESTAMP_LOW);
}

// get Method for high word of the time stamp
uint32_t RawTriggerTime::getTriggerTimeStampHigh() const {
  return BITFIELDGET(RAWBUFF32[WORD_TIMESTAMP_HIGH], MASK_TIMESTAMP_HIGH, SHIFT_TIMESTAMP_HIGH);
}

// get Method for 64 bit time stamp
uint64_t RawTriggerTime::getTriggerTimeStamp() const {
  uint64_t timestamp;
  register uint32_t low, high;


  low  = this->getTriggerTimeStampLow(); 
  high = this->getTriggerTimeStampHigh(); 
  timestamp=(uint64_t)high<<32 | low;

  return timestamp;
}

// get Method for the the seconds of the timeval
time_t RawTriggerTime::getTriggerTimeStampTvSec() const {
  return BITFIELDGET(RAWBUFF32[WORD_TIMESTAMP_SEC], MASK_TIMESTAMP_SEC, SHIFT_TIMESTAMP_SEC);
}

// get Method for the the microseconds of the timeval
suseconds_t RawTriggerTime::getTriggerTimeStampTvUsec() const {
  return BITFIELDGET(RAWBUFF32[WORD_TIMESTAMP_USEC], MASK_TIMESTAMP_USEC, SHIFT_TIMESTAMP_USEC);
}

// get Method for 64 bit time stamp, in timeval format
timeval RawTriggerTime::getTriggerTimeStampTv() const {
  timeval timestamp;

  timestamp.tv_usec = this->getTriggerTimeStampTvUsec(); 
  timestamp.tv_sec  = this->getTriggerTimeStampTvSec();

  return timestamp;
}

// get Method for clock source
uint32_t RawTriggerTime::getClockSource() const {
  return BITFIELDGET(RAWBUFF32[WORD_CLOCKSOURCE], MASK_CLOCKSOURCE, SHIFT_CLOCKSOURCE);
}

//
// SET METHODS
//
 
// set Method for low word of the time stamp
bool RawTriggerTime::setTriggerTimeStampLow(uint32_t timestamp_low){
  BITFIELDSET(RAWBUFF32[WORD_TIMESTAMP_LOW], timestamp_low, MASK_TIMESTAMP_LOW, SHIFT_TIMESTAMP_LOW);
  return true;
}

// set Method for high word of the time stamp
bool RawTriggerTime::setTriggerTimeStampHigh(uint32_t timestamp_high){
  BITFIELDSET(RAWBUFF32[WORD_TIMESTAMP_HIGH], timestamp_high, MASK_TIMESTAMP_HIGH, SHIFT_TIMESTAMP_HIGH);
  return true;
}

// set Method for 64bit time stamp
bool RawTriggerTime::setTriggerTimeStamp(uint64_t timestamp){
    register uint32_t low, high;

    low  = (uint32_t)timestamp;
    high = timestamp>>32;

    this->setTriggerTimeStampLow(low);
    this->setTriggerTimeStampHigh( high );
    
  return true;
}

// set Method for the timeval seconds
bool RawTriggerTime::setTriggerTimeStampTvSec(time_t timestamp_tv_sec){
  BITFIELDSET(RAWBUFF32[WORD_TIMESTAMP_SEC], timestamp_tv_sec, MASK_TIMESTAMP_SEC, SHIFT_TIMESTAMP_SEC);
  return true;
}

// set Method for the timeval microseconds
bool RawTriggerTime::setTriggerTimeStampTvUsec(suseconds_t timestamp_tv_usec){
  BITFIELDSET(RAWBUFF32[WORD_TIMESTAMP_USEC], timestamp_tv_usec, MASK_TIMESTAMP_USEC, SHIFT_TIMESTAMP_USEC);
  return true;
}

// set Method for 64bit time stamp, as a timeval
bool RawTriggerTime::setTriggerTimeStampTv(timeval timestamp){

    this->setTriggerTimeStampTvUsec(timestamp.tv_usec);
    this->setTriggerTimeStampTvSec(timestamp.tv_sec);
    
  return true;
}

// set Method for Trigger Clock Source
bool RawTriggerTime::setClockSource(trigClockSource source){
  BITFIELDSET(RAWBUFF32[WORD_CLOCKSOURCE], source, MASK_CLOCKSOURCE, SHIFT_CLOCKSOURCE);
  return true;
}

//
// Reset Methods (these methods may clear data)
//

// Reset the class
bool RawTriggerTime::reset(){
  this->resetBufferSource();
  this->clear();
  return true;
}

// clears all data in the buffer
bool RawTriggerTime::clear(){
  int max_data = this->sizeofdata();

  // Loop over the buffer and zero it
  for(int i=0; i< max_data; ++i){
    RAWBUFF32[i]=0;
  }
  return true;
}

// Reset the buffer to use for storage to point to the internal class buffer
bool RawTriggerTime::resetBufferSource(){
  this->setBufferSource(_trigTimeblock); // set the buffer pointer to point internally
  return true;
}

// Set the buffer to use for storage.  This buffer can be external.
bool RawTriggerTime::setBufferSource(void* input_buffer){
  _Buffer = input_buffer;
  return true;
}

//------------------------------
// INQUIRIY METHODS (CONST)
//

// Method to return the size of the data block in words
bool RawTriggerTime::isInternalBuffer() const {
  if((this->_Buffer) == (void*)(this->_trigTimeblock)){
    return true;
  }else{
    return false;
  }

  return false;
}


//
// WRITE METHODS
//
// Method to write the data out to a raw buffer
void* RawTriggerTime::writeData(void* buffer){
  uint32_t* write_buffer;

  if(buffer == 0){
    return 0;
  }

  write_buffer = (uint32_t*)buffer;

  for(uint32_t i = 0; i < this->sizeofdata(); ++i){
    *write_buffer = RAWBUFF32[i];
    write_buffer++;
  } // endfor

  return write_buffer;
}

//
// READ METHODS
//

// Method to read the data in from a raw buffer
void* RawTriggerTime::readData(void* buffer){
  uint32_t* read_buffer;

  if(buffer == 0){
    return 0;
  }

  read_buffer = (uint32_t*)buffer;

  for(uint32_t i = 0; i < this->sizeofdata(); ++i){
    RAWBUFF32[i] = *read_buffer;
    read_buffer++;
  } // endfor

  return read_buffer;
}

//
// SIZEOF METHODS
//

// Method to return the size of the data block in words
uint32_t RawTriggerTime::sizeofdata()const{
  return TIMEBLOCK_WORDS;
}

//
// Operators
//

// Method to return the size of the data block in words
bool RawTriggerTime::operator=(const RawTriggerTime& copy_in){
  return (this->copy(copy_in)); 
}

void RawTriggerTime::print() const{
    printf("++++ Trigger Time ++++\n");
    printf("Gen Timestamp Low:\t%u\n",this->getTriggerTimeStampLow() );
    printf("Gen Timestamp High:\t%u\n",this->getTriggerTimeStampHigh() );
    printf("Gen Timestamp:\t\t%Lu\n",(long long unsigned int)this->getTriggerTimeStamp() );
    time_t TrigSec = this->getTriggerTimeStampTvSec();
    char ctimestr[27]; ctime_r(&TrigSec,ctimestr);
    printf("Gen Time:\t\t%s",ctimestr);
    printf("Gen usec:\t\t%d\n",(uint32_t) this->getTriggerTimeStampTvUsec());
    printf("Clock Source:\t\t0x%02x\n",this->getClockSource() );

}
