#include <cppunit/extensions/HelperMacros.h>
#include <DAQDataFormats/DataFormatException.h>
#include <DAQDataFormats/RawTriggerMask.h>
#include <cstring>  /* memset */

#define PATTERN1 0x01234567
#define PATTERN2 0x76543210
#define PATTERN3 0x89abcdef
#define PATTERN4 0xfedcba98
#define PATTERN5 0xf0f0f0f0
#define PATTERN6 0xe1e1e1e1
#define PATTERN7 0xd2d2d2d2
#define PATTERN8 0xc3c3c3c3
#define PATTERN9 0xb4b4b4b4
#define PATTERN0 0xa5a5a5a5

#define BUFFER_WORDS 1000
#define SMALL_BUFFER_WORDS 1

using namespace daqdataformats;

class RawTriggerMaskTests : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(RawTriggerMaskTests);
    CPPUNIT_TEST(testBasicSetAndGet);
    CPPUNIT_TEST(testSetAndGetWithExternalBuffer);
    CPPUNIT_TEST(testCopyConstructorWithReference);
    CPPUNIT_TEST(testConstructionFromExternalBuffer);
    CPPUNIT_TEST(testNoDeleteOfExternalBuffer);
    CPPUNIT_TEST(testExtBuffToExtBuffCopy);
    CPPUNIT_TEST(testSmallExternalBuffer);
    CPPUNIT_TEST_SUITE_END();

public:
    void setUp();
    void tearDown();

    void testBasicSetAndGet();
    void testSetAndGetWithExternalBuffer();
    void testCopyConstructorWithReference();
    void testConstructionFromExternalBuffer();
    void testNoDeleteOfExternalBuffer();
    void testExtBuffToExtBuffCopy();
    void testSmallExternalBuffer();

private:

};

void
RawTriggerMaskTests::setUp()
{
    // flush stdout to get progress markers to display in realtime
    std::cout.flush();
}

void
RawTriggerMaskTests::tearDown()
{
    // flush stdout to get progress markers to display in realtime
    std::cout.flush();
}

void
RawTriggerMaskTests::testBasicSetAndGet()
{
    RawTriggerMask triggerMask;

    // test that we get the same values as what we set
    triggerMask.setMonteCarloFlag(0);
    CPPUNIT_ASSERT(triggerMask.getMonteCarloFlag() == 0);
    triggerMask.setMonteCarloFlag(1);
    CPPUNIT_ASSERT(triggerMask.getMonteCarloFlag() == 1);

    for (int testValue = 0; testValue < 64; ++testValue) {
        triggerMask.setTrigMod(testValue);
        CPPUNIT_ASSERT(triggerMask.getTrigMod() == testValue);
    }

    triggerMask.setTrigMask_Low(PATTERN3 & 0x00ffffff);
    CPPUNIT_ASSERT(triggerMask.getTrigMask_Low() == (PATTERN3 & 0x00ffffff));

    triggerMask.setTrigMask_High(PATTERN4);
    CPPUNIT_ASSERT(triggerMask.getTrigMask_High() == PATTERN4);

    triggerMask.setTrigMask_High2(PATTERN5);
    CPPUNIT_ASSERT(triggerMask.getTrigMask_High2() == PATTERN5);

    triggerMask.setPrescale(PATTERN6 & 0x7fffffff);
    CPPUNIT_ASSERT(triggerMask.getPrescale() == (PATTERN6 & 0x7fffffff));

    // do things in a different order to verify that there is no
    // cross-talk between the data members
    triggerMask.setMonteCarloFlag(1);
    triggerMask.setTrigMod(22);
    triggerMask.setTrigMask_Low(PATTERN1 & 0x00ffffff);
    triggerMask.setTrigMask_High(PATTERN2);
    triggerMask.setTrigMask_High2(PATTERN8);
    triggerMask.setPrescale(PATTERN9 & 0x7fffffff);

    CPPUNIT_ASSERT(triggerMask.getMonteCarloFlag() == 1);
    CPPUNIT_ASSERT(triggerMask.getTrigMod() == 22);
    CPPUNIT_ASSERT(triggerMask.getTrigMask_Low() == (PATTERN1 & 0x00ffffff));
    CPPUNIT_ASSERT(triggerMask.getTrigMask_High() == PATTERN2);
    CPPUNIT_ASSERT(triggerMask.getTrigMask_High2() == PATTERN8);
    CPPUNIT_ASSERT(triggerMask.getPrescale() == (PATTERN9 & 0x7fffffff));
}

void
RawTriggerMaskTests::testSetAndGetWithExternalBuffer()
{
    // create a trigger mask object based on an external buffer
    uint32_t buffer1[BUFFER_WORDS];
    memset(buffer1, 0, BUFFER_WORDS * sizeof(uint32_t));
    RawTriggerMask triggerMask((void*) buffer1, BUFFER_WORDS);

    // test that we get the same values as what we set
    triggerMask.setMonteCarloFlag(0);
    CPPUNIT_ASSERT(triggerMask.getMonteCarloFlag() == 0);
    triggerMask.setMonteCarloFlag(1);
    CPPUNIT_ASSERT(triggerMask.getMonteCarloFlag() == 1);

    for (int testValue = 0; testValue < 64; ++testValue) {
        triggerMask.setTrigMod(testValue);
        CPPUNIT_ASSERT(triggerMask.getTrigMod() == testValue);
    }

    triggerMask.setTrigMask_Low(PATTERN3 & 0x00ffffff);
    CPPUNIT_ASSERT(triggerMask.getTrigMask_Low() == (PATTERN3 & 0x00ffffff));

    triggerMask.setTrigMask_High(PATTERN4);
    CPPUNIT_ASSERT(triggerMask.getTrigMask_High() == PATTERN4);

    triggerMask.setTrigMask_High2(PATTERN5);
    CPPUNIT_ASSERT(triggerMask.getTrigMask_High2() == PATTERN5);

    triggerMask.setPrescale(PATTERN6 & 0x7fffffff);
    CPPUNIT_ASSERT(triggerMask.getPrescale() == (PATTERN6 & 0x7fffffff));

    // do things in a different order to verify that there is no
    // cross-talk between the data members
    triggerMask.setMonteCarloFlag(1);
    triggerMask.setTrigMod(22);
    triggerMask.setTrigMask_Low(PATTERN1 & 0x00ffffff);
    triggerMask.setTrigMask_High(PATTERN2);
    triggerMask.setTrigMask_High2(PATTERN8);
    triggerMask.setPrescale(PATTERN9 & 0x7fffffff);

    CPPUNIT_ASSERT(triggerMask.getMonteCarloFlag() == 1);
    CPPUNIT_ASSERT(triggerMask.getTrigMod() == 22);
    CPPUNIT_ASSERT(triggerMask.getTrigMask_Low() == (PATTERN1 & 0x00ffffff));
    CPPUNIT_ASSERT(triggerMask.getTrigMask_High() == PATTERN2);
    CPPUNIT_ASSERT(triggerMask.getTrigMask_High2() == PATTERN8);
    CPPUNIT_ASSERT(triggerMask.getPrescale() == (PATTERN9 & 0x7fffffff));

    // verify that we get the expected results after we clear the buffer
    memset(buffer1, 0, BUFFER_WORDS * sizeof(uint32_t));
    CPPUNIT_ASSERT(triggerMask.getMonteCarloFlag() == 0);
    CPPUNIT_ASSERT(triggerMask.getTrigMod() == 0);
    CPPUNIT_ASSERT(triggerMask.getTrigMask_Low() == 0);
    CPPUNIT_ASSERT(triggerMask.getTrigMask_High() == 0);
    CPPUNIT_ASSERT(triggerMask.getTrigMask_High2() == 0);
    CPPUNIT_ASSERT(triggerMask.getPrescale() == 0);
}

void
RawTriggerMaskTests::testCopyConstructorWithReference()
{
    RawTriggerMask triggerMask;

    // set values in the first trigger mask object
    triggerMask.setMonteCarloFlag(1);
    triggerMask.setTrigMod(17);
    triggerMask.setTrigMask_Low(PATTERN3 & 0x00ffffff);
    triggerMask.setTrigMask_High(PATTERN4);
    triggerMask.setTrigMask_High2(PATTERN7);
    triggerMask.setPrescale(PATTERN8 & 0x7fffffff);

    // create a second trigger mask object from the first
    RawTriggerMask maskCopy(triggerMask);

    // reset the values in the first instance
    triggerMask.setMonteCarloFlag(0);
    triggerMask.setTrigMod(0);
    triggerMask.setTrigMask_Low(0);
    triggerMask.setTrigMask_High(0);
    triggerMask.setTrigMask_High2(0);
    triggerMask.setPrescale(0);

    // verify that the copy has the original values
    CPPUNIT_ASSERT(maskCopy.getMonteCarloFlag() == 1);
    CPPUNIT_ASSERT(maskCopy.getTrigMod() == 17);
    CPPUNIT_ASSERT(maskCopy.getTrigMask_Low() == (PATTERN3 & 0x00ffffff));
    CPPUNIT_ASSERT(maskCopy.getTrigMask_High() == PATTERN4);
    CPPUNIT_ASSERT(maskCopy.getTrigMask_High2() == PATTERN7);
    CPPUNIT_ASSERT(maskCopy.getPrescale() == (PATTERN8 & 0x7fffffff));
}

void
RawTriggerMaskTests::testConstructionFromExternalBuffer()
{
    // create a buffer with sample trigger mask data
    uint32_t buffer[BUFFER_WORDS];
    memset(buffer, 0, BUFFER_WORDS * sizeof(uint32_t));
    buffer[0] |= SWAB32(0x80000000);
    buffer[0] |= SWAB32(0x2a000000);
    buffer[0] |= SWAB32(PATTERN1 & 0x00ffffff);
    buffer[1] |= SWAB32(PATTERN2);
    buffer[2] |= SWAB32(PATTERN3);
    buffer[3] |= SWAB32(PATTERN4 & 0x7fffffff);

    // create the the trigger mask object from the buffer
    RawTriggerMask triggerMask((void*) &buffer[0]);

    // verify that the copy has the original values
    CPPUNIT_ASSERT(triggerMask.getMonteCarloFlag() == 1);
    CPPUNIT_ASSERT(triggerMask.getTrigMod() == 42);
    CPPUNIT_ASSERT(triggerMask.getTrigMask_Low() == (PATTERN1 & 0x00ffffff));
    CPPUNIT_ASSERT(triggerMask.getTrigMask_High() == PATTERN2);
    CPPUNIT_ASSERT(triggerMask.getTrigMask_High2() == PATTERN3);
    CPPUNIT_ASSERT(triggerMask.getPrescale() == (PATTERN4 & 0x7fffffff));
}

void
RawTriggerMaskTests::testNoDeleteOfExternalBuffer()
{
    uint32_t* buffer = new uint32_t[BUFFER_WORDS];
    memset(buffer, 0, BUFFER_WORDS * sizeof(uint32_t));

    // create a trigger mask object based on the external buffer
    {
        RawTriggerMask triggerMask((void*) buffer, BUFFER_WORDS);
        triggerMask.setMonteCarloFlag(1);
        triggerMask.setTrigMod(42);
        triggerMask.setTrigMask_Low(PATTERN1 & 0x00ffffff);
        triggerMask.setTrigMask_High(PATTERN2);
        triggerMask.setTrigMask_High2(PATTERN3);
        triggerMask.setPrescale(PATTERN4 & 0x7fffffff);
    }

    // verify that the buffer is still valid after the mask object
    // goes out of scope and is deleted
    CPPUNIT_ASSERT(buffer[0] == SWAB32(0xaa000000 | (PATTERN1 & 0x00ffffff)));
    CPPUNIT_ASSERT(buffer[1] == SWAB32(PATTERN2));
    CPPUNIT_ASSERT(buffer[2] == SWAB32(PATTERN3));
    CPPUNIT_ASSERT(buffer[3] == SWAB32((PATTERN4 & 0x7fffffff)));

    delete[] buffer;
}

void
RawTriggerMaskTests::testExtBuffToExtBuffCopy()
{
    // create two trigger mask objects, both based on external buffers
    uint32_t buffer1[BUFFER_WORDS];
    uint32_t buffer2[BUFFER_WORDS];
    memset(buffer1, 0, BUFFER_WORDS * sizeof(uint32_t));
    memset(buffer2, 0, BUFFER_WORDS * sizeof(uint32_t));
    RawTriggerMask triggerMask((void*) buffer1, BUFFER_WORDS);
    RawTriggerMask maskCopy((void*) buffer2, BUFFER_WORDS);

    // set values in the first trigger mask object
    triggerMask.setMonteCarloFlag(1);
    triggerMask.setTrigMod(17);
    triggerMask.setTrigMask_Low(PATTERN3 & 0x00ffffff);
    triggerMask.setTrigMask_High(PATTERN4);
    triggerMask.setTrigMask_High2(PATTERN7);
    triggerMask.setPrescale(PATTERN8 & 0x7fffffff);

    // copy the data from the first object into the second
    maskCopy.copy(triggerMask);

    // reset the values in the first instance
    triggerMask.setMonteCarloFlag(0);
    triggerMask.setTrigMod(0);
    triggerMask.setTrigMask_Low(0);
    triggerMask.setTrigMask_High(0);
    triggerMask.setTrigMask_High2(0);
    triggerMask.setPrescale(0);

    // verify that the copy has the original values
    CPPUNIT_ASSERT(maskCopy.getMonteCarloFlag() == 1);
    CPPUNIT_ASSERT(maskCopy.getTrigMod() == 17);
    CPPUNIT_ASSERT(maskCopy.getTrigMask_Low() == (PATTERN3 & 0x00ffffff));
    CPPUNIT_ASSERT(maskCopy.getTrigMask_High() == PATTERN4);
    CPPUNIT_ASSERT(maskCopy.getTrigMask_High2() == PATTERN7);
    CPPUNIT_ASSERT(maskCopy.getPrescale() == (PATTERN8 & 0x7fffffff));
}

void
RawTriggerMaskTests::testSmallExternalBuffer()
{
    // create a buffer that it too small to hold mask information
    uint32_t buffer[SMALL_BUFFER_WORDS];
    memset(buffer, 0, SMALL_BUFFER_WORDS * sizeof(uint32_t));

    // create a mask object and expect an exception because the
    // buffer is too small
    try {
        RawTriggerMask triggerMask((void*) buffer, SMALL_BUFFER_WORDS);

        // if we get to the following assert, something failed (there
        // should have been an exception thrown
        CPPUNIT_ASSERT(0);
    }
    // an exception is the expected behavior, so we report success
    catch (DataFormatException& dfExcept) {
        CPPUNIT_ASSERT(1);
    }
}

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION(RawTriggerMaskTests, "DAQDataFormats");
CPPUNIT_REGISTRY_ADD_TO_DEFAULT("DAQDataFormats");
