#include <cppunit/extensions/HelperMacros.h>
#include <DAQDataFormats/DataFormatException.h>
#include <DAQDataFormats/RawTriggerRange.h>
#include <cstring>  /* memset */

#define PATTERN1 0x01234567
#define PATTERN2 0x76543210
#define PATTERN3 0x89abcdef
#define PATTERN4 0xfedcba98
#define PATTERN5 0xf0f0f0f0
#define PATTERN6 0xe1e1e1e1
#define PATTERN7 0xd2d2d2d2
#define PATTERN8 0xc3c3c3c3
#define PATTERN9 0xb4b4b4b4
#define PATTERN0 0xa5a5a5a5

#define BUFFER_WORDS 1000
#define SMALL_BUFFER_WORDS 1

using namespace daqdataformats;

class RawTriggerRangeTests : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(RawTriggerRangeTests);
    CPPUNIT_TEST(testBasicSetAndGet);
    CPPUNIT_TEST(testSetAndGetWithExternalBuffer);
    CPPUNIT_TEST(testCopyConstructorWithReference);
    CPPUNIT_TEST(testConstructionFromExternalBuffer);
    CPPUNIT_TEST(testNoDeleteOfExternalBuffer);
    CPPUNIT_TEST(testNoModsToExternalBuffer);
    CPPUNIT_TEST(testSmallExternalBuffer);
    CPPUNIT_TEST(testExtBuffToExtBuffCopy);
    CPPUNIT_TEST_SUITE_END();

public:
    void setUp();
    void tearDown();

    void testBasicSetAndGet();
    void testSetAndGetWithExternalBuffer();
    void testCopyConstructorWithReference();
    void testConstructionFromExternalBuffer();
    void testNoDeleteOfExternalBuffer();
    void testNoModsToExternalBuffer();
    void testSmallExternalBuffer();
    void testExtBuffToExtBuffCopy();

private:

};

void
RawTriggerRangeTests::setUp()
{
    // flush stdout to get progress markers to display in realtime
    std::cout.flush();
}

void
RawTriggerRangeTests::tearDown()
{
    // flush stdout to get progress markers to display in realtime
    std::cout.flush();
}

void
RawTriggerRangeTests::testBasicSetAndGet()
{
    RawTriggerRange triggerRange;

    // test that we get the same values as what we set
    triggerRange.setTriggerLength(PATTERN9);
    CPPUNIT_ASSERT(triggerRange.getTriggerLength() == PATTERN9);
}

void
RawTriggerRangeTests::testSetAndGetWithExternalBuffer()
{
    // create a trigger range object based on an external buffer
    uint32_t buffer1[BUFFER_WORDS];
    memset(buffer1, 0, BUFFER_WORDS * sizeof(uint32_t));
    RawTriggerRange triggerRange((void*) buffer1, BUFFER_WORDS);

    // test that we get the same values as what we set
    triggerRange.setTriggerLength(PATTERN9);
    CPPUNIT_ASSERT(triggerRange.getTriggerLength() == PATTERN9);

    // verify that we get the expected results after we clear the buffer
    memset(buffer1, 0, BUFFER_WORDS * sizeof(uint32_t));
    CPPUNIT_ASSERT(triggerRange.getTriggerLength() == 0);
}

void
RawTriggerRangeTests::testCopyConstructorWithReference()
{
    RawTriggerRange triggerRange;

    // set values in the first trigger range object
    triggerRange.setTriggerLength(PATTERN9);

    // create a second trigger range object from the first
    RawTriggerRange rangeCopy(triggerRange);

    // reset the values in the first instance
    triggerRange.setTriggerLength(0);

    // verify that the copy has the original values
    CPPUNIT_ASSERT(rangeCopy.getTriggerLength() == PATTERN9);
}

void
RawTriggerRangeTests::testConstructionFromExternalBuffer()
{
    // create a buffer with sample trigger range data
    uint32_t buffer[BUFFER_WORDS];
    memset(buffer, 0, BUFFER_WORDS * sizeof(uint32_t));
    buffer[0] |= SWAB32(PATTERN1);

    // create the the trigger range object from the buffer
    RawTriggerRange triggerRange((void*) &buffer[0]);

    // verify that the copy has the original values
    CPPUNIT_ASSERT(triggerRange.getTriggerLength() == PATTERN1);
}

void
RawTriggerRangeTests::testNoDeleteOfExternalBuffer()
{
    uint32_t* buffer = new uint32_t[BUFFER_WORDS];
    memset(buffer, 0, BUFFER_WORDS * sizeof(uint32_t));

    // create a trigger range object based on the external buffer
    {
        RawTriggerRange triggerRange((void*) buffer, BUFFER_WORDS);
        triggerRange.setTriggerLength(PATTERN1);
    }

    // verify that the buffer is still valid after the range object
    // goes out of scope and is deleted
    CPPUNIT_ASSERT( buffer[0] == SWAB32(PATTERN1) );

    delete[] buffer;
}

void
RawTriggerRangeTests::testNoModsToExternalBuffer()
{
    uint32_t* buffer = new uint32_t[BUFFER_WORDS];

    // initialize the buffer to all zeroes
    memset(buffer, 0, BUFFER_WORDS * sizeof(uint32_t));

    // create a trigger range object based on the external buffer
    {
        RawTriggerRange triggerRange((void*) buffer, BUFFER_WORDS);
    }

    // verify that the buffer is unchanged from its use by the range object
    for (uint32_t idx = 0; idx < BUFFER_WORDS; ++idx) {
        CPPUNIT_ASSERT(buffer[idx] == 0);
    }

    // initialize the buffer to all ones
    memset(buffer, 0xff, BUFFER_WORDS * sizeof(uint32_t));

    // create another trigger range object based on the external buffer
    {
        RawTriggerRange triggerRange((void*) buffer, BUFFER_WORDS);
    }

    // verify that the buffer is unchanged from its use by the range object
    for (uint32_t idx = 0; idx < BUFFER_WORDS; ++idx) {
        CPPUNIT_ASSERT(buffer[idx] == 0xffffffff);
    }

    delete[] buffer;
}

void
RawTriggerRangeTests::testSmallExternalBuffer()
{
    // this test isn't currently useful since the trigger range
    // data is simply one 32-bit word

#if 0
    // create a buffer that it too small to hold range information
    uint32_t buffer[SMALL_BUFFER_WORDS];
    memset(buffer, 0, SMALL_BUFFER_WORDS * sizeof(uint32_t));

    // create a range object and expect an exception because the
    // buffer is too small
    try {
        RawTriggerRange triggerRange((void*) buffer, SMALL_BUFFER_WORDS);

        // if we get to the following assert, something failed (there
        // should have been an exception thrown
        CPPUNIT_ASSERT(0);
    }
    // an exception is the expected behavior, so we report success
    catch (DataFormatException& dfExcept) {
        CPPUNIT_ASSERT(1);
    }
#endif
}

void
RawTriggerRangeTests::testExtBuffToExtBuffCopy()
{
    // create two trigger range objects, both based on external buffers
    uint32_t buffer1[BUFFER_WORDS];
    uint32_t buffer2[BUFFER_WORDS];
    memset(buffer1, 0, BUFFER_WORDS * sizeof(uint32_t));
    memset(buffer2, 0, BUFFER_WORDS * sizeof(uint32_t));
    RawTriggerRange triggerRange((void*) buffer1, BUFFER_WORDS);
    RawTriggerRange rangeCopy((void*) buffer2, BUFFER_WORDS);

    // set values in the first trigger range object
    triggerRange.setTriggerLength(PATTERN9);

    // copy the data from the first object into the second
    rangeCopy.copy(triggerRange);

    // reset the values in the first instance
    triggerRange.setTriggerLength(0);

    // verify that the copy has the original values
    CPPUNIT_ASSERT(rangeCopy.getTriggerLength() == PATTERN9);
}

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION(RawTriggerRangeTests, "DAQDataFormats");
CPPUNIT_REGISTRY_ADD_TO_DEFAULT("DAQDataFormats");
