#include <cppunit/extensions/HelperMacros.h>
#include <DAQDataFormats/DataFormatException.h>
#include <DAQDataFormats/RawTrigger.h>
#include <cstring>  /* memset */

// TODO test assignment/copy of trigger objects.

#define PATTERN1 0x01234567
#define PATTERN2 0x76543210
#define PATTERN3 0x89abcdef
#define PATTERN4 0xfedcba98
#define PATTERN5 0xf0f0f0f0
#define PATTERN6 0xe1e1e1e1
#define PATTERN7 0xd2d2d2d2
#define PATTERN8 0xc3c3c3c3
#define PATTERN9 0xb4b4b4b4
#define PATTERN0 0xa5a5a5a5

#define BUFFER_WORDS 1000
#define SMALL_BUFFER_WORDS 1

using namespace daqdataformats;

class RawTriggerTests : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(RawTriggerTests);
    CPPUNIT_TEST(testBasicSetAndGet);
    CPPUNIT_TEST(testSetAndGetWithExternalBuffer);
    CPPUNIT_TEST(testCopyConstructorWithReference);
    CPPUNIT_TEST(testShallowCopyMethod);
    //CPPUNIT_TEST(testConstructionFromExternalBuffer);
    CPPUNIT_TEST(testNoDeleteOfExternalBuffer);
    CPPUNIT_TEST(testNoModsToExternalBuffer);
    CPPUNIT_TEST(testSmallExternalBuffer);
    //CPPUNIT_TEST(testExtBuffToExtBuffCopy);
    CPPUNIT_TEST(testWriteAndReadData);
    CPPUNIT_TEST_SUITE_END();

public:
    void setUp();
    void tearDown();

    void testBasicSetAndGet();
    void testSetAndGetWithExternalBuffer();
    void testCopyConstructorWithReference();
    void testShallowCopyMethod();
    void testNoDeleteOfExternalBuffer();
    void testNoModsToExternalBuffer();
    void testSmallExternalBuffer();
    void testWriteAndReadData();

private:

};

void
RawTriggerTests::setUp()
{
    // flush stdout to get progress markers to display in realtime
    std::cout.flush();
}

void
RawTriggerTests::tearDown()
{
    // flush stdout to get progress markers to display in realtime
    std::cout.flush();
}

void
RawTriggerTests::testBasicSetAndGet()
{
    RawTrigger rawTrigger;

    uint64_t tmpValue;
    tmpValue = (PATTERN1 & 0x00ffffff);
    uint64_t sampleTime1 = PATTERN2 | (tmpValue << 32);
    tmpValue = (PATTERN3 & 0x00ffffff);
    uint64_t sampleTime2 = PATTERN4 | (tmpValue << 32);

    // ***********************************************
    // *** Create individual trigger components and
    // *** store them in a trigger object.
    // ***********************************************
    {
        // create a timing marker object with non-zero data
        RawTriggerTimingMarker timingMarker;
        timingMarker.setTriggerStart_baseClock(sampleTime1);
        timingMarker.setTriggerStart_833kHz(sampleTime2);

        // add the timing marker to the trigger object
        rawTrigger.setTriggerTimingMarker(timingMarker);

        // clear out the timing marker data to ensure that the
        // trigger object copied the data
        timingMarker.setTriggerStart_baseClock(0);
        timingMarker.setTriggerStart_833kHz(0);
    }
    {
        // create a trigger time object with non-zero data
        RawTriggerTime triggerTime;
        triggerTime.setClockSource(CLOCKSRC_GPS_ASHRIVER_1);
        triggerTime.setTriggerTimeStampLow(PATTERN5);
        triggerTime.setTriggerTimeStampHigh(PATTERN6 & MASK_TIMESTAMP_HIGH);
        triggerTime.setTriggerTimeStampTvUsec(PATTERN7);
        triggerTime.setTriggerTimeStampTvSec(PATTERN8);

        // add the trigger time to the trigger object
        rawTrigger.setTriggerTime(triggerTime);

        // clear out the trigger time data to ensure that the
        // trigger object copied the data
        triggerTime.setClockSource(CLOCKSRC_DUMMY);
        triggerTime.setTriggerTimeStampLow(0);
        triggerTime.setTriggerTimeStampHigh(0);
        triggerTime.setTriggerTimeStampTvUsec(0);
        triggerTime.setTriggerTimeStampTvSec(0);
    }
    {
        // create a trigger range object with non-zero data
        RawTriggerRange triggerRange;
        triggerRange.setTriggerLength(PATTERN0);

        // add the trigger range to the trigger object
        rawTrigger.setTriggerRange(triggerRange);

        // clear out the trigger range data to ensure that the
        // trigger object copied the data
        triggerRange.setTriggerLength(0);
    }
    {
        // create a trigger mask object with non-zero data
        RawTriggerMask triggerMask;
        triggerMask.setMonteCarloFlag(1);
        triggerMask.setTrigMod(22);
        triggerMask.setTrigMask_Low(PATTERN1 & 0x00ffffff);
        triggerMask.setTrigMask_High(PATTERN2);
        triggerMask.setTrigMask_High2(PATTERN8);
        triggerMask.setPrescale(PATTERN9 & 0x7fffffff);

        // add the trigger mask to the trigger object
        rawTrigger.setTriggerMask(triggerMask);

        // clear out the trigger mask data to ensure that the
        // trigger object copied the data
        triggerMask.setMonteCarloFlag(0);
        triggerMask.setTrigMod(0);
        triggerMask.setTrigMask_Low(0);
        triggerMask.setTrigMask_High(0);
        triggerMask.setTrigMask_High2(0);
        triggerMask.setPrescale(0);
    }
    {
        // create a trigger mask object with non-zero data
        RawTriggerHeader triggerHeader;
        triggerHeader.setHWFlag(1);
        triggerHeader.setSource(TRIGSRC_USERDEF);
        triggerHeader.setSourceID(TRIGSRC_ID_BEAMSTAMP_BOOSTER_2);
        triggerHeader.setSourceSubID(TRIGSRC_SUBID_PULSER_C6);
        triggerHeader.setVersion(42);
        triggerHeader.setTriggerNumberLow(PATTERN9);
        triggerHeader.setTriggerNumberHigh(PATTERN0 & 0x000000ff);
        triggerHeader.setDelimiter(PATTERN4 & 0x0000ffff);

        // add the trigger mask to the trigger object
        rawTrigger.setTriggerHeader(triggerHeader);

        // clear out the trigger mask data to ensure that the
        // trigger object copied the data
        triggerHeader.setHWFlag(0);
        triggerHeader.setSource(TRIGSRC_DUMMY);
        triggerHeader.setSourceID(TRIGSRC_ID_DUMMY);
        triggerHeader.setSourceSubID(TRIGSRC_SUBID_DUMMY);
        triggerHeader.setVersion(0);
        triggerHeader.setTriggerNumberLow(0);
        triggerHeader.setTriggerNumberHigh(0);
        triggerHeader.setDelimiter(0);
    }

    // *****************************************************
    // *** Verify that values pulled out of the composite
    // *** trigger object match what we put in.
    // *****************************************************
    {
        RawTriggerTimingMarker* timingMarkerPtr =
            rawTrigger.getTriggerTimingMarker();

        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStart_baseClock() == sampleTime1);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStart_833kHz() == sampleTime2);

        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartLow_baseClock() ==
                       PATTERN2);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartHigh_baseClock() ==
                       (PATTERN1 & 0x00ffffff));
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartLow_833kHz() ==
                       PATTERN4);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartHigh_833kHz() ==
                       (PATTERN3 & 0x00ffffff));
    }
    {
        RawTriggerTime* triggerTimePtr = rawTrigger.getTriggerTime();

        CPPUNIT_ASSERT(triggerTimePtr->getClockSource() ==
                       CLOCKSRC_GPS_ASHRIVER_1);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampLow() == PATTERN5);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampHigh() == 
	               (PATTERN6 & MASK_TIMESTAMP_HIGH));
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampTvUsec() == PATTERN7);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampTvSec() == PATTERN8);
    }
    {
        RawTriggerRange* triggerRangePtr = rawTrigger.getTriggerRange();

        CPPUNIT_ASSERT(triggerRangePtr->getTriggerLength() ==
                       PATTERN0);
    }
    {
        RawTriggerMask* triggerMaskPtr = rawTrigger.getTriggerMask();

        CPPUNIT_ASSERT(triggerMaskPtr->getMonteCarloFlag() == 1);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMod() == 22);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_Low() ==
                       (PATTERN1 & 0x00ffffff));
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_High() == PATTERN2);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_High2() == PATTERN8);
        CPPUNIT_ASSERT(triggerMaskPtr->getPrescale() ==
                       (PATTERN9 & 0x7fffffff));
    }
    {
        RawTriggerHeader* triggerHeaderPtr = rawTrigger.getTriggerHeader();

        CPPUNIT_ASSERT(triggerHeaderPtr->getHWFlag() == 1);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSource() == TRIGSRC_USERDEF);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSourceID() ==
                       TRIGSRC_ID_BEAMSTAMP_BOOSTER_2);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSourceSubID() ==
                       TRIGSRC_SUBID_PULSER_C6);
        CPPUNIT_ASSERT(triggerHeaderPtr->getVersion() == 42);
        CPPUNIT_ASSERT(triggerHeaderPtr->getTriggerNumberLow() == PATTERN9);
        CPPUNIT_ASSERT(triggerHeaderPtr->getTriggerNumberHigh() ==
                       (PATTERN0 & 0x000000ff));
    }
}

void
RawTriggerTests::testSetAndGetWithExternalBuffer()
{
    // create a trigger object based on an external buffer
    uint32_t buffer[BUFFER_WORDS];
    memset(buffer, 0, BUFFER_WORDS * sizeof(uint32_t));
    RawTrigger rawTrigger((void*) buffer, BUFFER_WORDS);

    uint64_t tmpValue;
    tmpValue = (PATTERN1 & 0x00ffffff);
    uint64_t sampleTime1 = PATTERN2 | (tmpValue << 32);
    tmpValue = (PATTERN3 & 0x00ffffff);
    uint64_t sampleTime2 = PATTERN4 | (tmpValue << 32);

    // ***********************************************
    // *** Create individual trigger components and
    // *** store them in a trigger object.
    // ***********************************************
    {
        // create a timing marker object with non-zero data
        RawTriggerTimingMarker timingMarker;
        timingMarker.setTriggerStart_baseClock(sampleTime1);
        timingMarker.setTriggerStart_833kHz(sampleTime2);

        // add the timing marker to the trigger object
        rawTrigger.setTriggerTimingMarker(timingMarker);

        // clear out the timing marker data to ensure that the
        // trigger object copied the data
        timingMarker.setTriggerStart_baseClock(0);
        timingMarker.setTriggerStart_833kHz(0);
    }
    {
        // create a trigger time object with non-zero data
        RawTriggerTime triggerTime;
        triggerTime.setClockSource(CLOCKSRC_GPS_ASHRIVER_1);
        triggerTime.setTriggerTimeStampLow(PATTERN5);
        triggerTime.setTriggerTimeStampHigh(PATTERN6 & MASK_TIMESTAMP_HIGH);
        triggerTime.setTriggerTimeStampTvUsec(PATTERN7);
        triggerTime.setTriggerTimeStampTvSec(PATTERN8);

        // add the trigger time to the trigger object
        rawTrigger.setTriggerTime(triggerTime);

        // clear out the trigger time data to ensure that the
        // trigger object copied the data
        triggerTime.setClockSource(CLOCKSRC_DUMMY);
        triggerTime.setTriggerTimeStampLow(0);
        triggerTime.setTriggerTimeStampHigh(0);
        triggerTime.setTriggerTimeStampTvUsec(0);
        triggerTime.setTriggerTimeStampTvSec(0);
    }
    {
        // create a trigger range object with non-zero data
        RawTriggerRange triggerRange;
        triggerRange.setTriggerLength(PATTERN0);

        // add the trigger range to the trigger object
        rawTrigger.setTriggerRange(triggerRange);

        // clear out the trigger range data to ensure that the
        // trigger object copied the data
        triggerRange.setTriggerLength(0);
    }
    {
        // create a trigger mask object with non-zero data
        RawTriggerMask triggerMask;
        triggerMask.setMonteCarloFlag(1);
        triggerMask.setTrigMod(22);
        triggerMask.setTrigMask_Low(PATTERN1 & 0x00ffffff);
        triggerMask.setTrigMask_High(PATTERN2);
        triggerMask.setTrigMask_High2(PATTERN8);
        triggerMask.setPrescale(PATTERN9 & 0x7fffffff);

        // add the trigger mask to the trigger object
        rawTrigger.setTriggerMask(triggerMask);

        // clear out the trigger mask data to ensure that the
        // trigger object copied the data
        triggerMask.setMonteCarloFlag(0);
        triggerMask.setTrigMod(0);
        triggerMask.setTrigMask_Low(0);
        triggerMask.setTrigMask_High(0);
        triggerMask.setTrigMask_High2(0);
        triggerMask.setPrescale(0);
    }
    {
        // create a trigger mask object with non-zero data
        RawTriggerHeader triggerHeader;
        triggerHeader.setHWFlag(1);
        triggerHeader.setSource(TRIGSRC_USERDEF);
        triggerHeader.setSourceID(TRIGSRC_ID_BEAMSTAMP_BOOSTER_2);
        triggerHeader.setSourceSubID(TRIGSRC_SUBID_PULSER_C6);
        triggerHeader.setVersion(42);
        triggerHeader.setTriggerNumberLow(PATTERN9);
        triggerHeader.setTriggerNumberHigh(PATTERN0 & 0x000000ff);
        triggerHeader.setDelimiter(PATTERN4 & 0x0000ffff);

        // add the trigger mask to the trigger object
        rawTrigger.setTriggerHeader(triggerHeader);

        // clear out the trigger mask data to ensure that the
        // trigger object copied the data
        triggerHeader.setHWFlag(0);
        triggerHeader.setSource(TRIGSRC_DUMMY);
        triggerHeader.setSourceID(TRIGSRC_ID_DUMMY);
        triggerHeader.setSourceSubID(TRIGSRC_SUBID_DUMMY);
        triggerHeader.setVersion(0);
        triggerHeader.setTriggerNumberLow(0);
        triggerHeader.setTriggerNumberHigh(0);
        triggerHeader.setDelimiter(0);
    }

    // *****************************************************
    // *** Verify that values pulled out of the composite
    // *** trigger object match what we put in.
    // *****************************************************
    {
        RawTriggerTimingMarker* timingMarkerPtr =
            rawTrigger.getTriggerTimingMarker();

        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStart_baseClock() == sampleTime1);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStart_833kHz() == sampleTime2);

        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartLow_baseClock() ==
                       PATTERN2);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartHigh_baseClock() ==
                       (PATTERN1 & 0x00ffffff));
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartLow_833kHz() ==
                       PATTERN4);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartHigh_833kHz() ==
                       (PATTERN3 & 0x00ffffff));
    }
    {
        RawTriggerTime* triggerTimePtr = rawTrigger.getTriggerTime();

        CPPUNIT_ASSERT(triggerTimePtr->getClockSource() ==
                       CLOCKSRC_GPS_ASHRIVER_1);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampLow() == PATTERN5);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampHigh() == 
	               (PATTERN6 & MASK_TIMESTAMP_HIGH));
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampTvUsec() == PATTERN7);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampTvSec() == PATTERN8);
    }
    {
        RawTriggerRange* triggerRangePtr = rawTrigger.getTriggerRange();

        CPPUNIT_ASSERT(triggerRangePtr->getTriggerLength() ==
                       PATTERN0);
    }
    {
        RawTriggerMask* triggerMaskPtr = rawTrigger.getTriggerMask();

        CPPUNIT_ASSERT(triggerMaskPtr->getMonteCarloFlag() == 1);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMod() == 22);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_Low() ==
                       (PATTERN1 & 0x00ffffff));
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_High() == PATTERN2);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_High2() == PATTERN8);
        CPPUNIT_ASSERT(triggerMaskPtr->getPrescale() ==
                       (PATTERN9 & 0x7fffffff));
    }
    {
        RawTriggerHeader* triggerHeaderPtr = rawTrigger.getTriggerHeader();

        CPPUNIT_ASSERT(triggerHeaderPtr->getHWFlag() == 1);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSource() == TRIGSRC_USERDEF);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSourceID() ==
                       TRIGSRC_ID_BEAMSTAMP_BOOSTER_2);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSourceSubID() ==
                       TRIGSRC_SUBID_PULSER_C6);
        CPPUNIT_ASSERT(triggerHeaderPtr->getVersion() == 42);
        CPPUNIT_ASSERT(triggerHeaderPtr->getTriggerNumberLow() == PATTERN9);
        CPPUNIT_ASSERT(triggerHeaderPtr->getTriggerNumberHigh() ==
                       (PATTERN0 & 0x000000ff));
    }

    // *****************************************************
    // *** Verify that we get the expected results after
    // *** we clear the buffer
    // *****************************************************
    memset(buffer, 0, BUFFER_WORDS * sizeof(uint32_t));
    {
        RawTriggerTimingMarker* timingMarkerPtr =
            rawTrigger.getTriggerTimingMarker();
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartLow_baseClock() == 0);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartHigh_baseClock() == 0);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartLow_833kHz() == 0);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartHigh_833kHz() == 0);
    }
    {
        RawTriggerTime* triggerTimePtr = rawTrigger.getTriggerTime();
        CPPUNIT_ASSERT(triggerTimePtr->getClockSource() == 0);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampLow() == 0);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampHigh() == 0);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampTvUsec() == 0);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampTvSec() == 0);
    }
    {
        RawTriggerRange* triggerRangePtr = rawTrigger.getTriggerRange();
        CPPUNIT_ASSERT(triggerRangePtr->getTriggerLength() == 0);
    }
    {
        RawTriggerMask* triggerMaskPtr = rawTrigger.getTriggerMask();
        CPPUNIT_ASSERT(triggerMaskPtr->getMonteCarloFlag() == 0);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMod() == 0);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_Low() == 0);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_High() == 0);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_High2() == 0);
        CPPUNIT_ASSERT(triggerMaskPtr->getPrescale() == 0);
    }
    {
        RawTriggerHeader* triggerHeaderPtr = rawTrigger.getTriggerHeader();
        CPPUNIT_ASSERT(triggerHeaderPtr->getHWFlag() == 0);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSource() == 0);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSourceID() == 0);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSourceSubID() == 0);
        CPPUNIT_ASSERT(triggerHeaderPtr->getVersion() == 0);
        CPPUNIT_ASSERT(triggerHeaderPtr->getTriggerNumberLow() == 0);
        CPPUNIT_ASSERT(triggerHeaderPtr->getTriggerNumberHigh() == 0);
    }
}

void
RawTriggerTests::testCopyConstructorWithReference()
{
    RawTrigger rawTrigger;

    uint64_t tmpValue;
    tmpValue = (PATTERN1 & 0x00ffffff);
    uint64_t sampleTime1 = PATTERN2 | (tmpValue << 32);
    tmpValue = (PATTERN3 & 0x00ffffff);
    uint64_t sampleTime2 = PATTERN4 | (tmpValue << 32);

    // set values in the first trigger object
    {
        // create a timing marker object with non-zero data
        RawTriggerTimingMarker timingMarker;
        timingMarker.setTriggerStart_baseClock(sampleTime1);
        timingMarker.setTriggerStart_833kHz(sampleTime2);

        // add the timing marker to the trigger object
        rawTrigger.setTriggerTimingMarker(timingMarker);
    }
    {
        // create a trigger time object with non-zero data
        RawTriggerTime triggerTime;
        triggerTime.setClockSource(CLOCKSRC_GPS_ASHRIVER_1);
        triggerTime.setTriggerTimeStampLow(PATTERN5);
        triggerTime.setTriggerTimeStampHigh(PATTERN6 & MASK_TIMESTAMP_HIGH);
        triggerTime.setTriggerTimeStampTvUsec(PATTERN7);
        triggerTime.setTriggerTimeStampTvSec(PATTERN8);

        // add the trigger time to the trigger object
        rawTrigger.setTriggerTime(triggerTime);
    }
    {
        // create a trigger range object with non-zero data
        RawTriggerRange triggerRange;
        triggerRange.setTriggerLength(PATTERN0);

        // add the trigger range to the trigger object
        rawTrigger.setTriggerRange(triggerRange);
    }
    {
        // create a trigger mask object with non-zero data
        RawTriggerMask triggerMask;
        triggerMask.setMonteCarloFlag(1);
        triggerMask.setTrigMod(22);
        triggerMask.setTrigMask_Low(PATTERN1 & 0x00ffffff);
        triggerMask.setTrigMask_High(PATTERN2);
        triggerMask.setTrigMask_High2(PATTERN8);
        triggerMask.setPrescale(PATTERN9 & 0x7fffffff);

        // add the trigger mask to the trigger object
        rawTrigger.setTriggerMask(triggerMask);
    }
    {
        // create a trigger mask object with non-zero data
        RawTriggerHeader triggerHeader;
        triggerHeader.setHWFlag(1);
        triggerHeader.setSource(TRIGSRC_USERDEF);
        triggerHeader.setSourceID(TRIGSRC_ID_BEAMSTAMP_BOOSTER_2);
        triggerHeader.setSourceSubID(TRIGSRC_SUBID_PULSER_C6);
        triggerHeader.setVersion(42);
        triggerHeader.setTriggerNumberLow(PATTERN9);
        triggerHeader.setTriggerNumberHigh(PATTERN0 & 0x000000ff);
        triggerHeader.setDelimiter(PATTERN4 & 0x0000ffff);

        // add the trigger mask to the trigger object
        rawTrigger.setTriggerHeader(triggerHeader);
    }

    // create a second trigger object from the first
    RawTrigger triggerCopy(rawTrigger);

    // reset the values in the first instance
    {
        RawTriggerTimingMarker* timingMarkerPtr =
            rawTrigger.getTriggerTimingMarker();
        timingMarkerPtr->setTriggerStart_baseClock(0);
        timingMarkerPtr->setTriggerStart_833kHz(0);
    }
    {
        RawTriggerTime* triggerTimePtr = rawTrigger.getTriggerTime();
        triggerTimePtr->setClockSource(CLOCKSRC_DUMMY);
        triggerTimePtr->setTriggerTimeStampLow(0);
        triggerTimePtr->setTriggerTimeStampHigh(0);
        triggerTimePtr->setTriggerTimeStampTvUsec(0);
        triggerTimePtr->setTriggerTimeStampTvSec(0);
    }
    {
        RawTriggerRange* triggerRangePtr = rawTrigger.getTriggerRange();
        triggerRangePtr->setTriggerLength(0);
    }
    {
        RawTriggerMask* triggerMaskPtr = rawTrigger.getTriggerMask();
        triggerMaskPtr->setMonteCarloFlag(0);
        triggerMaskPtr->setTrigMod(0);
        triggerMaskPtr->setTrigMask_Low(0);
        triggerMaskPtr->setTrigMask_High(0);
        triggerMaskPtr->setTrigMask_High2(0);
        triggerMaskPtr->setPrescale(0);
    }
    {
        RawTriggerHeader* triggerHeaderPtr = rawTrigger.getTriggerHeader();
        triggerHeaderPtr->setHWFlag(0);
        triggerHeaderPtr->setSource(TRIGSRC_DUMMY);
        triggerHeaderPtr->setSourceID(TRIGSRC_ID_DUMMY);
        triggerHeaderPtr->setSourceSubID(TRIGSRC_SUBID_DUMMY);
        triggerHeaderPtr->setVersion(0);
        triggerHeaderPtr->setTriggerNumberLow(0);
        triggerHeaderPtr->setTriggerNumberHigh(0);
        triggerHeaderPtr->setDelimiter(0);
    }

    // verify that the copy has the original values
    {
        RawTriggerTimingMarker* timingMarkerPtr =
            triggerCopy.getTriggerTimingMarker();

        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStart_baseClock() == sampleTime1);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStart_833kHz() == sampleTime2);

        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartLow_baseClock() ==
                       PATTERN2);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartHigh_baseClock() ==
                       (PATTERN1 & 0x00ffffff));
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartLow_833kHz() ==
                       PATTERN4);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartHigh_833kHz() ==
                       (PATTERN3 & 0x00ffffff));
    }
    {
        RawTriggerTime* triggerTimePtr = triggerCopy.getTriggerTime();

        CPPUNIT_ASSERT(triggerTimePtr->getClockSource() ==
                       CLOCKSRC_GPS_ASHRIVER_1);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampLow() == PATTERN5);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampHigh() == 
                       (PATTERN6 & MASK_TIMESTAMP_HIGH));
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampTvUsec() == PATTERN7);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampTvSec() == PATTERN8);
    }
    {
        RawTriggerRange* triggerRangePtr = triggerCopy.getTriggerRange();

        CPPUNIT_ASSERT(triggerRangePtr->getTriggerLength() ==
                       PATTERN0);
    }
    {
        RawTriggerMask* triggerMaskPtr = triggerCopy.getTriggerMask();

        CPPUNIT_ASSERT(triggerMaskPtr->getMonteCarloFlag() == 1);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMod() == 22);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_Low() ==
                       (PATTERN1 & 0x00ffffff));
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_High() == PATTERN2);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_High2() == PATTERN8);
        CPPUNIT_ASSERT(triggerMaskPtr->getPrescale() ==
                       (PATTERN9 & 0x7fffffff));
    }
    {
        RawTriggerHeader* triggerHeaderPtr = triggerCopy.getTriggerHeader();

        CPPUNIT_ASSERT(triggerHeaderPtr->getHWFlag() == 1);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSource() == TRIGSRC_USERDEF);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSourceID() ==
                       TRIGSRC_ID_BEAMSTAMP_BOOSTER_2);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSourceSubID() ==
                       TRIGSRC_SUBID_PULSER_C6);
        CPPUNIT_ASSERT(triggerHeaderPtr->getVersion() == 42);
        CPPUNIT_ASSERT(triggerHeaderPtr->getTriggerNumberLow() == PATTERN9);
        CPPUNIT_ASSERT(triggerHeaderPtr->getTriggerNumberHigh() ==
                       (PATTERN0 & 0x000000ff));
    }
}

void
RawTriggerTests::testShallowCopyMethod()
{
    RawTrigger rawTrigger;

    uint64_t tmpValue;
    tmpValue = (PATTERN1 & 0x00ffffff);
    uint64_t sampleTime1 = PATTERN2 | (tmpValue << 32);
    tmpValue = (PATTERN3 & 0x00ffffff);
    uint64_t sampleTime2 = PATTERN4 | (tmpValue << 32);

    // set values in the first trigger object
    {
        // create a timing marker object with non-zero data
        RawTriggerTimingMarker timingMarker;
        timingMarker.setTriggerStart_baseClock(sampleTime1);
        timingMarker.setTriggerStart_833kHz(sampleTime2);

        // add the timing marker to the trigger object
        rawTrigger.setTriggerTimingMarker(timingMarker);
    }
    {
        // create a trigger time object with non-zero data
        RawTriggerTime triggerTime;
        triggerTime.setClockSource(CLOCKSRC_GPS_ASHRIVER_1);
        triggerTime.setTriggerTimeStampLow(PATTERN5);
        triggerTime.setTriggerTimeStampHigh(PATTERN6 & MASK_TIMESTAMP_HIGH);
        triggerTime.setTriggerTimeStampTvUsec(PATTERN7);
        triggerTime.setTriggerTimeStampTvSec(PATTERN8);

        // add the trigger time to the trigger object
        rawTrigger.setTriggerTime(triggerTime);
    }
    {
        // create a trigger range object with non-zero data
        RawTriggerRange triggerRange;
        triggerRange.setTriggerLength(PATTERN0);

        // add the trigger range to the trigger object
        rawTrigger.setTriggerRange(triggerRange);
    }
    {
        // create a trigger mask object with non-zero data
        RawTriggerMask triggerMask;
        triggerMask.setMonteCarloFlag(1);
        triggerMask.setTrigMod(22);
        triggerMask.setTrigMask_Low(PATTERN1 & 0x00ffffff);
        triggerMask.setTrigMask_High(PATTERN2);
        triggerMask.setTrigMask_High2(PATTERN8);
        triggerMask.setPrescale(PATTERN9 & 0x7fffffff);

        // add the trigger mask to the trigger object
        rawTrigger.setTriggerMask(triggerMask);
    }
    {
        // create a trigger mask object with non-zero data
        RawTriggerHeader triggerHeader;
        triggerHeader.setHWFlag(1);
        triggerHeader.setSource(TRIGSRC_USERDEF);
        triggerHeader.setSourceID(TRIGSRC_ID_BEAMSTAMP_BOOSTER_2);
        triggerHeader.setSourceSubID(TRIGSRC_SUBID_PULSER_C6);
        triggerHeader.setVersion(42);
        triggerHeader.setTriggerNumberLow(PATTERN9);
        triggerHeader.setTriggerNumberHigh(PATTERN0 & 0x000000ff);
        triggerHeader.setDelimiter(PATTERN4 & 0x0000ffff);

        // add the trigger mask to the trigger object
        rawTrigger.setTriggerHeader(triggerHeader);
    }

    // create a second trigger object and do the shallow copy
    RawTrigger triggerCopy;
    triggerCopy.copy(rawTrigger.getBuffer(), rawTrigger.sizeofdata());

    // verify that the copy has the same values as the original
    {
        RawTriggerTimingMarker* timingMarkerPtr =
            triggerCopy.getTriggerTimingMarker();

        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStart_baseClock() == sampleTime1);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStart_833kHz() == sampleTime2);

        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartLow_baseClock() ==
                       PATTERN2);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartHigh_baseClock() ==
                       (PATTERN1 & 0x00ffffff));
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartLow_833kHz() ==
                       PATTERN4);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartHigh_833kHz() ==
                       (PATTERN3 & 0x00ffffff));
    }
    {
        RawTriggerTime* triggerTimePtr = triggerCopy.getTriggerTime();

        CPPUNIT_ASSERT(triggerTimePtr->getClockSource() ==
                       CLOCKSRC_GPS_ASHRIVER_1);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampLow() == PATTERN5);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampHigh() == 
                       (PATTERN6 & MASK_TIMESTAMP_HIGH));
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampTvUsec() == PATTERN7);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampTvSec() == PATTERN8);
    }
    {
        RawTriggerRange* triggerRangePtr = triggerCopy.getTriggerRange();

        CPPUNIT_ASSERT(triggerRangePtr->getTriggerLength() ==
                       PATTERN0);
    }
    {
        RawTriggerMask* triggerMaskPtr = triggerCopy.getTriggerMask();

        CPPUNIT_ASSERT(triggerMaskPtr->getMonteCarloFlag() == 1);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMod() == 22);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_Low() ==
                       (PATTERN1 & 0x00ffffff));
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_High() == PATTERN2);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_High2() == PATTERN8);
        CPPUNIT_ASSERT(triggerMaskPtr->getPrescale() ==
                       (PATTERN9 & 0x7fffffff));
    }
    {
        RawTriggerHeader* triggerHeaderPtr = triggerCopy.getTriggerHeader();

        CPPUNIT_ASSERT(triggerHeaderPtr->getHWFlag() == 1);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSource() == TRIGSRC_USERDEF);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSourceID() ==
                       TRIGSRC_ID_BEAMSTAMP_BOOSTER_2);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSourceSubID() ==
                       TRIGSRC_SUBID_PULSER_C6);
        CPPUNIT_ASSERT(triggerHeaderPtr->getVersion() == 42);
        CPPUNIT_ASSERT(triggerHeaderPtr->getTriggerNumberLow() == PATTERN9);
        CPPUNIT_ASSERT(triggerHeaderPtr->getTriggerNumberHigh() ==
                       (PATTERN0 & 0x000000ff));
    }

    // do the copy again with a slightly larger size (to trigger a realloc)
    triggerCopy.copy(rawTrigger.getBuffer(), 16+rawTrigger.sizeofdata());

    // verify that the copy has the same values as the original
    {
        RawTriggerTimingMarker* timingMarkerPtr =
            triggerCopy.getTriggerTimingMarker();

        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStart_baseClock() == sampleTime1);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStart_833kHz() == sampleTime2);

        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartLow_baseClock() ==
                       PATTERN2);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartHigh_baseClock() ==
                       (PATTERN1 & 0x00ffffff));
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartLow_833kHz() ==
                       PATTERN4);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartHigh_833kHz() ==
                       (PATTERN3 & 0x00ffffff));
    }
    {
        RawTriggerTime* triggerTimePtr = triggerCopy.getTriggerTime();

        CPPUNIT_ASSERT(triggerTimePtr->getClockSource() ==
                       CLOCKSRC_GPS_ASHRIVER_1);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampLow() == PATTERN5);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampHigh() == 
	               (PATTERN6 & MASK_TIMESTAMP_HIGH));
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampTvUsec() == PATTERN7);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampTvSec() == PATTERN8);
    }
    {
        RawTriggerRange* triggerRangePtr = triggerCopy.getTriggerRange();

        CPPUNIT_ASSERT(triggerRangePtr->getTriggerLength() ==
                       PATTERN0);
    }
    {
        RawTriggerMask* triggerMaskPtr = triggerCopy.getTriggerMask();

        CPPUNIT_ASSERT(triggerMaskPtr->getMonteCarloFlag() == 1);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMod() == 22);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_Low() ==
                       (PATTERN1 & 0x00ffffff));
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_High() == PATTERN2);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_High2() == PATTERN8);
        CPPUNIT_ASSERT(triggerMaskPtr->getPrescale() ==
                       (PATTERN9 & 0x7fffffff));
    }
    {
        RawTriggerHeader* triggerHeaderPtr = triggerCopy.getTriggerHeader();

        CPPUNIT_ASSERT(triggerHeaderPtr->getHWFlag() == 1);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSource() == TRIGSRC_USERDEF);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSourceID() ==
                       TRIGSRC_ID_BEAMSTAMP_BOOSTER_2);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSourceSubID() ==
                       TRIGSRC_SUBID_PULSER_C6);
        CPPUNIT_ASSERT(triggerHeaderPtr->getVersion() == 42);
        CPPUNIT_ASSERT(triggerHeaderPtr->getTriggerNumberLow() == PATTERN9);
        CPPUNIT_ASSERT(triggerHeaderPtr->getTriggerNumberHigh() ==
                       (PATTERN0 & 0x000000ff));
    }

    // reset the values in the first instance
    {
        RawTriggerTimingMarker* timingMarkerPtr =
            rawTrigger.getTriggerTimingMarker();
        timingMarkerPtr->setTriggerStart_baseClock(0);
        timingMarkerPtr->setTriggerStart_833kHz(0);
    }
    {
        RawTriggerTime* triggerTimePtr = rawTrigger.getTriggerTime();
        triggerTimePtr->setClockSource(CLOCKSRC_DUMMY);
        triggerTimePtr->setTriggerTimeStampLow(0);
        triggerTimePtr->setTriggerTimeStampHigh(0);
        triggerTimePtr->setTriggerTimeStampTvUsec(0);
        triggerTimePtr->setTriggerTimeStampTvSec(0);
    }
    {
        RawTriggerRange* triggerRangePtr = rawTrigger.getTriggerRange();
        triggerRangePtr->setTriggerLength(0);
    }
    {
        RawTriggerMask* triggerMaskPtr = rawTrigger.getTriggerMask();
        triggerMaskPtr->setMonteCarloFlag(0);
        triggerMaskPtr->setTrigMod(0);
        triggerMaskPtr->setTrigMask_Low(0);
        triggerMaskPtr->setTrigMask_High(0);
        triggerMaskPtr->setTrigMask_High2(0);
        triggerMaskPtr->setPrescale(0);
    }
    {
        RawTriggerHeader* triggerHeaderPtr = rawTrigger.getTriggerHeader();
        triggerHeaderPtr->setHWFlag(0);
        triggerHeaderPtr->setSource(TRIGSRC_DUMMY);
        triggerHeaderPtr->setSourceID(TRIGSRC_ID_DUMMY);
        triggerHeaderPtr->setSourceSubID(TRIGSRC_SUBID_DUMMY);
        triggerHeaderPtr->setVersion(0);
        triggerHeaderPtr->setTriggerNumberLow(0);
        triggerHeaderPtr->setTriggerNumberHigh(0);
        triggerHeaderPtr->setDelimiter(0);
    }

    // verify that the copy now has reset values, too
    {
        RawTriggerTimingMarker* timingMarkerPtr =
            triggerCopy.getTriggerTimingMarker();

        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStart_baseClock() == 0);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStart_833kHz() == 0);

        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartLow_baseClock() == 0);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartHigh_baseClock() == 0);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartLow_833kHz() == 0);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartHigh_833kHz() == 0);
    }
    {
        RawTriggerTime* triggerTimePtr = triggerCopy.getTriggerTime();

        CPPUNIT_ASSERT(triggerTimePtr->getClockSource() == CLOCKSRC_DUMMY);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampLow() == 0);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampHigh() == 0);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampTvUsec() == 0);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampTvSec() == 0);
    }
    {
        RawTriggerRange* triggerRangePtr = triggerCopy.getTriggerRange();

        CPPUNIT_ASSERT(triggerRangePtr->getTriggerLength() == 0);
    }
    {
        RawTriggerMask* triggerMaskPtr = triggerCopy.getTriggerMask();

        CPPUNIT_ASSERT(triggerMaskPtr->getMonteCarloFlag() == 0);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMod() == 0);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_Low() == 0);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_High() == 0);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_High2() == 0);
        CPPUNIT_ASSERT(triggerMaskPtr->getPrescale() == 0);
    }
    {
        RawTriggerHeader* triggerHeaderPtr = triggerCopy.getTriggerHeader();

        CPPUNIT_ASSERT(triggerHeaderPtr->getHWFlag() == 0);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSource() == TRIGSRC_DUMMY);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSourceID() == TRIGSRC_ID_DUMMY);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSourceSubID() ==
                       TRIGSRC_SUBID_DUMMY);
        CPPUNIT_ASSERT(triggerHeaderPtr->getVersion() == 0);
        CPPUNIT_ASSERT(triggerHeaderPtr->getTriggerNumberLow() == 0);
        CPPUNIT_ASSERT(triggerHeaderPtr->getTriggerNumberHigh() == 0);
    }
}

void
RawTriggerTests::testNoDeleteOfExternalBuffer()
{
    uint32_t* buffer = new uint32_t[BUFFER_WORDS];
    memset(buffer, 0, BUFFER_WORDS * sizeof(uint32_t));

    // create a trigger object based on the external buffer
    {
        RawTrigger rawTrigger((void*) buffer, BUFFER_WORDS);
        RawTriggerHeader* triggerHeaderPtr = rawTrigger.getTriggerHeader();
        triggerHeaderPtr->setHWFlag(1);
    }

    // verify that the buffer is still valid after the trigger object
    // goes out of scope and is deleted
    CPPUNIT_ASSERT(buffer[1] == SWAB32(0x80000000));

    delete[] buffer;
}

void
RawTriggerTests::testNoModsToExternalBuffer()
{
    uint32_t* buffer = new uint32_t[BUFFER_WORDS];

    // initialize the buffer to all zeroes
    memset(buffer, 0, BUFFER_WORDS * sizeof(uint32_t));

    // create a trigger object based on the external buffer
    {
        RawTrigger rawTrigger((void*) buffer, BUFFER_WORDS);
    }

    // verify that the buffer is unchanged from its use by the trigger object
    for (uint32_t idx = 0; idx < BUFFER_WORDS; ++idx) {
        CPPUNIT_ASSERT(buffer[idx] == 0);
    }

    // initialize the buffer to all ones
    memset(buffer, 0xff, BUFFER_WORDS * sizeof(uint32_t));

    // create another trigger object based on the external buffer
    {
        RawTrigger rawTrigger((void*) buffer, BUFFER_WORDS);
    }

    // verify that the buffer is unchanged from its use by the trigger object
    for (uint32_t idx = 0; idx < BUFFER_WORDS; ++idx) {
        CPPUNIT_ASSERT(buffer[idx] == 0xffffffff);
    }

    delete[] buffer;
}

void
RawTriggerTests::testSmallExternalBuffer()
{
    // create a buffer that it too small to hold trigger information
    uint32_t buffer[SMALL_BUFFER_WORDS];
    memset(buffer, 0, SMALL_BUFFER_WORDS * sizeof(uint32_t));

    // create a trigger object and expect an exception because the
    // buffer is too small
    try {
        RawTrigger trigger((void*) buffer, SMALL_BUFFER_WORDS);

        // if we get to the following assert, something failed (there
        // should have been an exception thrown
        CPPUNIT_ASSERT(0);
    }
    // an exception is the expected behavior, so we report success
    catch (DataFormatException& dfExcept) {
        CPPUNIT_ASSERT(1);
    }
}

void
RawTriggerTests::testWriteAndReadData()
{
    RawTrigger rawTrigger;

    uint64_t tmpValue;
    tmpValue = (PATTERN1 & 0x00ffffff);
    uint64_t sampleTime1 = PATTERN2 | (tmpValue << 32);
    tmpValue = (PATTERN3 & 0x00ffffff);
    uint64_t sampleTime2 = PATTERN4 | (tmpValue << 32);

    // ***********************************************
    // *** Create individual trigger components and
    // *** store them in a trigger object.
    // ***********************************************
    {
        RawTriggerTimingMarker timingMarker;
        timingMarker.setTriggerStart_baseClock(sampleTime1);
        timingMarker.setTriggerStart_833kHz(sampleTime2);

        rawTrigger.setTriggerTimingMarker(timingMarker);
    }
    {
        RawTriggerTime triggerTime;
        triggerTime.setClockSource(CLOCKSRC_GPS_ASHRIVER_1);
        triggerTime.setTriggerTimeStampLow(PATTERN5);
        triggerTime.setTriggerTimeStampHigh(PATTERN6 & MASK_TIMESTAMP_HIGH);
        triggerTime.setTriggerTimeStampTvUsec(PATTERN7);
        triggerTime.setTriggerTimeStampTvSec(PATTERN8);

        rawTrigger.setTriggerTime(triggerTime);
    }
    {
        RawTriggerRange triggerRange;
        triggerRange.setTriggerLength(PATTERN0);

        rawTrigger.setTriggerRange(triggerRange);
    }
    {
        RawTriggerMask triggerMask;
        triggerMask.setMonteCarloFlag(1);
        triggerMask.setTrigMod(22);
        triggerMask.setTrigMask_Low(PATTERN1 & 0x00ffffff);
        triggerMask.setTrigMask_High(PATTERN2);
        triggerMask.setTrigMask_High2(PATTERN8);
        triggerMask.setPrescale(PATTERN9 & 0x7fffffff);

        rawTrigger.setTriggerMask(triggerMask);
    }
    {
        RawTriggerHeader triggerHeader;
        triggerHeader.setHWFlag(1);
        triggerHeader.setSource(TRIGSRC_USERDEF);
        triggerHeader.setSourceID(TRIGSRC_ID_BEAMSTAMP_BOOSTER_2);
        triggerHeader.setSourceSubID(TRIGSRC_SUBID_PULSER_C6);
        triggerHeader.setVersion(42);
        triggerHeader.setTriggerNumberLow(PATTERN9);
        triggerHeader.setTriggerNumberHigh(PATTERN0 & 0x000000ff);
        triggerHeader.setDelimiter(PATTERN4 & 0x0000ffff);

        rawTrigger.setTriggerHeader(triggerHeader);
    }

    // ******************************************************
    // *** Copy the first trigger object to another by
    // *** way of a temporary buffer.  (This also tests
    // *** the writeData and readData methods, of course.)
    // ******************************************************
    uint32_t tmpBuffer[BUFFER_WORDS];
    memset(tmpBuffer, 0, BUFFER_WORDS * sizeof(uint32_t));
    rawTrigger.writeData(tmpBuffer);

    RawTrigger triggerCopy;
    triggerCopy.readData(tmpBuffer);

    //uint32_t* bufPtr = (uint32_t*) rawTrigger.getBuffer();
    //for (int idx = 0; idx < 16; ++idx) {
    //    std::cout << idx << " 0x" << std::hex << bufPtr[idx] << std::dec << std::endl;
    //}

    // **************************************************
    // *** Verify that values pulled out of the copied
    // *** trigger object match what we put in.
    // **************************************************
    {
        RawTriggerTimingMarker* timingMarkerPtr =
            triggerCopy.getTriggerTimingMarker();

        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStart_baseClock() == sampleTime1);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStart_833kHz() == sampleTime2);

        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartLow_baseClock() ==
                       PATTERN2);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartHigh_baseClock() ==
                       (PATTERN1 & 0x00ffffff));
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartLow_833kHz() ==
                       PATTERN4);
        CPPUNIT_ASSERT(timingMarkerPtr->getTriggerStartHigh_833kHz() ==
                       (PATTERN3 & 0x00ffffff));
    }
    {
        RawTriggerTime* triggerTimePtr = triggerCopy.getTriggerTime();

        CPPUNIT_ASSERT(triggerTimePtr->getClockSource() ==
                       CLOCKSRC_GPS_ASHRIVER_1);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampLow() == PATTERN5);
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampHigh() == 
	               (PATTERN6 & MASK_TIMESTAMP_HIGH));
        CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampTvUsec() == PATTERN7);
 CPPUNIT_ASSERT(triggerTimePtr->getTriggerTimeStampTvSec() == PATTERN8);
    }
    {
        RawTriggerRange* triggerRangePtr = triggerCopy.getTriggerRange();

        CPPUNIT_ASSERT(triggerRangePtr->getTriggerLength() ==
                       PATTERN0);
    }
    {
        RawTriggerMask* triggerMaskPtr = triggerCopy.getTriggerMask();

        CPPUNIT_ASSERT(triggerMaskPtr->getMonteCarloFlag() == 1);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMod() == 22);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_Low() ==
                       (PATTERN1 & 0x00ffffff));
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_High() == PATTERN2);
        CPPUNIT_ASSERT(triggerMaskPtr->getTrigMask_High2() == PATTERN8);
        CPPUNIT_ASSERT(triggerMaskPtr->getPrescale() ==
                       (PATTERN9 & 0x7fffffff));
    }
    {
        RawTriggerHeader* triggerHeaderPtr = triggerCopy.getTriggerHeader();

        CPPUNIT_ASSERT(triggerHeaderPtr->getHWFlag() == 1);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSource() == TRIGSRC_USERDEF);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSourceID() ==
                       TRIGSRC_ID_BEAMSTAMP_BOOSTER_2);
        CPPUNIT_ASSERT(triggerHeaderPtr->getSourceSubID() ==
                       TRIGSRC_SUBID_PULSER_C6);
        CPPUNIT_ASSERT(triggerHeaderPtr->getVersion() == 42);
        CPPUNIT_ASSERT(triggerHeaderPtr->getTriggerNumberLow() == PATTERN9);
        CPPUNIT_ASSERT(triggerHeaderPtr->getTriggerNumberHigh() ==
                       (PATTERN0 & 0x000000ff));
    }
}

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION(RawTriggerTests, "DAQDataFormats");
CPPUNIT_REGISTRY_ADD_TO_DEFAULT("DAQDataFormats");
