#include <cppunit/extensions/HelperMacros.h>
#include <DAQDataFormats/DataFormatException.h>
#include <DAQDataFormats/RawTriggerTime.h>
#include <cstring>  /* memset */

#define PATTERN1 0x01234567
#define PATTERN2 0x76543210
#define PATTERN3 0x89abcdef
#define PATTERN4 0xfedcba98
#define PATTERN5 0xf0f0f0f0
#define PATTERN6 0xe1e1e1e1
#define PATTERN7 0xd2d2d2d2
#define PATTERN8 0xc3c3c3c3
#define PATTERN9 0xb4b4b4b4
#define PATTERN0 0xa5a5a5a5

#define BUFFER_WORDS 1000
#define SMALL_BUFFER_WORDS 1

using namespace daqdataformats;

class RawTriggerTimeTests : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(RawTriggerTimeTests);
    CPPUNIT_TEST(testBasicSetAndGet);
    CPPUNIT_TEST(testSetAndGetWithExternalBuffer);
    CPPUNIT_TEST(testCopyConstructorFromReference);
    CPPUNIT_TEST(testConstructionFromExternalBuffer);
    CPPUNIT_TEST(testNoDeleteOfExternalBuffer);
    CPPUNIT_TEST(testExtBuffToExtBuffCopy);
    CPPUNIT_TEST(testSmallExternalBuffer);
    CPPUNIT_TEST_SUITE_END();

public:
    void setUp();
    void tearDown();

    void testBasicSetAndGet();
    void testSetAndGetWithExternalBuffer();
    void testCopyConstructorFromReference();
    void testConstructionFromExternalBuffer();
    void testNoDeleteOfExternalBuffer();
    void testExtBuffToExtBuffCopy();
    void testSmallExternalBuffer();

private:

};

void
RawTriggerTimeTests::setUp()
{
    // flush stdout to get progress markers to display in realtime
    std::cout.flush();
}

void
RawTriggerTimeTests::tearDown()
{
    // flush stdout to get progress markers to display in realtime
    std::cout.flush();
}

void
RawTriggerTimeTests::testBasicSetAndGet()
{
    RawTriggerTime triggerTime;

    // test the clock source
    triggerTime.setClockSource(CLOCKSRC_GPS_FNAL_NUMI_1);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_GPS_FNAL_NUMI_1);
    triggerTime.setClockSource(CLOCKSRC_GPS_FNAL_NUMI_2);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_GPS_FNAL_NUMI_2);
    triggerTime.setClockSource(CLOCKSRC_GPS_FNAL_NUMI_3);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_GPS_FNAL_NUMI_3);
    triggerTime.setClockSource(CLOCKSRC_GPS_ASHRIVER_1);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_GPS_ASHRIVER_1);
    triggerTime.setClockSource(CLOCKSRC_GPS_ASHRIVER_2);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_GPS_ASHRIVER_2);
    // Depricated Clock Sources: See TriggerDefines.h for current list
#if 0
    triggerTime.setClockSource(CLOCKSRC_GPS_ASHRIVER_3);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_GPS_ASHRIVER_3);
    triggerTime.setClockSource(CLOCKSRC_MTU_NEARDET_1);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_MTU_NEARDET_1);
    triggerTime.setClockSource(CLOCKSRC_MTU_NEARDET_2);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_MTU_NEARDET_2);
    triggerTime.setClockSource(CLOCKSRC_MTU_NEARDET_3);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_MTU_NEARDET_3);
#endif
    triggerTime.setClockSource(CLOCKSRC_DUMMY);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_DUMMY);

    // test the low and high timestamps
    triggerTime.setTriggerTimeStampLow(PATTERN1);
    triggerTime.setTriggerTimeStampHigh(PATTERN2 & MASK_TIMESTAMP_HIGH);

    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampLow() == PATTERN1);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampHigh() == 
                   (PATTERN2 & MASK_TIMESTAMP_HIGH));

    // shuffle values
    triggerTime.setTriggerTimeStampLow(PATTERN4);
    triggerTime.setTriggerTimeStampHigh(PATTERN3 & MASK_TIMESTAMP_HIGH);

    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampLow() == PATTERN4);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampHigh() == 
                   (PATTERN3 & MASK_TIMESTAMP_HIGH));

    // try a different order
    triggerTime.setTriggerTimeStampHigh(PATTERN2 & MASK_TIMESTAMP_HIGH);
    triggerTime.setTriggerTimeStampLow(PATTERN3);

    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampHigh() == 
                   (PATTERN2 & MASK_TIMESTAMP_HIGH));
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampLow() == PATTERN3);

    // test a full time value
    uint64_t tmpValue = (PATTERN5 & MASK_TIMESTAMP_HIGH);
    uint64_t fullValue1 = PATTERN6 | (tmpValue << 32);
    triggerTime.setTriggerTimeStamp(fullValue1);

    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStamp() == fullValue1);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampLow() == PATTERN6);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampHigh() ==
                   (PATTERN5 & MASK_TIMESTAMP_HIGH));

    // test the timeval timestamp
    triggerTime.setTriggerTimeStampTvUsec(PATTERN1);
    triggerTime.setTriggerTimeStampTvSec(PATTERN2);

    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvUsec() == PATTERN1);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvSec() == PATTERN2);

    // shuffle values
    triggerTime.setTriggerTimeStampTvUsec(PATTERN4);
    triggerTime.setTriggerTimeStampTvSec(PATTERN3);

    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvUsec() == PATTERN4);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvSec() == PATTERN3);

    // try a different order
    triggerTime.setTriggerTimeStampTvSec(PATTERN2);
    triggerTime.setTriggerTimeStampTvUsec(PATTERN3);

    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvSec() == PATTERN2);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvUsec() == PATTERN3);

    // test a full timeval
    timeval tmpTimeVal;
    tmpTimeVal.tv_sec = PATTERN5;
    tmpTimeVal.tv_usec = PATTERN6;
    triggerTime.setTriggerTimeStampTv(tmpTimeVal);

    // can't compare timevals directly, so need to break it out and
    // compare the parts
    timeval newtmpTimeVal = triggerTime.getTriggerTimeStampTv();
    CPPUNIT_ASSERT(newtmpTimeVal.tv_usec == tmpTimeVal.tv_usec);
    CPPUNIT_ASSERT(newtmpTimeVal.tv_sec == tmpTimeVal.tv_sec);
    // perhaps redundant, but make sure the same data is in there
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvUsec() == PATTERN6);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvSec() == PATTERN5);
}

void
RawTriggerTimeTests::testSetAndGetWithExternalBuffer()
{
    // create a trigger time object based on an external buffer
    uint32_t buffer1[BUFFER_WORDS];
    memset(buffer1, 0, BUFFER_WORDS * sizeof(uint32_t));
    RawTriggerTime triggerTime((void*) buffer1, BUFFER_WORDS);

    // test the clock source
    // Depricated Clock Sources: See TriggerDefines.h for current list
#if 0
    triggerTime.setClockSource(CLOCKSRC_GPS_FNAL_1);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_GPS_FNAL_1);
    triggerTime.setClockSource(CLOCKSRC_GPS_FNAL_2);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_GPS_FNAL_2);
    triggerTime.setClockSource(CLOCKSRC_GPS_FNAL_3);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_GPS_FNAL_3);
    triggerTime.setClockSource(CLOCKSRC_GPS_ASHRIVER_1);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_GPS_ASHRIVER_1);
    triggerTime.setClockSource(CLOCKSRC_GPS_ASHRIVER_2);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_GPS_ASHRIVER_2);
    triggerTime.setClockSource(CLOCKSRC_GPS_ASHRIVER_3);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_GPS_ASHRIVER_3);
    triggerTime.setClockSource(CLOCKSRC_MTU_NEARDET_1);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_MTU_NEARDET_1);
    triggerTime.setClockSource(CLOCKSRC_MTU_NEARDET_2);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_MTU_NEARDET_2);
    triggerTime.setClockSource(CLOCKSRC_MTU_NEARDET_3);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_MTU_NEARDET_3);
    triggerTime.setClockSource(CLOCKSRC_DUMMY);
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_DUMMY);
#endif

    // test the low and high timestamps
    triggerTime.setTriggerTimeStampLow(PATTERN1);
    triggerTime.setTriggerTimeStampHigh(PATTERN2 & MASK_TIMESTAMP_HIGH);

    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampLow() == PATTERN1);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampHigh() == 
                   (PATTERN2 & MASK_TIMESTAMP_HIGH));

    // shuffle values
    triggerTime.setTriggerTimeStampLow(PATTERN4);
    triggerTime.setTriggerTimeStampHigh(PATTERN3 & MASK_TIMESTAMP_HIGH);

    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampLow() == PATTERN4);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampHigh() == 
                   (PATTERN3 & MASK_TIMESTAMP_HIGH));

    // try a different order
    triggerTime.setTriggerTimeStampHigh(PATTERN2 & MASK_TIMESTAMP_HIGH);
    triggerTime.setTriggerTimeStampLow(PATTERN3);

    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampHigh() == 
                   (PATTERN2 & MASK_TIMESTAMP_HIGH));
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampLow() == PATTERN3);

    // test a full time value
    uint64_t tmpValue = (PATTERN5 & MASK_TIMESTAMP_HIGH);
    uint64_t fullValue1 = PATTERN6 | (tmpValue << 32);
    triggerTime.setTriggerTimeStamp(fullValue1);

    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStamp() == fullValue1);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampLow() == PATTERN6);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampHigh() ==
                   (PATTERN5 & MASK_TIMESTAMP_HIGH));

    // test the low and high timestamps, timeval
    triggerTime.setTriggerTimeStampTvUsec(PATTERN1);
    triggerTime.setTriggerTimeStampTvSec(PATTERN2);

    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvUsec() == PATTERN1);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvSec() == PATTERN2);

    // shuffle values
    triggerTime.setTriggerTimeStampTvUsec(PATTERN4);
    triggerTime.setTriggerTimeStampTvSec(PATTERN3);

    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvUsec() == PATTERN4);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvSec() == PATTERN3);

    // try a different order
    triggerTime.setTriggerTimeStampTvSec(PATTERN2);
    triggerTime.setTriggerTimeStampTvUsec(PATTERN3);

    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvSec() == PATTERN2);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvUsec() == PATTERN3);

    // test a full timeval
    timeval tmpTimeVal;
    tmpTimeVal.tv_sec = PATTERN5;
    tmpTimeVal.tv_usec = PATTERN6;
    triggerTime.setTriggerTimeStampTv(tmpTimeVal);

    // can't compare timevals directly, so need to break it out and
    // compare the parts
    timeval newtmpTimeVal = triggerTime.getTriggerTimeStampTv();
    CPPUNIT_ASSERT(newtmpTimeVal.tv_usec == tmpTimeVal.tv_usec);
    CPPUNIT_ASSERT(newtmpTimeVal.tv_sec == tmpTimeVal.tv_sec);
    // perhaps redundant, but make sure the same data is in there
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvUsec() == PATTERN6);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvSec() == PATTERN5);

    // verify that we get the expected results after we clear the buffer
    memset(buffer1, 0, BUFFER_WORDS * sizeof(uint32_t));
    CPPUNIT_ASSERT(triggerTime.getClockSource() == 0);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampHigh() == 0);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampLow() == 0);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvSec() == 0);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvUsec() == 0);
}

void
RawTriggerTimeTests::testCopyConstructorFromReference()
{
    RawTriggerTime triggerTime;

    // set values in the first trigger time object
    triggerTime.setClockSource(CLOCKSRC_GPS_FNAL_NUMI_3);
    triggerTime.setTriggerTimeStampLow(PATTERN5);
    triggerTime.setTriggerTimeStampHigh(PATTERN6 & MASK_TIMESTAMP_HIGH);
    triggerTime.setTriggerTimeStampTvUsec(PATTERN3);
    triggerTime.setTriggerTimeStampTvSec(PATTERN4);

    // create a second trigger time object from the first
    RawTriggerTime timeCopy(triggerTime);

    // reset the values in the first instance
    triggerTime.setClockSource(CLOCKSRC_DUMMY);
    triggerTime.setTriggerTimeStampLow(0);
    triggerTime.setTriggerTimeStampHigh(0);
    triggerTime.setTriggerTimeStampTvUsec(0);
    triggerTime.setTriggerTimeStampTvSec(0);

    // verify that the copy has the original values
    CPPUNIT_ASSERT(timeCopy.getClockSource() == CLOCKSRC_GPS_FNAL_NUMI_3);
    CPPUNIT_ASSERT(timeCopy.getTriggerTimeStampLow() == PATTERN5);
    CPPUNIT_ASSERT(timeCopy.getTriggerTimeStampHigh() == 
                   (PATTERN6 & MASK_TIMESTAMP_HIGH));
    CPPUNIT_ASSERT(timeCopy.getTriggerTimeStampTvUsec() == PATTERN3);
    CPPUNIT_ASSERT(timeCopy.getTriggerTimeStampTvSec() == PATTERN4);
}

void
RawTriggerTimeTests::testConstructionFromExternalBuffer()
{
    // create a buffer with sample trigger time data
    uint32_t buffer[BUFFER_WORDS];
    memset(buffer, 0, BUFFER_WORDS * sizeof(uint32_t));
    buffer[0] |= SWAB32(PATTERN1);
    buffer[1] |= SWAB32(PATTERN2 & MASK_TIMESTAMP_HIGH);
    buffer[2] |= SWAB32(PATTERN3);
    buffer[3] |= SWAB32(PATTERN4);
    buffer[4] |= SWAB32(CLOCKSRC_DUMMY);

    // create the the trigger time object from the buffer
    RawTriggerTime triggerTime((void*) &buffer[0]);

    // verify that the copy has the original values
    CPPUNIT_ASSERT(triggerTime.getClockSource() == CLOCKSRC_DUMMY);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampLow() == PATTERN1);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampHigh() == 
                   (PATTERN2 & MASK_TIMESTAMP_HIGH));
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvSec() == PATTERN3);
    CPPUNIT_ASSERT(triggerTime.getTriggerTimeStampTvUsec() == PATTERN4);
}

void
RawTriggerTimeTests::testNoDeleteOfExternalBuffer()
{
    uint32_t* buffer = new uint32_t[BUFFER_WORDS];
    memset(buffer, 0, BUFFER_WORDS * sizeof(uint32_t));

    // create a trigger time object based on the external buffer
    {
        RawTriggerTime triggerTime((void*) buffer, BUFFER_WORDS);
        triggerTime.setClockSource(CLOCKSRC_DUMMY);
        triggerTime.setTriggerTimeStampLow(PATTERN1);
        triggerTime.setTriggerTimeStampHigh(PATTERN2 & MASK_TIMESTAMP_HIGH);
        triggerTime.setTriggerTimeStampTvSec(PATTERN3);
        triggerTime.setTriggerTimeStampTvUsec(PATTERN4);
    }

    // verify that the buffer is still valid after the time object
    // goes out of scope and is deleted
    CPPUNIT_ASSERT(buffer[0] == SWAB32(PATTERN1));
    CPPUNIT_ASSERT(buffer[1] == SWAB32(PATTERN2 & MASK_TIMESTAMP_HIGH));
    CPPUNIT_ASSERT(buffer[2] == SWAB32(PATTERN3));
    CPPUNIT_ASSERT(buffer[3] == SWAB32(PATTERN4));
    CPPUNIT_ASSERT(buffer[4] == SWAB32(CLOCKSRC_DUMMY));

    delete[] buffer;
}

void
RawTriggerTimeTests::testExtBuffToExtBuffCopy()
{
    // create two trigger time objects, both based on external buffers
    uint32_t buffer1[BUFFER_WORDS];
    uint32_t buffer2[BUFFER_WORDS];
    memset(buffer1, 0, BUFFER_WORDS * sizeof(uint32_t));
    memset(buffer2, 0, BUFFER_WORDS * sizeof(uint32_t));
    RawTriggerTime triggerTime((void*) buffer1, BUFFER_WORDS);
    RawTriggerTime timeCopy((void*) buffer2, BUFFER_WORDS);

    // set values in the first trigger time object
    triggerTime.setClockSource(CLOCKSRC_GPS_FNAL_NUMI_3);
    triggerTime.setTriggerTimeStampLow(PATTERN5);
    triggerTime.setTriggerTimeStampHigh(PATTERN6 & MASK_TIMESTAMP_HIGH);
    triggerTime.setTriggerTimeStampTvUsec(PATTERN3);
    triggerTime.setTriggerTimeStampTvSec(PATTERN4);

    // copy the data from the first object into the second
    timeCopy.copy(triggerTime);

    // reset the values in the first instance
    triggerTime.setClockSource(CLOCKSRC_DUMMY);
    triggerTime.setTriggerTimeStampLow(0);
    triggerTime.setTriggerTimeStampHigh(0);
    triggerTime.setTriggerTimeStampTvUsec(0);
    triggerTime.setTriggerTimeStampTvSec(0);

    // verify that the copy has the original values
    CPPUNIT_ASSERT(timeCopy.getClockSource() == CLOCKSRC_GPS_FNAL_NUMI_3);
    CPPUNIT_ASSERT(timeCopy.getTriggerTimeStampLow() == PATTERN5);
    CPPUNIT_ASSERT(timeCopy.getTriggerTimeStampHigh() == 
                   (PATTERN6 & MASK_TIMESTAMP_HIGH));
    CPPUNIT_ASSERT(timeCopy.getTriggerTimeStampTvUsec() == PATTERN3);
    CPPUNIT_ASSERT(timeCopy.getTriggerTimeStampTvSec() == PATTERN4);
}

void
RawTriggerTimeTests::testSmallExternalBuffer()
{
    // create a buffer that is too small to hold time information
    uint32_t buffer[SMALL_BUFFER_WORDS];
    memset(buffer, 0, SMALL_BUFFER_WORDS * sizeof(uint32_t));

    // create a time object and expect an exception because the
    // buffer is too small
    try {
        RawTriggerTime triggerTime((void*) buffer, SMALL_BUFFER_WORDS);

        // if we get to the following assert, something failed (there
        // should have been an exception thrown
        CPPUNIT_ASSERT(0);
    }
    // an exception is the expected behavior, so we report success
    catch (DataFormatException& dfExcept) {
        CPPUNIT_ASSERT(1);
    }
}

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION(RawTriggerTimeTests, "DAQDataFormats");
CPPUNIT_REGISTRY_ADD_TO_DEFAULT("DAQDataFormats");
