#include <cppunit/extensions/HelperMacros.h>
#include <DAQDataFormats/DataFormatException.h>
#include <DAQDataFormats/RawTriggerTimingMarker.h>
#include <cstring>  /* memset */

#define PATTERN1 0x01234567
#define PATTERN2 0x76543210
#define PATTERN3 0x89abcdef
#define PATTERN4 0xfedcba98
#define PATTERN5 0xf0f0f0f0
#define PATTERN6 0xe1e1e1e1
#define PATTERN7 0xd2d2d2d2
#define PATTERN8 0xc3c3c3c3
#define PATTERN9 0xb4b4b4b4
#define PATTERN0 0xa5a5a5a5

#define BUFFER_WORDS 1000
#define SMALL_BUFFER_WORDS 1

using namespace daqdataformats;

class RawTriggerTimingMarkerTests : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(RawTriggerTimingMarkerTests);
    CPPUNIT_TEST(testBasicSetAndGet);
    CPPUNIT_TEST(testSetAndGetWithExternalBuffer);
    CPPUNIT_TEST(testCopyConstructorFromReference);
    CPPUNIT_TEST(testConstructionFromExternalBuffer);
    CPPUNIT_TEST(testNoDeleteOfExternalBuffer);
    CPPUNIT_TEST(testExtBuffToExtBuffCopy);
    CPPUNIT_TEST(testSmallExternalBuffer);
    CPPUNIT_TEST_SUITE_END();

public:
    void setUp();
    void tearDown();

    void testBasicSetAndGet();
    void testSetAndGetWithExternalBuffer();
    void testCopyConstructorFromReference();
    void testConstructionFromExternalBuffer();
    void testNoDeleteOfExternalBuffer();
    void testExtBuffToExtBuffCopy();
    void testSmallExternalBuffer();

private:

};

void
RawTriggerTimingMarkerTests::setUp()
{
    // flush stdout to get progress markers to display in realtime
    std::cout.flush();
}

void
RawTriggerTimingMarkerTests::tearDown()
{
    // flush stdout to get progress markers to display in realtime
    std::cout.flush();
}

void
RawTriggerTimingMarkerTests::testBasicSetAndGet()
{
    RawTriggerTimingMarker timingMarker;

    // test that we get the same value as what we set
    timingMarker.setTriggerStartLow_baseClock(PATTERN1);
    timingMarker.setTriggerStartHigh_baseClock(PATTERN2 & 0x00ffffff);
    timingMarker.setTriggerStartLow_833kHz(PATTERN3);
    timingMarker.setTriggerStartHigh_833kHz(PATTERN4 & 0x00ffffff);

    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_baseClock() == PATTERN1);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_baseClock() ==
                   (PATTERN2 & 0x00ffffff));
    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_833kHz() == PATTERN3);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_833kHz() ==
                   (PATTERN4 & 0x00ffffff));

    // redo the test with things shuffled a bit
    timingMarker.setTriggerStartLow_baseClock(PATTERN2);
    timingMarker.setTriggerStartHigh_baseClock(PATTERN3 & 0x00ffffff);
    timingMarker.setTriggerStartLow_833kHz(PATTERN4);
    timingMarker.setTriggerStartHigh_833kHz(PATTERN1 & 0x00ffffff);

    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_baseClock() == PATTERN2);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_baseClock() ==
                   (PATTERN3 & 0x00ffffff));
    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_833kHz() == PATTERN4);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_833kHz() ==
                   (PATTERN1 & 0x00ffffff));

    // does the order matter? (hopefully not)
    timingMarker.setTriggerStartHigh_baseClock(PATTERN3 & 0x00ffffff);
    timingMarker.setTriggerStartLow_baseClock(PATTERN4);
    timingMarker.setTriggerStartHigh_833kHz(PATTERN1 & 0x00ffffff);
    timingMarker.setTriggerStartLow_833kHz(PATTERN2);

    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_baseClock() ==
                   (PATTERN3 & 0x00ffffff));
    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_baseClock() == PATTERN4);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_833kHz() ==
                   (PATTERN1 & 0x00ffffff));
    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_833kHz() == PATTERN2);

    // test full values
    uint64_t tmpValue = (PATTERN2 & 0x00ffffff);
    uint64_t fullValue1 = PATTERN1 | (tmpValue << 32);
    tmpValue = (PATTERN3 & 0x00ffffff);
    uint64_t fullValue2 = PATTERN4 | (tmpValue << 32);
    timingMarker.setTriggerStart_baseClock(fullValue1);
    timingMarker.setTriggerStart_833kHz(fullValue2);

    CPPUNIT_ASSERT(timingMarker.getTriggerStart_baseClock() == fullValue1);
    CPPUNIT_ASSERT(timingMarker.getTriggerStart_833kHz() == fullValue2);

    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_baseClock() == PATTERN1);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_baseClock() ==
                   (PATTERN2 & 0x00ffffff));
    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_833kHz() == PATTERN4);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_833kHz() ==
                   (PATTERN3 & 0x00ffffff));
}

void
RawTriggerTimingMarkerTests::testSetAndGetWithExternalBuffer()
{
    // create a timing marker object based on an external buffer
    uint32_t buffer1[BUFFER_WORDS];
    memset(buffer1, 0, BUFFER_WORDS * sizeof(uint32_t));
    RawTriggerTimingMarker timingMarker((void*) buffer1, BUFFER_WORDS);

    // test that we get the same value as what we set
    timingMarker.setTriggerStartLow_baseClock(PATTERN1);
    timingMarker.setTriggerStartHigh_baseClock(PATTERN2 & 0x00ffffff);
    timingMarker.setTriggerStartLow_833kHz(PATTERN3);
    timingMarker.setTriggerStartHigh_833kHz(PATTERN4 & 0x00ffffff);

    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_baseClock() == PATTERN1);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_baseClock() ==
                   (PATTERN2 & 0x00ffffff));
    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_833kHz() == PATTERN3);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_833kHz() ==
                   (PATTERN4 & 0x00ffffff));

    // redo the test with things shuffled a bit
    timingMarker.setTriggerStartLow_baseClock(PATTERN2);
    timingMarker.setTriggerStartHigh_baseClock(PATTERN3 & 0x00ffffff);
    timingMarker.setTriggerStartLow_833kHz(PATTERN4);
    timingMarker.setTriggerStartHigh_833kHz(PATTERN1 & 0x00ffffff);

    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_baseClock() == PATTERN2);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_baseClock() ==
                   (PATTERN3 & 0x00ffffff));
    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_833kHz() == PATTERN4);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_833kHz() ==
                   (PATTERN1 & 0x00ffffff));

    // does the order matter? (hopefully not)
    timingMarker.setTriggerStartHigh_baseClock(PATTERN3 & 0x00ffffff);
    timingMarker.setTriggerStartLow_baseClock(PATTERN4);
    timingMarker.setTriggerStartHigh_833kHz(PATTERN1 & 0x00ffffff);
    timingMarker.setTriggerStartLow_833kHz(PATTERN2);

    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_baseClock() ==
                   (PATTERN3 & 0x00ffffff));
    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_baseClock() == PATTERN4);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_833kHz() ==
                   (PATTERN1 & 0x00ffffff));
    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_833kHz() == PATTERN2);

    // test full values
    uint64_t tmpValue = (PATTERN2 & 0x00ffffff);
    uint64_t fullValue1 = PATTERN1 | (tmpValue << 32);
    tmpValue = (PATTERN3 & 0x00ffffff);
    uint64_t fullValue2 = PATTERN4 | (tmpValue << 32);
    timingMarker.setTriggerStart_baseClock(fullValue1);
    timingMarker.setTriggerStart_833kHz(fullValue2);

    CPPUNIT_ASSERT(timingMarker.getTriggerStart_baseClock() == fullValue1);
    CPPUNIT_ASSERT(timingMarker.getTriggerStart_833kHz() == fullValue2);

    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_baseClock() == PATTERN1);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_baseClock() ==
                   (PATTERN2 & 0x00ffffff));
    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_833kHz() == PATTERN4);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_833kHz() ==
                   (PATTERN3 & 0x00ffffff));

    // verify that we get the expected results after we clear the buffer
    memset(buffer1, 0, BUFFER_WORDS * sizeof(uint32_t));
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_baseClock() == 0);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_baseClock() == 0);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_833kHz() == 0);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_833kHz() == 0);
}

void
RawTriggerTimingMarkerTests::testCopyConstructorFromReference()
{
    RawTriggerTimingMarker timingMarker;

    // set values in the first timing marker object
    timingMarker.setTriggerStartLow_baseClock(PATTERN1);
    timingMarker.setTriggerStartHigh_baseClock(PATTERN2 & 0x00ffffff);
    timingMarker.setTriggerStartLow_833kHz(PATTERN3);
    timingMarker.setTriggerStartHigh_833kHz(PATTERN4 & 0x00ffffff);

    // create a second timing marker object from the first
    RawTriggerTimingMarker timingMarkerCopy(timingMarker);

    // reset the values in the first instance
    timingMarker.setTriggerStartLow_baseClock(0);
    timingMarker.setTriggerStartHigh_baseClock(0);
    timingMarker.setTriggerStartLow_833kHz(0);
    timingMarker.setTriggerStartHigh_833kHz(0);

    // verify that the copy has the original values
    CPPUNIT_ASSERT(timingMarkerCopy.getTriggerStartLow_baseClock() == PATTERN1);
    CPPUNIT_ASSERT(timingMarkerCopy.getTriggerStartHigh_baseClock() ==
                   (PATTERN2 & 0x00ffffff));
    CPPUNIT_ASSERT(timingMarkerCopy.getTriggerStartLow_833kHz() == PATTERN3);
    CPPUNIT_ASSERT(timingMarkerCopy.getTriggerStartHigh_833kHz() ==
                   (PATTERN4 & 0x00ffffff));
}

void
RawTriggerTimingMarkerTests::testConstructionFromExternalBuffer()
{
    // create a buffer with sample timing marker data
    uint32_t buffer[BUFFER_WORDS];
    memset(buffer, 0, BUFFER_WORDS * sizeof(uint32_t));
    buffer[0] |= SWAB32(PATTERN1);
    buffer[1] |= SWAB32(PATTERN2 & 0x00ffffff);
    buffer[2] |= SWAB32(PATTERN3);
    buffer[3] |= SWAB32(PATTERN4 & 0x00ffffff);

    // create the the timing marker object from the buffer
    RawTriggerTimingMarker timingMarker((void*) &buffer[0]);

    // verify that the copy has the original values
    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_baseClock() == PATTERN1);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_baseClock() ==
                   (PATTERN2 & 0x00ffffff));
    CPPUNIT_ASSERT(timingMarker.getTriggerStartLow_833kHz() == PATTERN3);
    CPPUNIT_ASSERT(timingMarker.getTriggerStartHigh_833kHz() ==
                   (PATTERN4 & 0x00ffffff));
}

void
RawTriggerTimingMarkerTests::testNoDeleteOfExternalBuffer()
{
    uint32_t* buffer = new uint32_t[BUFFER_WORDS];
    memset(buffer, 0, BUFFER_WORDS * sizeof(uint32_t));

    // create a trigger timingMarker object based on the external buffer
    {
        RawTriggerTimingMarker timingMarker((void*) buffer, BUFFER_WORDS);
        timingMarker.setTriggerStartLow_baseClock(PATTERN1);
        timingMarker.setTriggerStartHigh_baseClock(PATTERN2 & 0x00ffffff);
        timingMarker.setTriggerStartLow_833kHz(PATTERN3);
        timingMarker.setTriggerStartHigh_833kHz(PATTERN4 & 0x00ffffff);
    }

    // verify that the buffer is still valid after the timingMarker object
    // goes out of scope and is deleted
    CPPUNIT_ASSERT(buffer[0] == SWAB32(PATTERN1));
    CPPUNIT_ASSERT(buffer[1] == SWAB32((PATTERN2 & 0x00ffffff)));
    CPPUNIT_ASSERT(buffer[2] == SWAB32(PATTERN3));
    CPPUNIT_ASSERT(buffer[3] == SWAB32((PATTERN4 & 0x00ffffff)));

    delete[] buffer;
}

void
RawTriggerTimingMarkerTests::testExtBuffToExtBuffCopy()
{
    // create two timing marker objects, both based on external buffers
    uint32_t buffer1[BUFFER_WORDS];
    uint32_t buffer2[BUFFER_WORDS];
    memset(buffer1, 0, BUFFER_WORDS * sizeof(uint32_t));
    memset(buffer2, 0, BUFFER_WORDS * sizeof(uint32_t));
    RawTriggerTimingMarker timingMarker((void*) buffer1, BUFFER_WORDS);
    RawTriggerTimingMarker timingMarkerCopy((void*) buffer2, BUFFER_WORDS);

    // set values in the first timing marker object
    timingMarker.setTriggerStartLow_baseClock(PATTERN1);
    timingMarker.setTriggerStartHigh_baseClock(PATTERN2 & 0x00ffffff);
    timingMarker.setTriggerStartLow_833kHz(PATTERN3);
    timingMarker.setTriggerStartHigh_833kHz(PATTERN4 & 0x00ffffff);

    // copy the data from the first object into the second
    timingMarkerCopy.copy(timingMarker);

    // reset the values in the first instance
    timingMarker.setTriggerStartLow_baseClock(0);
    timingMarker.setTriggerStartHigh_baseClock(0);
    timingMarker.setTriggerStartLow_833kHz(0);
    timingMarker.setTriggerStartHigh_833kHz(0);

    // verify that the copy has the original values
    CPPUNIT_ASSERT(timingMarkerCopy.getTriggerStartLow_baseClock() == PATTERN1);
    CPPUNIT_ASSERT(timingMarkerCopy.getTriggerStartHigh_baseClock() ==
                   (PATTERN2 & 0x00ffffff));
    CPPUNIT_ASSERT(timingMarkerCopy.getTriggerStartLow_833kHz() == PATTERN3);
    CPPUNIT_ASSERT(timingMarkerCopy.getTriggerStartHigh_833kHz() ==
                   (PATTERN4 & 0x00ffffff));
}

void
RawTriggerTimingMarkerTests::testSmallExternalBuffer()
{
    // create a buffer that it too small to hold timingMarker information
    uint32_t buffer[SMALL_BUFFER_WORDS];
    memset(buffer, 0, SMALL_BUFFER_WORDS * sizeof(uint32_t));

    // create a timingMarker object and expect an exception because the
    // buffer is too small
    try {
        RawTriggerTimingMarker timingMarker((void*) buffer, SMALL_BUFFER_WORDS);

        // if we get to the following assert, something failed (there
        // should have been an exception thrown
        CPPUNIT_ASSERT(0);
    }
    // an exception is the expected behavior, so we report success
    catch (DataFormatException& dfExcept) {
        CPPUNIT_ASSERT(1);
    }
}

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION(RawTriggerTimingMarkerTests, "DAQDataFormats");
CPPUNIT_REGISTRY_ADD_TO_DEFAULT("DAQDataFormats");
